/*
 * $Id: NewOMDOMWriter.java,v 1.4 2017/07/04 08:04:36 schreine Exp $
 *
 * Copyright (c) 2001-2004, RIACA, Technische Universiteit Eindhoven (TU/e).
 * All Rights Reserved.
 *
 * ---------------------------------------------------------------------------
 *
 *  The contents of this file are subject to the RIACA Public License
 *  Version 1.0 (the "License"). You may not use this file except in
 *  compliance with the License. A copy of the License is available at
 *  http://www.riaca.win.tue.nl/
 *
 *  Alternatively, the contents of this file may be used under the terms
 *  of the GNU Lesser General Public license (the "LGPL license"), in which 
 *  case the provisions of the LGPL license are applicable instead of those 
 *  above. A copy of the LGPL license is available at http://www.gnu.org/
 *
 *  The Original Code is ROML -- the RIACA OpenMath Library. The Initial
 *  Developer of the Original Code is Manfred N. Riem.  Portions created
 *  by Manfred N. Riem are Copyright (c) 2001. All Rights Reserved.
 *
 *  Contributor(s):
 *
 *      Ernesto Reinaldo Barreiro, Arjeh M. Cohen, Hans Cuypers, Hans Sterk,
 *      Olga Caprotti
 *
 *  Modifications by Wolfgang Schreiner (2005):
 *  
 *      Change constructor to take the DOM document used for node
 *      creation as a parameter as well as namespace URI and prefix.
 *      Fix namespace handling by replacing createElement by createElementNS.
 *      
 *      Fix bug in writeNode() when writing OMBinding.
 *      Fix bug in writeNode() when writing OMAttribution.
 *      
 * ---------------------------------------------------------------------------
 */
package fmrisc.External;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import nl.tue.win.riaca.openmath.lang.OMApplication;
import nl.tue.win.riaca.openmath.lang.OMAttribution;
import nl.tue.win.riaca.openmath.lang.OMBinding;
import nl.tue.win.riaca.openmath.lang.OMByteArray;
import nl.tue.win.riaca.openmath.lang.OMError;
import nl.tue.win.riaca.openmath.lang.OMFloat;
import nl.tue.win.riaca.openmath.lang.OMForeign;
import nl.tue.win.riaca.openmath.lang.OMInteger;
import nl.tue.win.riaca.openmath.lang.OMObject;
import nl.tue.win.riaca.openmath.lang.OMRoot;
import nl.tue.win.riaca.openmath.lang.OMReference;
import nl.tue.win.riaca.openmath.lang.OMString;
import nl.tue.win.riaca.openmath.lang.OMSymbol;
import nl.tue.win.riaca.openmath.lang.OMVariable;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * An OMDOMWriter. <p>
 *
 * @author Manfred N. Riem (mriem@manorrock.org)
 * @version $Revision: 1.4 $
 */
public final class NewOMDOMWriter {
    /**
     * Stores the namespace prefix. <p>
     */
    protected String prefix;
    protected String uri;
    
    /**
     * Stores the document reference we use to create elements. <p>
     */
    protected Document document;

    /**
     * Constructor. 
     * @param document the document.
     * @param uri the uri.
     * @param prefix the prefix.
     */
    public NewOMDOMWriter(Document document, String uri, String prefix) {
      this.document = document;
      this.uri = uri;
      this.prefix = prefix;
    }

    /**
     * Set the namespace prefix. <p>
     *
     * @param newPrefix the namespace prefix to generate.
     */
    public void setNamespacePrefix(String newPrefix) {
        prefix = newPrefix;
    }

    /**
     * Get the namespace prefix. <p>
     *
     * @return the namespace prefix.
     */
    public String getNamespacePrefix() {
        return prefix;
    }

    /**
     * Write out a DOM document. <p>
     *
     * @param object the object to write out.
     * @return the generated DOM document.
     * @throws IOException when a serious error occurs.
     */
    public Document write(OMObject object) throws IOException {
        try {
            if (object instanceof OMRoot) {
                Node node = writeNode(object);
                document.appendChild(node);
            }
            else {
                Node node = document.createElementNS(uri, prefix + ":OMOBJ");
            
                //if (prefix != null) {
                //    node.setPrefix(prefix);
                //}
            
                node.appendChild(writeNode(object));
                document.appendChild(node);
            }
            return document;
        } 
        catch(Exception exception) {
            exception.printStackTrace(System.err);
            throw new IOException("Unable to write out DOM tree:" + 
                exception.getMessage());
        }
    }

    /**
     * Write out a DOM document node. <p>
     *
     * @param object the object to write out.
     * @throws IOException when a serious error occurs.
     * @return the generated DOM node.
     */
    public Node writeNode(OMObject object) throws IOException {
        Node node  = null;
        Node child = null;
        
        if (document != null) {
            /*
             * Create the element with basic data.
             */
            if (object instanceof OMApplication) {
                node = document.createElementNS(uri, prefix + ":OMA");
                
                OMApplication application = (OMApplication) object;
                Enumeration   enumeration = application.getElements().elements();
                
                for (; enumeration.hasMoreElements();) {
                    child = writeNode((OMObject) enumeration.nextElement());
                    node.appendChild(child);
                }
            }

            if (object instanceof OMAttribution) {
                node = document.createElementNS(uri, prefix + ":OMATTR");

                OMAttribution attribution = (OMAttribution)object;
                Hashtable     attributions = attribution.getAttributions();
                Enumeration   keys         = attributions.keys();
                Enumeration   values       = attributions.elements();
                Element       pairs        = document.createElementNS(uri, prefix + ":OMATP");
                
                for (; keys.hasMoreElements();) {
                    OMObject key   = (OMObject) keys.nextElement();
                    OMObject value = (OMObject) values.nextElement();
                    
                    pairs.appendChild(writeNode(key));
                    pairs.appendChild(writeNode(value));
                }

                node.appendChild(pairs);
                node.appendChild(writeNode(attribution.getConstructor()));
            }

            if (object instanceof OMByteArray) {
                node = document.createElementNS(uri, prefix + ":OMB");
                
                OMByteArray byteArray = (OMByteArray) object;
                Node        text      = document.createTextNode(
                    byteArray.getByteArrayAsString());
                
                node.appendChild(text);
            }
            
            if (object instanceof OMBinding) {
                node = document.createElementNS(uri, prefix + ":OMBIND");
                
                OMBinding   binding     = (OMBinding)object;
                Element     variables   = document.createElementNS(uri,
                 prefix + ":OMBVAR");
                Enumeration enumeration = binding.getVariables().elements();
                
                node.appendChild(writeNode(binding.getBinder()));
                
                for (; enumeration.hasMoreElements();) {
                    OMObject element = (OMObject) enumeration.nextElement();
                    variables.appendChild(writeNode(element));
                }
                
                node.appendChild(variables);
                node.appendChild(writeNode(binding.getBody()));
            }
            
            if (object instanceof OMError) {
                node = document.createElementNS(uri, prefix + ":OME");

                OMError     error       = (OMError) object;
                Enumeration enumeration = error.getElements().elements();
                
                node.appendChild(writeNode(error.getSymbol()));
                
                for (; enumeration.hasMoreElements();) {
                    child = writeNode((OMObject) enumeration.nextElement());
                    node.appendChild(child);
                }
            }
            
            if (object instanceof OMFloat) {
                node = document.createElementNS(uri, prefix + ":OMF");
            }
            
            if (object instanceof OMForeign) {
                node = document.createElementNS(uri, prefix + ":OMFOREIGN");
                
                //OMForeign foreign = (OMForeign) object;
            }
            
            if (object instanceof OMInteger) {
                node = document.createElementNS(uri, prefix + ":OMI");
                
                OMInteger integer = (OMInteger) object;
                Node      text    = document.createTextNode(
                    integer.getInteger());
                
                node.appendChild(text);
            }
            
            if (object instanceof OMReference) {
                node = document.createElementNS(uri, prefix + ":OMR");
            }
            
            if (object instanceof OMRoot) {
                node = document.createElementNS(uri, prefix + ":OMOBJ");
            }

            if (object instanceof OMString) {
                node = document.createElementNS(uri, prefix + ":OMSTR");
                
                OMString string = (OMString) object;
                Node     text   = document.createTextNode(string.getString());

                node.appendChild(text);
            }
            
            if (object instanceof OMSymbol) {
                node = document.createElementNS(uri, prefix + ":OMS");
            }
            
            if (object instanceof OMVariable) {
                node = document.createElementNS(uri, prefix + ":OMV");
            }
            
            //if (prefix != null) {
            //    node.setPrefix(prefix);
            //}
            
            /*
             * Add the attributes.
             */
            Hashtable   attributes = object.getAttributes();
            Enumeration keys       = attributes.keys();
            Enumeration values     = attributes.elements();
            Element     element    = (Element) node;
            
            for (; keys.hasMoreElements();) {
                String key   = (String) keys.nextElement();
                String value = (String) values.nextElement();
                element.setAttribute(key, value);
            }
            
            return node;
        }
        
        throw new IOException("Unable to write out DOM node");
    }
}
