// ----------------------------------------------------------------------------
// ClassReader.java
// A reader of a class.
// $Id: ClassReader.java,v 1.12 2017/07/04 08:04:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import org.antlr.runtime.*;

import fmrisc.ProgramExplorer.Parser.*;
import fmrisc.ProgramExplorer.Syntax.Program.*;

/**
 * 
 * A reader of a class.
 *
 */
public class ClassReader
{
  private Source source;
  private ErrorStream error;
  private ClassDeclaration decl;
  
  /****************************************************************************
   * Create a reader of a class declaration.
   * @param source the source for the class declaration.
   * @param error the stream where to report errors.
   **************************************************************************/
  public ClassReader(Source source, ErrorStream error)
  {
    this.source = source;
    this.error = error;
    this.decl = null;
  }
  
  /****************************************************************************
   * Get source for class reader.
   * @return the source.
   ***************************************************************************/
  public Source getSource()
  {
    return source;
  }
  
  /****************************************************************************
   * Indicate whether declaration has been modified.
   * @return true if declaration has been modified.
   ***************************************************************************/
  public boolean isModified()
  {
    return source.isModified();
  }
  
  /***************************************************************************
   * Read the class declaration.
   * @return the declaration.
   **************************************************************************/
  public ClassDeclaration read()
  {
    if (decl != null && !isModified()) 
      return decl;
    decl = null;
    String text = source.getText();
    if (text == null) return decl;
    SourcePosition pos = error.getPosition();
    try
    {
      error.clearErrors(source);
      source.clearPositions();
      int n = error.getErrorNumber();
      CharStream input = new ANTLRStringStream(text);
      MiniJavaLexer lexer = new MiniJavaLexer(input, error, source);
      CommonTokenStream tokens = new CommonTokenStream(lexer);
      MiniJavaParser parser = new MiniJavaParser(tokens, error, source);
      decl = parser.classdecl();
      int m = error.getErrorNumber();
      if (m > n)
      {
        error.explainError("could not parse file \"" + 
            source.getFile().getAbsolutePath() + "\"");
        decl = null;
      }       
    }
    catch (RecognitionException e)
    {
      error.newError("could not parse file \"" + 
          source.getFile().getAbsolutePath() + "\"");
      decl = null;
    }
    error.setPosition(pos);
    return decl;
  }
}
