// ----------------------------------------------------------------------------
// PrettyMathML.java
// Pretty-pinting abstract syntax trees in HTML/MathML.
// $Id: PrettyMathML.java,v 1.18 2012/10/18 16:12:58 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import org.w3c.dom.*;

import fmrisc.ProgramExplorer.InOut.Breaks.Info;
import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;

/**
 * 
 * Pretty-pinting abstract syntax trees in HTML/MathML.
 *
 */
public final class PrettyMathML extends ASTVisitorBase
{
  // see class Prettyprinter for more information.
  
  private final int MAXLEVEL = 100; // maximum indentation level
  private final int MINFREE = 30;   // no further breaks if this is reached
  private final int INDENT = 3;     // amount of indentation per level
  
  private int width;         // the suggested line width
  private Document document; // the document in which to create new nodes
  private Node root;         // the root to append math nodes to
  private Parent mroot;      // the current root math node
  
  private boolean lineMode = false;  // we are in line mode
  private int level = 0;             // the current indentation level
  private int pos = 0;               // the current column position
  
  // the "parameters" of visit()  
  private Info current;   // the information for the currently visited node
  private Parent cparent; // the node to which to currently append information to

  // a reference to the current parent node
  // its content can be updated
  private final static class Parent
  {
    Node node;
    Parent (Node node)
    {
      this.node = node;
    }
  }
  
  /****************************************************************************
   * Create converter that creates DOM nodes in the specified document.
   * @param width the maximum line width (suggestion, not a hard bound).
   * @param document the document in which to create the DOM nodes.
   ***************************************************************************/
  public PrettyMathML(Document document, int width)
  {
    this.document = document;
    this.width = width;
  }

  /****************************************************************************
   * Set pretty appender to initial state.
   ***************************************************************************/
  public void reset()
  {
    lineMode = false;
    pos = 0;
  }
  
  /***************************************************************************
   * Convert formula to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param formula the expression to convert.
   **************************************************************************/
  public void appendFormula(Node parent, Formula formula)
  {
    reset();
    root = parent;
    mroot = new Parent(createMathNode(document));
    parent.appendChild(mroot.node);
    append(mroot, Breaks.compute(formula, true));
  }
    
  /***************************************************************************
   * Convert term to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param term the expression to convert.
   **************************************************************************/
  public void appendTerm(Node parent, Term term)
  {
    reset();
    root = parent;
    mroot = new Parent(createMathNode(document));
    parent.appendChild(mroot.node);
    append(mroot, Breaks.compute(term, true));
  }
  
  /***************************************************************************
   * Convert nameto DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param name the expression to convert.
   **************************************************************************/
  public void appendName(Node parent, Name name)
  {
    reset();
    root = parent;
    mroot = new Parent(createMathNode(document));
    parent.appendChild(mroot.node);
    appendName(mroot, name);
  }
  
  // --------------------------------------------------------------------------
  //
  // auxiliary methods
  //
  // --------------------------------------------------------------------------
  
  /***************************************************************************
   * Create MathML node in document.
   * @param document the document in which to create the node.
   * @return the MathML node.
   **************************************************************************/
  private static Element createMathNode(Document document)
  {
    Element mnode = document.createElement("math");
    mnode.setAttribute("xmlns", "http://www.w3.org/1998/Math/MathML");
    return mnode;
  }
  
  /***************************************************************************
   * Start new line and indent to the appropriate position.
   * @param overflow true iff line is an overflow line
   **************************************************************************/
  private void resetLine(Parent parent, boolean overflow)
  {
    root.appendChild(document.createElement("br"));
    int indent = level*INDENT;
    if (overflow) indent = indent+2*INDENT;
    StringBuffer buffer = new StringBuffer(indent);
    for (int i=0; i<indent; i++)
      buffer.append('\u00A0');
    root.appendChild(document.createTextNode(buffer.toString()));
    mroot.node = createMathNode(document);
    root.appendChild(mroot.node);
    pos = indent;
  }
  
  /***************************************************************************
   * If in block mode, indent to the apropriate position
   * by appending space to the parent node.
   **************************************************************************/
  private void indent(Parent parent)
  {
    int indent = level*INDENT;
    StringBuffer buffer = new StringBuffer(Math.max(0,indent-pos));
    for (int i=pos; i<indent; i++)
      buffer.append('\u00A0');
    String s = buffer.toString();
    appendTextNode(parent, s, s.length());
    pos = indent;
  }
  
  /****************************************************************************
   * Change indentation level.
   * @param increase true if level is increased (decreased otherwise)
   * @param parent the parent node
   ***************************************************************************/
  private void changeLevel(boolean increase, Parent parent)
  {
    if (lineMode) return;
    if (increase) 
    {
      level++;
      indent(parent);
    }
    else
      level--;
  }
  
  /***************************************************************************
   * Start new (non-overflow)line and indent it appropriately.
   * @param parent the parent node.
   ***************************************************************************/
  private void newLine(Parent parent)
  {
    if (lineMode) return;
    resetLine(parent, false);
  }
  
  // --------------------------------------------------------------------------
  //
  // append() methods
  //
  // --------------------------------------------------------------------------
  
  /***************************************************************************
   * Create new OMDoc node with denoted name and append it to parent node.
   * @param parent the parent of the new node.
   * @param name the name of the new node.
   * @return the new node.
   **************************************************************************/
  private Element appendNode(Parent parent, String name)
  {
    Element node = document.createElement(name);
    parent.node.appendChild(node);
    return node;
  }
  
  /***************************************************************************
   * Create new DOM text node with given text and append it into parent node.
   * @param parent the parent of the new node.
   * @param text the text of the node.
   * @param len the length which the text is assumed to consume
   *        (typically text.length() but may be larger to make up for
   *         surrounding space).
   **************************************************************************/
  private void appendTextNode(Parent parent, String text, int len)
  {
    if (width-pos < len) 
    {
      resetLine(parent, true);
    }
    Node node = document.createTextNode(text);
    parent.node.appendChild(node);
    pos = pos+len;
  }
  
  /***************************************************************************
   * Append non-breakable space to parent node.
   * @param parent the parent node.
   **************************************************************************/
  private void appendSpaceSeparator(Parent parent)
  {
    appendTextNode(parent, "\u00a0", 1);
  }
  
  /***************************************************************************
   * Convert expression to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param exp the expression to be converted.
   ***************************************************************************/
  private void appendNumber(Parent parent, IntLiteral object)
  {
    Node node = appendNode(parent, "mn");
    String s = object.getName();
    appendTextNode(new Parent(node), s, s.length());
  }
  
  /***************************************************************************
   * Convert identifier to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param name the name of the identifier to be converted.
   ***************************************************************************/
  private void appendIdentifier(Parent parent, String name)
  {
    Element node = appendNode(parent, "mi");
    appendTextNode(new Parent(node), name, name.length());
  }
  
  /***************************************************************************
   * Convert identifier to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param name the name of the identifier to be converted.
   ***************************************************************************/
  private void appendConstant(Parent parent, String name)
  {
    Element node = appendNode(parent, "mi");
    appendTextNode(new Parent(node), name, name.length());
    node.setAttribute("mathcolor", "blue");
  }
  
  /***************************************************************************
   * Convert name to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param name the name to be converted.
   ***************************************************************************/
  private void appendName(Parent parent, Name name)
  {
    Identifier[] idents = name.idents;
    int n = idents.length;
    for (int i=0; i<n; i++)
    {
      appendIdentifier(parent, idents[i]);
      if (i < n-1) appendOperator(parent, ".");
    }
  }
  
  /***************************************************************************
   * Convert identifier to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param object the identifier to be converted.
   ***************************************************************************/
  private void appendIdentifier(Parent parent, Identifier object)
  {
    String name = object.toString();
    int p = name.indexOf('_');
    if (p == -1 || p == 0 || p == name.length()-1)
    {
      appendIdentifier(parent, name);
      return;
    }
    String base = name.substring(0, p);
    String index = name.substring(p+1);
    Element node = appendNode(parent, "msub");
    appendIdentifier(new Parent(node), base);
    appendIdentifier(new Parent(node), index);
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   * @return the created node.
   **************************************************************************/
  private Element appendOperator(Parent parent, String text)
  {
    Element node = appendNode(parent, "mo");
    appendTextNode(new Parent(node), text, text.length());
    return node;
  }
 
  /***************************************************************************
   * Append to parent operator node with denoted text assuming that it
   * takes some more space than denoted by the text itself (2 chars extra).
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   * @return the created node.
   **************************************************************************/
  private Element appendWideOperator(Parent parent, String text)
  {
    Element node = appendNode(parent, "mo");
    appendTextNode(new Parent(node), text, text.length()+1);
    return node;
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   * @return the created node.
   **************************************************************************/
  private Element appendPrefixFunction(Parent parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "1px");
    node.setAttribute("rspace", "0px");
    return node;
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   * @return the created node.
   **************************************************************************/
  private Element appendInfixFunction(Parent parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "1px");
    node.setAttribute("rspace", "1px");
    return node;
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text, wide version.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   * @return the created node.
   **************************************************************************/
  private Element appendWideInfixFunction(Parent parent, String text)
  {
    Element node = appendWideOperator(parent, text);
    node.setAttribute("lspace", "1px");
    node.setAttribute("rspace", "1px");
    return node;
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendInfixPredicate(Parent parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "3px");
    node.setAttribute("rspace", "3px");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendKeyword(Parent parent, String text)
  {
    Element node = appendOperator(parent, text.toLowerCase());
    node.setAttribute("lspace", "0px");
    node.setAttribute("rspace", "0px");
    node.setAttribute("mathcolor", "blue");
  }
  
  /***************************************************************************
   * Append to parent postxfix operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendPostfix(Parent parent, String text)
  {
    Element node0 = appendOperator(parent, text.toLowerCase());
    node0.setAttribute("lspace", "0px");
    node0.setAttribute("rspace", "0px");
    Element node1 = appendOperator(parent, PostfixApplication.operator);
    node1.setAttribute("lspace", "0px");
    node1.setAttribute("rspace", "0px");
  }
  
  /***************************************************************************
   * Append to parent postxfix operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   * @param info the argument of the operator node
   **************************************************************************/
  private void appendPostfix(Parent parent, String text, Info info)
  {
    Element node0 = appendOperator(parent, text.toLowerCase());
    node0.setAttribute("lspace", "0px");
    node0.setAttribute("rspace", "0px");
    appendOperator(parent, "(");
    append(parent, info);
    appendOperator(parent, ")");
    Element node1 = appendOperator(parent, PostfixApplication.operator);
    node1.setAttribute("lspace", "0px");
    node1.setAttribute("rspace", "0px");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendInfixConnective(Parent parent, String text)
  {
    Element node = appendOperator(parent, text.toLowerCase());
    node.setAttribute("lspace", "4px");
    node.setAttribute("rspace", "4px");
    node.setAttribute("mathcolor", "blue");
    //node.setAttribute("mathvariant", "bold");
    //Element snode = appendNode(parent, "mspace");
    //snode.setAttribute("linebreak", "goodbreak");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendPrefixConnective(Parent parent, String text)
  {
    Element node = appendOperator(parent, text.toLowerCase());
    node.setAttribute("lspace", "0px");
    node.setAttribute("rspace", "3px");
    node.setAttribute("mathcolor", "blue");
    //node.setAttribute("mathvariant", "bold");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendQuantifier(Parent parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("rspace", "0px");
    node.setAttribute("mathcolor", "red");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text, wide version.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendWideConnector(Parent parent, String text)
  {
    Element node = appendWideOperator(parent, text);
    node.setAttribute("lspace", "1px");
    node.setAttribute("rspace", "1px");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendPunctuation(Parent parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "0px");
    node.setAttribute("rspace", "2px");
  }
  
  /***************************************************************************
   * Append expression occurring as a left subexpression of an outer expression,
   * if necessary with parentheses.
   * @param parent the parent to which the expression is to be appended.
   * @param info the info for the expression to be appended.
   * @param outer the outer expression.
   *************************************************************************/
  private void appendLeftPar(Parent parent, Info info, Expression outer)
  {
    Expression exp = (Expression)info.tree;
    boolean parentheses =
      exp.getPrecedence() < outer.getPrecedence() &&
      !(exp.getClass() == outer.getClass() && exp.isLeftAssociative());
    if (parentheses) appendOperator(parent, "(");
    append(parent, info);
    if (parentheses) appendOperator(parent, ")");
  }
  
  /***************************************************************************
   * Append expression occurring as a right subexpression of an outer expression,
   * if necessary with parentheses.
   * @param info the info for the expression to be appended.
   * @param outer the outer expression.
   *************************************************************************/
  private void appendRightPar(Parent parent, Info info, Expression outer)
  {
    Expression exp = (Expression)info.tree;
    boolean parentheses =
      exp.getPrecedence() < outer.getPrecedence() &&
      !(exp.getClass() == outer.getClass() && exp.isRightAssociative());
    if (parentheses) appendOperator(parent, "(");
    append(parent, info);
    if (parentheses) appendOperator(parent, ")");
  }
  
  /****************************************************************************
   * Append tree based on previously computed breaks information.
   * @param parent the parent to which to append the information.
   * @param info the information on which the printing is based.
   ***************************************************************************/
  private void append(Parent parent, Info info)
  {
    AST tree = info.tree;
    // int len = info.getLength();
    int free = width-pos;
    
    /*
    // enough space left, we can print the simple way
    if (len <= free)
    {
      tree.print(out);
      pos = pos+len;
      return;
    }
    */
    
    // we may switch to line mode
    boolean oldMode = lineMode;
    if (info.getLength() <= free || level > MAXLEVEL || free < MINFREE || tree instanceof Type)
      lineMode = true;

    // the core printing method
    current = info;
    cparent = parent;
    tree.accept(this);
    
    // restore original mode
    lineMode = oldMode;
  }
  
  /****************************************************************************
   * Append sequence of trees.
   * @param parent the parent to which to append.
   * @param info the sequence to be appended.
   * @param from the index of the first element to be appended.
   * @param to the index of the last element to be appended.
   * @param separator the separator appended between elements.
   * @param nl true iff a new line is started before each element
   ***************************************************************************/
  private void append(Parent parent, Info[] info, int from, int to,
    String separator, boolean nl)
  {
    int n = to-from;
    
    // check whether sequence fits in single line
    int l = Breaks.length(info);
    if (n>1) l = l+(n-1)*(separator.length()+1);
    boolean single = (pos+l <= width);
    boolean mode = lineMode;
    if (single) 
    {
      // newLine();
      lineMode = true;
    }
    
    for (int i=from; i<to; i++)
    {
      if (!single && nl) newLine(parent);
      append(parent, info[i]);
      if (i<to-1) { appendWideOperator(parent, separator); }
    }
    
    // restore mode
    lineMode = mode;
  }
  
  // --------------------------------------------------------------------------
  //
  // the visitor methods.
  //
  // --------------------------------------------------------------------------
  
  public void visit(Identifier exp)
  {
    Parent parent = cparent;
    appendIdentifier(parent, exp);
    // return true;
  }
 
  public void visit(Name exp)
  {
    Parent parent = cparent;
    appendName(parent, exp);
    // return true;
  }
  
  public void visit(ApplicationExpression exp)
  {
    // information for current node
    Info info = current;
    Parent parent = cparent;
    
    // we append applications in line mode
    boolean mode = lineMode;
    lineMode = true;
    
    // append the application
    Info[] infos = info.getChildren();
    append(parent, infos[0]);
    appendOperator(parent, "(");
    append(parent, infos, 1, infos.length, ",", false);
    appendOperator(parent, ")");
    
    // restore the line mode
    lineMode = mode;
    // return true;
  }
  
  public void visit(PlusTerm exp)
  { 
    appendBinaryTerm(exp, exp.getOperator());
    // return true;
  }
  
  public void visit(MinusTerm exp)
  { 
    appendBinaryTerm(exp, exp.getOperator());
    // return true;
  }
  
  public void visit(BitOrTerm exp)
  { 
    appendBinaryTerm(exp, exp.getOperator());
    // return true;
  }
  
  public void visit(BitAndTerm exp)
  { 
    appendBinaryTerm(exp, exp.getOperator());
    // return true;
  }
  
  public void visit(TimesTerm exp)
  { 
    appendBinaryTerm(exp, "\u00B7");
    // return true;
  }
  
  private void appendBinaryTerm(BinaryTerm exp, String op)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // terms appended in line mode
    boolean mode = lineMode;
    lineMode = true;
    
    // append expression
    Info[] children = info.getChildren();
    appendLeftPar(parent, children[0], exp);
    appendInfixFunction(parent, op);
    appendRightPar(parent, children[1], exp);
    
    // restore the line mode
    lineMode = mode;
  }
  
  public void visit(DividesTerm exp)
  { 
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // terms appended in line mode
    boolean mode = lineMode;
    lineMode = true;
    
    // append expression
    Info[] children = info.getChildren();
    Node node = appendNode(parent, "mfrac");
    Node num = appendNode(new Parent(node), "mrow");
    append(new Parent(num), children[0]);
    Node denom = appendNode(new Parent(node), "mrow");
    append(new Parent(denom), children[1]);
    
    // restore the line mode
    lineMode = mode;
    // return true;
  }
  
  public void visit(PowerTerm exp)
  { 
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // terms appended in line mode
    boolean mode = lineMode;
    lineMode = true;
    
    // append expression
    Info[] children = info.getChildren();
    Node node = appendNode(parent, "msup");
    Node base = appendNode(new Parent(node), "mrow");
    appendLeftPar(new Parent(base), children[0], exp);
    Node power = appendNode(new Parent(node), "mrow");
    appendRightPar(new Parent(power), children[1], exp);
    
    // restore the line mode
    lineMode = mode;
    // return true;
  }
  
  public void visit(EqualsFormula exp)
  { 
    appendBinaryAtomicFormula(exp, exp.getOperator());
    // return true;
  }
  
  public void visit(GreaterEqualFormula exp)
  { 
    appendBinaryAtomicFormula(exp, "\u2265");
    // return true;
  }
  
  public void visit(GreaterFormula exp)
  { 
    appendBinaryAtomicFormula(exp, exp.getOperator());
    // return true;
  }
  
  public void visit(LessEqualFormula exp)
  { 
    appendBinaryAtomicFormula(exp, "\u2264");
    // return true;
  }
  
  public void visit(LessFormula exp)
  { 
    appendBinaryAtomicFormula(exp, exp.getOperator());
    // return true;
  }
  
  public void visit(NotEqualsFormula exp)
  { 
    appendBinaryAtomicFormula(exp, "\u2260");
    // return true;
  }

  public void visit(SimilarFormula exp)
  { 
    appendBinaryAtomicFormula(exp, exp.getOperator());
    // return true;
  }
  
  private void appendBinaryAtomicFormula(BinaryAtomicFormula exp, String op)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // terms appended in line mode
    boolean mode = lineMode;
    lineMode = true;
    
    // append expression
    Info[] children = info.getChildren();
    append(parent, children[0]); //appendLeftPar(parent, children[0], exp);
    appendInfixPredicate(parent, op);
    append(parent, children[1]); //appendRightPar(parent, children[1], exp);
    
    // restore the line mode
    lineMode = mode;
  }
  
  public void visit(AndFormula exp)
  { 
    appendAndOrFormula(exp, "\u2227");
    // return true;
  }
  
  public void visit(OrFormula exp)
  { 
    appendAndOrFormula(exp, "\u2228");
    // return true;
  }
  
  private void appendAndOrFormula(BinaryFormula exp, String operator)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // derived information
    BinaryFormula outer = exp;
    Info[] infos = info.getChildren();
    int olen = operator.length();
    
    // iterate over subformulas of same type
    changeLevel(true, parent);
    while (true)
    {
      while(true)
      {
        appendLeftPar(parent, infos[0], outer);
        Formula second = outer.getSecond();
        if (second.getClass() != exp.getClass()) break;
        Info[] infos0 = infos[1].getChildren();
        if (2+olen+infos0[0].getLength()+INDENT > width-pos) break;
        appendInfixConnective(parent, operator);
        info = infos[1];
        outer = (BinaryFormula)second;
        infos = info.getChildren();
      }
      changeLevel(false, parent);
      newLine(parent);
      appendInfixConnective(parent, operator);
      Formula second = outer.getSecond();
      if (second.getClass() != exp.getClass()) break;
      if (infos[1].getLength()+INDENT <= width-pos) break;
      changeLevel(true, parent);
      newLine(parent);
      info = infos[1];
      outer = (BinaryFormula)second;
      infos = info.getChildren();
    }
    
    // append last subformula
    changeLevel(true, parent);
    newLine(parent);
    appendRightPar(parent, infos[1], outer);
    changeLevel(false, parent);
  }
  
  public void visit(EquivalentFormula exp)
  { 
    appendEquivFormula(exp, "\u21D4");
    // return true;
  }
  
  public void visit(NotEquivalentFormula exp)
  { 
    appendEquivFormula(exp, "xor");
    // return true;
  }
  
  private void appendEquivFormula(BinaryFormula exp, String operator)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append expression
    Info[] children = info.getChildren();
    changeLevel(true, parent);
    appendLeftPar(parent, children[0], exp);
    changeLevel(false, parent);
    newLine(parent);
    appendInfixConnective(parent, operator);
    changeLevel(true, parent);
    newLine(parent);
    appendRightPar(parent, children[1], exp);
    changeLevel(false, parent);
  }
  
  public void visit(ImpliesFormula exp)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // iterate as long as second formula is of same type
    Info[] infos = info.getChildren();
    changeLevel(true, parent);
    Formula outer = (Formula)info.tree;
    while (true)
    {
      appendLeftPar(parent, infos[0], outer);
      changeLevel(false, parent);
      newLine(parent);
      appendInfixConnective(parent, "\u21D2");
      if (!(infos[1].tree instanceof ImpliesFormula)) break;
      if (infos[1].getLength()+INDENT <= width-pos) break;
      changeLevel(true, parent);
      newLine(parent);
      info = infos[1];
      infos = info.getChildren();
      outer = (Formula)info.tree;
    }
    changeLevel(true, parent);
    newLine(parent);
    appendRightPar(parent, infos[1], outer);
    changeLevel(false, parent);
    
    // we are done
    // return true;
  }
  
  public void visit(UnaryTerm exp)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // terms appended in line mode
    boolean mode = lineMode;
    lineMode = true;
    
    // append expression
    Info[] children = info.getChildren();
    appendPrefixFunction(parent, exp.getOperator());
    appendRightPar(parent, children[0], exp);
    
    // restore the line mode
    lineMode = mode;
    // return true;
  }
  
  public void visit(NotFormula exp)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append expression
    Info[] children = info.getChildren();
    appendPrefixConnective(parent, "\u00AC");
    appendRightPar(parent, children[0], exp);
    
    // we are done
    // return true;
  }
  
  public void visit(ForallFormula exp)
  {
    appendQuantifiedExpression(exp, "\u2200");
    // return true;
  }
  
  public void visit(ExistsFormula exp)
  {
    appendQuantifiedExpression(exp, "\u2203");
    // return true;
  }
  
  public void visit(LambdaTerm exp)
  {
    appendQuantifiedExpression(exp, "\u03BB");
    // return true;
  }
  
  public void visit(ArrayTerm exp)
  {
    appendQuantifiedExpression(exp, "array");
    // return true;
  }
  
  private void appendQuantifiedExpression(QuantifiedExpression exp, 
    String quantifier)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append the expression
    Info[] infos = info.getChildren();
    int n = infos.length;
    appendQuantifier(parent, quantifier);
    boolean mode = lineMode;
    lineMode = true;
    append(parent, infos, 0, n-1, ",", false);
    lineMode = mode;
    appendPunctuation(parent, ":");
    changeLevel(true, parent);
    newLine(parent);
    appendRightPar(parent, infos[n-1], exp);
    changeLevel(false, parent);
  }
  
  public void visit(ArrayType exp)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append the expression
    Info[] infos = info.getChildren();
    appendPrefixConnective(parent, ArrayType.arrayKeyword);
    append(parent, infos[0]);
    appendInfixConnective(parent, ArrayType.baseKeyword);
    appendRightPar(parent, infos[1], exp);
    
    // we are done
    // return true;
  }
  
  public void visit(BooleanType exp)
  {
    // information for this node
    Parent parent = cparent;
    
    Element node = document.createElement("mo");
    node.setAttribute("mathvariant", "double-struck");
    parent.node.appendChild(node);
    if (Main.toLogical())
      appendTextNode(new Parent(node), "LOGICAL", 1);
    else
      appendTextNode(new Parent(node), "B", 1);
    // return true;
  }
  
  public void visit(BitType exp)
  {
    // information for this node
    Parent parent = cparent;
    
    Element node = document.createElement("mo");
    node.setAttribute("mathvariant", "double-struck");
    parent.node.appendChild(node);
    if (Main.toLogical())
      appendTextNode(new Parent(node), "B", 1);
    else
      appendTextNode(new Parent(node), "BIT", 1);
    // return true;
  }
  
  public void visit(NatType exp)
  {
    appendAtomicType(exp, "\u2115");
    // return true;
  }
  
  public void visit(IntType exp)
  {
    appendAtomicType(exp, "\u2124");
    // return true;
  }
  
  public void visit(RealType exp)
  {
    appendAtomicType(exp, "\u211D");
    // return true;
  }
  
  private void appendAtomicType(AtomicType exp, String op)
  {
    // information for this node
    Parent parent = cparent;
    
    appendOperator(parent, op);
  }
 
  public void visit(BooleanLiteral exp)
  {
    // information for this node
    Parent parent = cparent;
    
    if (Main.toLogical())
    {
      if (exp instanceof False)
        appendConstant(parent, "logicalfalse");
      else
        appendConstant(parent, "logicaltrue");
      return;
    }
    appendConstant(parent, exp.getName().toLowerCase());
    // return true;
  }
  
  public void visit(BitLiteral exp)
  {
    // information for this node
    Parent parent = cparent;
    
    if (Main.toLogical())
    {
      if (exp instanceof BitFalse)
        appendConstant(parent, "false");
      else
        appendConstant(parent, "true");
      return;
    }
    appendConstant(parent, exp.getName().toLowerCase());
    // return true;
  }
  
  public void visit(IntLiteral exp)
  {
    // information for this node
    Parent parent = cparent;
    
    appendNumber(parent, exp);
    // return true;
  }
  
  public void visit(StateLiteral exp)
  {
    // information for this node
    Parent parent = cparent;
    
    appendConstant(parent, exp.getName().toLowerCase());
    // return true;
  }
  
  public void visit(StringLiteral exp)
  {
    // information for this node
    Parent parent = cparent;
    
    appendOperator(parent, "\"" + exp.getName() + "\"");
    // return true;
  }
  
  public void visit(FunctionType exp)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // derived information
    Info[] infos = info.getChildren();
    int n = infos.length;
    
    // special case of a predicate
    Type range = exp.getRange();
    if (range instanceof BooleanType)
    {
      appendKeyword(parent, "predicate");
      appendOperator(parent, "(");
      append(parent, infos, 0, n-1, ",", false);
      appendOperator(parent, ")");
      // return true;
    }
    
    // general case of a function
    if (n == 2)
      appendLeftPar(parent, infos[0], exp);
    else
    {
      appendOperator(parent, "(");
      append(parent, infos, 0, n-1, ", ", false);
      appendOperator(parent, ")");
    }
    appendWideInfixFunction(parent, "\u2192");
    appendRightPar(parent, infos[n-1], exp);
    
    // we are done
    // return true;
  }
  
  public void visit(IfThenElseExpression exp)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // derived information
    Info[] infos = info.getChildren();
    boolean mode = lineMode;
    
    // append the expression
    appendInfixConnective(parent, IfThenElseExpression.ifKeyword); 
    lineMode = true;
    append(parent, infos[0]);
    lineMode = mode;
    appendInfixConnective(parent, IfThenElseExpression.thenKeyword);
    changeLevel(true, parent);
    newLine(parent); append(parent, infos[1]);
    changeLevel(false, parent);
    while (((IfThenElseExpression)info.tree).getElseBranch() 
        instanceof IfThenElseExpression)
    {
      info = infos[2];
      infos = info.getChildren();
      newLine(parent);
      appendInfixConnective(parent, IfThenElseExpression.elsifKeyword);
      lineMode = true;
      append(parent, infos[0]);
      lineMode = mode;
      appendInfixConnective(parent, IfThenElseExpression.thenKeyword);
      changeLevel(true, parent);
      newLine(parent); append(parent, infos[1]);
      changeLevel(false, parent);
    }
    newLine(parent); appendInfixConnective(parent, IfThenElseExpression.elseKeyword);
    changeLevel(true, parent);
    newLine(parent); append(parent, infos[2]);
    changeLevel(false, parent);
    newLine(parent); appendInfixConnective(parent, IfThenElseExpression.endifKeyword);
    
    // we are done
    // return true;
  }
  
  public void visit(LetExpression exp)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // derived information
    Info[] infos = info.getChildren();
    int n = infos.length;
    
    // append expression
    appendPrefixConnective(parent, LetExpression.letKeyword);
    changeLevel(true, parent);
    append(parent, infos, 0, n-1, ", ", true);
    changeLevel(false, parent);
    newLine(parent); appendInfixConnective(parent, LetExpression.inKeyword);
    changeLevel(true, parent);
    newLine(parent);
    appendRightPar(parent, infos[n-1], exp);
    changeLevel(false, parent);
    
    // we are done
    // return true;
  }
  
  public void visit(PostfixApplication exp)
  { 
    // if (exp instanceof StateThrowsException)
    //   return visit((StateThrowsException)exp);
    
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // derived information
    Info[] infos = info.getChildren();
    
    // append expression
    appendPostfix(parent, exp.getIdentifier().getString());
    appendRightPar(parent, infos[0], exp);
    
    // we are done
    // return true;
  }
  
  public void visit(StateThrowsException exp)
  { 
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // derived information
    Info[] infos = info.getChildren();
    
    // append expression
    appendPostfix(parent, exp.getIdentifier().getString(), infos[1]);
    appendRightPar(parent, infos[0], exp);

    // we are done
    // return true;
  }
  
  public void visit(ProgramVariableBase exp)
  {     
    // information for this node
    Parent parent = cparent;
    
    boolean mode = lineMode;
    lineMode = true;
    appendKeyword(parent, exp.getKeyword());
    appendSpaceSeparator(parent);
    appendName(parent, exp.getName());
    lineMode = mode;
    // return true;
  }
  
  public void visit(WritesOnlyFormula exp)
  { 
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // special case
    if (exp instanceof ReadsOnlyFormula)
    {
      appendKeyword(parent, ReadsOnlyFormula.keyword);
      return;
    }
    
    // general case
    Info[] infos = info.getChildren();
    boolean mode = lineMode;
    lineMode = true;
    appendKeyword(parent, WritesOnlyFormula.keyword);
    appendSpaceSeparator(parent);
    append(parent, infos, 0, infos.length, ",", false);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(RecordTerm exp)
  { 
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append expression
    Info[] infos = info.getChildren();
    boolean mode = lineMode;
    lineMode = true;
    appendOperator(parent, RecordTerm.beginToken);
    append(parent, infos, 0, infos.length, ",", false);
    appendOperator(parent, RecordTerm.endToken);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(RecordType exp)
  { 
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append expression
    Info[] infos = info.getChildren();
    boolean mode = lineMode;
    lineMode = true;
    appendOperator(parent, RecordType.beginToken);
    append(parent, infos, 0, infos.length, ",", false);
    appendOperator(parent, RecordType.endToken);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(Reference exp)
  { 
    // information for this node
    Parent parent = cparent;
    
    appendName(parent, exp.getName());
    // return true;
  }
  
  // append expression
  public void visit(SelectionTerm exp)
  { 
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append expression
    Info[] infos = info.getChildren();
    boolean mode = lineMode;
    lineMode = true;
    appendLeftPar(parent, infos[0], exp);
    append(parent, infos[1]);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(SelectorIdentifier exp)
  { 
    // information for this node
    Parent parent = cparent;
    
    // append expression
    appendOperator(parent, ".");
    appendIdentifier(parent, exp.getIdentifier());
    // return true;
  }
  
  public void visit(SelectorIndex exp)
  { 
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append expression
    Info[] infos = info.getChildren();
    appendOperator(parent, "[");
    append(parent, infos[0]);
    appendOperator(parent, "]");

    // we are done
    // return true;
  }
  
  public void visit(SelectorNumber exp)
  { 
    // information for this node
    Parent parent = cparent;
    
    // append expression
    appendOperator(parent, ".");
    appendNumber(parent, exp.getNumber());
    // return true;
  }
  
  public void visit(StateType exp)
  { 
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append expression
    Info[] infos = info.getChildren();
    
    // type tag
    // appendOperator(parent, "state");
    Element node = document.createElement("mo");
    node.setAttribute("mathvariant", "double-struck");
    parent.node.appendChild(node);
    appendTextNode(new Parent(node), "S", 1);
    
    if (infos != null)
    {
      boolean mode = lineMode;
      lineMode = true;
      appendOperator(parent, "(");
      append(parent, infos[0]);
      appendOperator(parent, ")");
      lineMode = mode;
    }
    
    // we are done
    // return true;
  }
  
  public void visit(SubrangeType exp)
  { 
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append expression
    Info[] infos = info.getChildren();
    boolean mode = lineMode;
    lineMode = true;
    appendOperator(parent, SubrangeType.beginToken);
    append(parent, infos[0]);
    appendOperator(parent, SubrangeType.separatorToken);
    append(parent, infos[1]);
    appendOperator(parent, SubrangeType.endToken);
    lineMode = mode;
    
    // we are done
    // return true;
  } 
  
  public void visit(SubType exp)
  { 
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append expression
    Info[] infos = info.getChildren();
    boolean mode = lineMode;
    lineMode = true;
    appendOperator(parent, SubType.keyword);
    appendOperator(parent, "(");
    append(parent, infos[0]);
    appendOperator(parent, ")");
    lineMode = mode;
    
    // we are done
    // return true;
  }  
    
  public void visit(TupleTerm exp)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append expression
    Info[] infos = info.getChildren();
    boolean mode = lineMode;
    lineMode = true;
    appendOperator(parent, TupleTerm.beginToken);
    append(parent, infos, 0, infos.length, ",", false);
    appendOperator(parent, TupleTerm.endToken);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(TupleType exp)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append expression
    Info[] infos = info.getChildren();
    boolean mode = lineMode;
    lineMode = true;
    appendOperator(parent, TupleTerm.beginToken);
    append(parent, infos, 0, infos.length, ",", false);
    appendOperator(parent, TupleTerm.endToken);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(TypedIdentifier exp)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append expression
    Info[] infos = info.getChildren();
    boolean mode = lineMode;
    lineMode = true;
    append(parent, infos[0]);
    appendWideConnector(parent, "\u2208");
    append(parent, infos[1]);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(UpdateTerm exp)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append expression
    Info[] infos = info.getChildren();
    int n = infos.length;
    boolean mode = lineMode;
    lineMode = true;
    appendLeftPar(parent,infos[0], exp);
    appendInfixConnective(parent, UpdateTerm.withKeyword);
    append(parent, infos, 1, n-1, "", false);
    appendInfixConnective(parent, UpdateTerm.assignToken);
    appendRightPar(parent, infos[n-1], exp);
    lineMode = mode;
    
    // we are done
    // return true;
  }
  
  public void visit(ValueDeclarationInitialized def)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append declaration
    Info[] infos = info.getChildren();
    append(parent, infos[0]);
    appendWideConnector(parent, "\u2208");
    append(parent, infos[1]);
    appendInfixPredicate(parent, "=");
    changeLevel(true, parent);
    newLine(parent); 
    append(parent, infos[2]);
    changeLevel(false, parent);
    
    // we are done
    // return true;
  }
  
  public void visit(ValueDefinitionClass def)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append declaration
    Info[] infos = info.getChildren();
    append(parent, infos[0]);
    appendInfixPredicate(parent, "=");
    changeLevel(true, parent);
    newLine(parent); 
    append(parent, infos[1]);
    changeLevel(false, parent);
    
    // we are done
    // return true;
  }
  
  public void visit(ValuedIdentifier exp)
  {
    // information for this node
    Info info = current;
    Parent parent = cparent;
    
    // append expression
    Info[] infos = info.getChildren();
    boolean mode = lineMode;
    lineMode = true;
    append(parent, infos[0]);
    appendOperator(parent, ":=");
    append(parent, infos[1]);
    lineMode = mode;
    
    // we are done
    // return true;
  }
}
