// ----------------------------------------------------------------------------
// RefSubstitution.java
// Reference substitution P[T1/I1,...]
// $Id: RefSubstitution.java,v 1.10 2012/12/16 18:09:40 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.ewuni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Judgements;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Semantics.*;

/**
 * 
 * Reference substitution P[T1/I1,...]
 *
 */
public final class RefSubstitution extends ASTCloner
{
  /***************************************************************************
   * Perform substitution of identifiers by terms.
   * @param map the mapping of identifier names (strings) to terms.
   * @param exp the expression where to perform the substitution.
   * @return the substituted formula (=exp, if substitution map is empty)
   **************************************************************************/
  public static AST perform(Map<String, Term> map, AST exp)
  {
    if (map.isEmpty()) return exp;
    RefSubstitution subst = new RefSubstitution(map);
    return subst.clone0(exp);
  }
  
  // the substitution map
  private final Map<String, Term> map;
  
  /**************************************************************************
   * Construct a substitution
   * @param map a mapping from identifier names (strings) to terms.
   *************************************************************************/
  private RefSubstitution(Map<String, Term> map)
  {
    this.map = map;
  }
  
  /****************************************************************************
   * Process expression
   * @param exp the expression to be substituted.
   ***************************************************************************/
  public void visit(QuantifiedExpression exp) 
  { 
    Map<String, Term> map0 = new MapClass<String, Term>();
    map0.putAll(map);
    TypedIdentifier[] vars = exp.getVariables();
    for (TypedIdentifier var : vars)
      map0.remove(var.getName().getString());
    Expression body0 = (Expression)RefSubstitution.perform(map0, exp.getBody());
    if (exp instanceof ForallFormula)
      result = ForallFormula.construct(vars, (Formula)body0);
    else if (exp instanceof ExistsFormula)
      result = ExistsFormula.construct(vars, (Formula)body0);
    else if (exp instanceof ArrayTerm)
    {
      result = ArrayTerm.construct(vars, (Term)body0);
      ((Term)result).setType(((Term)exp).getType());
    }
    else if (exp instanceof LambdaFormula)
      result = LambdaFormula.construct(vars, (Formula)body0);
    else if (exp instanceof LambdaTerm)
    {
      result = LambdaTerm.construct(vars, (Term)body0);
      ((Term)result).setType(((Term)exp).getType());
    }
    else
      Main.internalError("unknown quantified expression");
  }
  
  /****************************************************************************
   * Process expression
   * @param exp the expression to be substituted.
   ***************************************************************************/
  public void visit(ExistsFormula exp) 
  {
    visit((QuantifiedExpression)exp);
  }
  
  /****************************************************************************
   * Process expression
   * @param exp the expression to be substituted.
   ***************************************************************************/
  public void visit(ForallFormula exp) 
  {
    visit((QuantifiedExpression)exp);
  }
  
  /****************************************************************************
   * Process expression
   * @param exp the expression to be substituted.
   ***************************************************************************/
  public void visit(ArrayTerm exp) 
  {
    visit((QuantifiedExpression)exp);
  }
  
  /****************************************************************************
   * Process expression
   * @param exp the expression to be substituted.
   ***************************************************************************/
  public void visit(LambdaFormula exp) 
  {
    visit((QuantifiedExpression)exp);
  }
  
  /****************************************************************************
   * Process expression
   * @param exp the expression to be substituted.
   ***************************************************************************/
  public void visit(LambdaTerm exp) 
  {
    visit((QuantifiedExpression)exp);
  }
  
  /****************************************************************************
   * Process expression
   * @param exp the expression to be substituted.
   ***************************************************************************/
  public void visit(LetExpression exp) 
  { 
    Map<String, Term> map0 = new MapClass<String, Term>();
    map0.putAll(map);
    ValueDefinition[] defs = (ValueDefinition[])exp.getDefinitions();
    int n = defs.length;
    for (int i=0; i<n; i++)
    {
      ValueDefinition def = defs[i];
      defs[i] = (ValueDefinition)RefSubstitution.perform(map0, def);
      map0.remove(def.getName().getString());
    }
    Expression body0 = (Expression)RefSubstitution.perform(map0, exp.getBody());
    if (exp instanceof LetFormula)
      result = LetFormula.construct(defs, (Formula)body0);
    else if (exp instanceof LetTerm)
    {
      result = LetTerm.construct(defs, (Term)body0);
      ((Term)result).setType(((Term)exp).getType());
    }
    else
      Main.internalError("unknown let expression");
  }
  
  /****************************************************************************
   * Process expression
   * @param exp the expression to be substituted.
   ***************************************************************************/
  public void visit(LetFormula exp) 
  { 
    visit((LetExpression)exp);
  }
  
  /****************************************************************************
   * Process expression
   * @param exp the expression to be substituted.
   ***************************************************************************/
  public void visit(LetTerm exp) 
  { 
    visit((LetExpression)exp);
  }
  
  /****************************************************************************
   * Process expression
   * @param exp the expression to be substituted.
   ***************************************************************************/
  public void visit(Reference exp) 
  { 
    Name name = exp.getName();
    Identifier[] idents = name.idents;
    Identifier ident = idents[0];
    if (!(ident.getSymbol() instanceof ValueSymbol))
    {
      super.visit(exp);
      return;
    }
    String key = ident.getString();
    Term term = map.get(key);
    if (term == null) 
    {
      super.visit(exp);
      return;
    }
    Term exp0 = term;
    int n = idents.length;
    for (int i=1; i<n; i++)
    {
      Identifier ident0 = idents[i];
      Selector sel = SelectorIdentifier.construct(ident0);
      ValueSymbol symbol = (ValueSymbol)ident0.getSymbol();
      exp0 = SelectionTerm.construct(exp0, sel);
      exp0.setType(symbol.getType());
    }
    result = exp0;
  }
}
