// ----------------------------------------------------------------------------
// Satisfies.java
// The judgement "se, Is, VS |- C: F"
// $Id: Satisfies.java,v 1.68 2013/06/26 16:24:26 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General private License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General private License for more details.
//
// You should have received a copy of the GNU General private License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Judgements;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Judgements.Substitution.Value;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Program.IdentifierSelector;
import fmrisc.ProgramExplorer.Syntax.Program.LeftValue;
import fmrisc.ProgramExplorer.Syntax.Program.TypeExpression;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Tasks.*;

import java.util.*;

/**
 * 
 * The judgement "se, Is, VS |- C: F"
 *
 */
public final class Satisfies
{ 
  // the context of the judgement processor
  private final ParamSymbol method;        // the method of the judgements
  private final StateType stype;           // the state type
  private final TaskFolder invariants;     // the invariants folder
  private final Vector<TheorySymbol> theories;     // the referenced theories
  private final Declaration[][] decls;             // the current declarations
  private final ErrorStream out;                   // the error output stream
  
  // the auxiliary judgement processors            
  private final ExpressionLogic logic; // translate expression to term/formula

  // the string type (program and logic)
  private final TypeExpression stringProgramType;
  private final Type stringLogicType;
  
  // stack of for-loop updates
  private Vector<fmrisc.ProgramExplorer.Syntax.Program.Statement> forLoopUpdates = 
      new Vector<fmrisc.ProgramExplorer.Syntax.Program.Statement>();
  
  /***************************************************************************
   * Create a judgement processor.
   * @param method the method in which the judged commands arise.
   * @param stype the type of the state for the context of a command.
   * @param invariants the folder to which to add invariant tasks.
   * @param theories the referenced theories (first is base theory)
   * @param decls a sequence of declaration lists.
   * @param out the stream to which to write error messages.
   **************************************************************************/
  public Satisfies(ParamSymbol method,
    StateType stype, TaskFolder invariants, 
    Vector<TheorySymbol> theories, Declaration[][] decls, ErrorStream out)
  {
    this.method = method;
    this.stype = stype;
    this.invariants = invariants;
    this.theories = theories;
    this.decls = decls;
    this.out = out;
    this.logic = new ExpressionLogic(theories.firstElement());
    method.setSatisfies(this);
    this.stringProgramType = 
        theories.firstElement().getTypeTable().getStringType();
    this.stringLogicType = 
        method.getClassSymbol().getPackage().getTypeTranslator().getStringType();
  }
  
  /***************************************************************************
   * Get the state type of this judgement processor
   * @return the state type.
   **************************************************************************/
  public StateType getStateType()
  {
    return stype;
  }
  
  /***************************************************************************
   * Get the string type of this judgement processor
   * @return the string type.
   **************************************************************************/
  public Type getStringType()
  {
    return stringLogicType;
  }
  
  /***************************************************************************
   * Get expression translator of this judgement processor
   * @return the expression translator
   **************************************************************************/
  public ExpressionLogic getExpressionLogic()
  {
    return logic;
  }
  
  /***************************************************************************
   * Extend judgement to the denoted set of modifiable variables.
   * @param S the judgement.
   * @param vars the set of modifiable variables
   * @return the extended judgement
   **************************************************************************/
  public StatJudgement extendFrame(StatJudgement S, Set<VariableSymbol> vars)
  {
    Formula formula = S.formula;
    Set<VariableSymbol> variables = new SetClass<VariableSymbol>(S.variables);
    for (VariableSymbol v : vars)
    {
      if (!variables.contains(v))
      {
        Name name = getReference(v);
        Formula eq = 
          EqualsFormula.construct(
              postVariable(name),
              preVariable(name));
        formula = AndFormula.construct(formula, eq);
        variables.add(v);
      }
    }
    StatJudgement S0 = new StatJudgement(S.getPrecondition(), formula, 
        S.executes, S.continues, S.breaks, S.returns, 
        S.exceptions, variables);
    S0.setTermination(S.getTermination());
    return S0;
  }
  
  /***************************************************************************
   * Apply the judgement "se, Is, Vs |- C: F"
   * to compute the specification F of command C and 
   * to generate proving tasks as a side effect
   * (see Section 4.12 of "A Program Calculus", se, Is, Vs are not needed 
   * because identifiers have already been resolved to symbols).
   * @param C the command C
   * @return the specification F of command C.
   **************************************************************************/
  public StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.Statement C)
  {
    StatJudgement S = null;
    if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.AssertionStatement) 
      S = derive((fmrisc.ProgramExplorer.Syntax.Program.AssertionStatement)C);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.AssignmentStatement) 
      S = derive((fmrisc.ProgramExplorer.Syntax.Program.AssignmentStatement)C);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.BlockStatement) 
      S = derive((fmrisc.ProgramExplorer.Syntax.Program.BlockStatement)C);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.BreakStatement) 
      S = derive((fmrisc.ProgramExplorer.Syntax.Program.BreakStatement)C);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.CallStatement) 
      S = derive((fmrisc.ProgramExplorer.Syntax.Program.CallStatement)C);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.ConditionalStatement) 
      S = derive((fmrisc.ProgramExplorer.Syntax.Program.ConditionalStatement)C);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.ContinueStatement) 
      S = derive((fmrisc.ProgramExplorer.Syntax.Program.ContinueStatement)C);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.EmptyStatement) 
      S = derive((fmrisc.ProgramExplorer.Syntax.Program.EmptyStatement)C);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.ForLoopStatement) 
    {
      forLoopUpdates.add(((fmrisc.ProgramExplorer.Syntax.Program.ForLoopStatement)C).update);
      S = derive((fmrisc.ProgramExplorer.Syntax.Program.ForLoopStatement)C);
      forLoopUpdates.remove(forLoopUpdates.size()-1);
    }
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.ReturnStatement) 
      S = derive((fmrisc.ProgramExplorer.Syntax.Program.ReturnStatement)C);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.TryCatchStatement) 
      S = derive((fmrisc.ProgramExplorer.Syntax.Program.TryCatchStatement)C);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.VariableStatement) 
      S = derive((fmrisc.ProgramExplorer.Syntax.Program.VariableStatement)C);
    else if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.WhileLoopStatement) 
    {
      forLoopUpdates.add(null);
      S = derive((fmrisc.ProgramExplorer.Syntax.Program.WhileLoopStatement)C);
      forLoopUpdates.remove(forLoopUpdates.size()-1);
    }
    else
      Main.internalError("unknown kind of statement");
    StatementSpec annotation = (StatementSpec)C.getAnnotation();
    if (annotation != null)
    {
      Formula assertion = annotation.getAssertion();
      if (assertion != null)
      {
        assertion = Normalization.perform(assertion);
        S.setPrecondition(andFormula(assertion, S.getPrecondition()));
      }
    }
    C.setJudgement(S);
    return S;
  }
  
  /****************************************************************************
   * The assertion statement "assert E"
   ***************************************************************************/
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.AssertionStatement C)
  {
    Vector<Formula> pre = new Vector<Formula>();
    Formula cond = logic.deriveFormula(C.cond, pre);
    pre.add(cond);
    return new StatJudgement(
        andFormula(pre), cond,
        true, false, false, false, null, null);
  }
  
  /****************************************************************************
   * The assignment statement "lval = exp"
   ***************************************************************************/
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.AssignmentStatement C)
  {
    return deriveAssignment(C.lval, C.exp);
  }
  
  /****************************************************************************
   * The assignment statement "lval = exp"
   * @param lval the left value of the assignment.
   * @param exp the right value of the assignment.
   ***************************************************************************/
  private StatJudgement deriveAssignment(
    fmrisc.ProgramExplorer.Syntax.Program.LeftValue lval,
    fmrisc.ProgramExplorer.Syntax.Program.ValueExpression exp)
  {
    // construct the update value
    Vector<Formula> pre = new Vector<Formula>();
    Term value = logic.deriveTerm(exp, pre);
    
    // get name of updated variable and the selectors to be processed
    Name name = getVariableName(lval, pre);
    Selector[] sels = selectors;
    if (sels != null)
    {
      Type type = value.getType();
      value = UpdateTerm.construct(preVariable(name), sels, value);
      value.setType(type);
    }
    
    // construct the specification
    Set<VariableSymbol> vars = new SetClass<VariableSymbol>();
    vars.add((VariableSymbol)name.getSymbol());
    return new StatJudgement(
        andFormula(pre),
        // implied from the frame
        // AndFormula.construct(
            EqualsFormula.construct(
                postVariable(name), value),
        // StateExecutes.construct(stateLiteral(false))),
        true, false, false, false, null, vars);
  }
  
  // selectors that remain to be processed (null, if none)
  private Selector[] selectors;
  
  /****************************************************************************
   * Get from left value the name of the updated variable.
   * Sets selectors to those selectors that remain to be processed.
   * @param lval the left value.
   * @param pre the vector to which to add preconditions.
   * @return the name of the updated variable.
   ***************************************************************************/
  private Name getVariableName(
    fmrisc.ProgramExplorer.Syntax.Program.LeftValue lval,
    Vector<Formula> pre)
  { 
    // we construct the name of the variable, its symbol,
    // and the selectors to be processed
    Vector<Identifier> idents = new Vector<Identifier>();
    VariableSymbol symbol = null;
    Vector<Selector> selectors0 = new Vector<Selector>();
    
    // the base expression constructed so far and its type
    Term base = null;
    Type btype = null;
    
    // process identifiers of unique name
    Name name = NameUtils.uniqueName(Name.construct(lval.name));
    int n0=name.idents.length;
    for (int i=0; i<n0; i++)
    {
      Identifier ident = name.idents[i];
      idents.add(ident);
      Symbol symbol0 = ident.getSymbol(); 
      if (symbol0 instanceof VariableSymbol) 
      {
        symbol = (VariableSymbol)symbol0;
        Type type = symbol.getLogicType();
        base = preVariable(Name.construct(idents));
        btype = type;
        for (int j=i+1; j<n0; j++)
        {
          Identifier ident0 = name.idents[j];
          IdentifierSelector sel = IdentifierSelector.construct(ident0);
          sel.setLogicBaseType(type);
          for (Selector sel0 : logic.deriveSelector(base, sel, pre))
          {
            selectors0.add(sel0);
            base = SelectionTerm.construct(base, sel0);
            btype = selectedType(btype, sel0);
            base.setType(btype);
          }
          ObjectVariableSymbol symbol1 = (ObjectVariableSymbol)ident0.getSymbol();
          type = symbol1.getLogicType();
        }
        break;
      }
    }
      
    // process selectors
    int n = lval.selectors.length;
    if (symbol != null)
    {
      // all selectors remain to be processed
      for (int i=0; i<n; i++)
      {
        for (Selector sel0 : logic.deriveSelector(base, lval.selectors[i], pre))
        {
          selectors0.add(sel0);
          base = SelectionTerm.construct(base, sel0);
          btype = selectedType(btype, sel0);
          base.setType(btype);
        }
      }
    }
    else
    {
      for (int i=0; i<n; i++)
      {
        fmrisc.ProgramExplorer.Syntax.Program.Selector sel = lval.selectors[i];
        if (sel instanceof fmrisc.ProgramExplorer.Syntax.Program.IdentifierSelector)
        {
          Identifier ident = 
            ((fmrisc.ProgramExplorer.Syntax.Program.IdentifierSelector)sel).name;
          idents.add(ident);
          Symbol symbol0 = ident.getSymbol();
          if (symbol0 instanceof VariableSymbol) 
          {
            symbol = (VariableSymbol)symbol0;
            for (int j=i+1; j<n; j++)
            {
              for (Selector sel0 : logic.deriveSelector(base, lval.selectors[j], pre))
              {
                selectors0.add(sel0);
                base = SelectionTerm.construct(base, sel0);
                btype = selectedType(btype, sel0);
                base.setType(btype);
              }
            }
            break;
          }
        }
      }
    }
    
    // must have found the variable
    if (symbol == null) Main.internalError("variable selector not found");
    
    // construct result
    if (selectors0.isEmpty())
      selectors = null;
    else
    {
      selectors = new Selector[selectors0.size()];
      selectors0.toArray(selectors);
    }
    return Name.construct(idents);
  }
  
  /**************************************************************************
   * Get type selected from base type by selector.
   * @param btype the base type. 
   * @param sel the selector.
   * @return the selected type.
   *************************************************************************/
  private Type selectedType(Type btype, Selector sel)
  {
    if (sel instanceof SelectorIndex)
      return ((ArrayType)btype).getBase();
    if (sel instanceof SelectorIdentifier)
    {
      Identifier ident = ((SelectorIdentifier)sel).getIdentifier();
      ValueSymbol symbol = (ValueSymbol)ident.getSymbol();
      return symbol.getType();
    }
    if (sel instanceof SelectorNumber)
    {
      int number = Integer.parseInt(((SelectorNumber)sel).
          getNumber().getName());
      return ((TupleType)btype).getBase()[number];
    }
    Main.internalError("unknown selector");
    return null;
  }
  
  /****************************************************************************
   * The block statement "{ ... }"
   * The rule is derived by combining the rules of local variable blocks
   * with that of statement sequences. Since we refer to variables by symbols
   * rather than by identifiers, only one block rule has to be considered.
   ***************************************************************************/
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.BlockStatement C)
  {
    // determine set of locally declared variables
    Set<VariableSymbol> Vs0 = new SetClass<VariableSymbol>();
    addLocal(Vs0, C.stats);
    
    // compute specification of statement sequence
    StatJudgement spec = null;
    int n = C.stats.length;
    for (int i=n-1; i>=0; i--)
    {
      // the next statement to be processed
      fmrisc.ProgramExplorer.Syntax.Program.Statement C0 = C.stats[i];
      
      // skip statement, if it does not have any effect
      if (C0 instanceof fmrisc.ProgramExplorer.Syntax.Program.EmptyStatement)
      {
        continue;
      }
      
      // the specification to be derived
      StatJudgement spec0 = null;
        
      if (C0 instanceof fmrisc.ProgramExplorer.Syntax.Program.VariableStatement)
      {
        fmrisc.ProgramExplorer.Syntax.Program.VariableStatement C1 =
          (fmrisc.ProgramExplorer.Syntax.Program.VariableStatement)C0;
        
        if (C1.exp == null) 
        {
          // without initialization expression, declaration has no effect
          C1.setJudgement(derive(C1));
          continue;
        }
        // with initialization expression, effect is that of an assignment
        spec0 = deriveAssignment(
            LeftValue.construct(C1.name, 
                new fmrisc.ProgramExplorer.Syntax.Program.Selector[0]), 
                C1.exp);
        C1.setJudgement(spec0);
      }
      else if (C0 instanceof fmrisc.ProgramExplorer.Syntax.Program.DeclCallStatement)
      {        
        fmrisc.ProgramExplorer.Syntax.Program.DeclCallStatement C1 =
          (fmrisc.ProgramExplorer.Syntax.Program.DeclCallStatement)C0;
        
        // effect is that of a call
        spec0 = deriveCallStatement(Name.construct(C1.name), null, null, C1);
        C1.setJudgement(spec0);
      }
      else
      {
        // the standard case
        spec0 = derive(C0);
      }
      
      // combine with previously derived ones
      if (spec == null)
        spec = spec0;
      else
        spec = combineSeq(spec0, spec);
    }
    
    // the block may not contain any statement with an effect
    if (spec == null)
    {
      return new StatJudgement(
          null,
          ExpressionLogic.booleanFormula(true),
          true, false, false, false, null, null);
    }
    
    // unused names for mathematical variables substituting program variables
    Formula formula = spec.formula;
    Collection<String> used = UsedVariables.compute(formula);
    int vn = Vs0.size();
    VariableSymbol[] symbols = new VariableSymbol[vn];
    Vs0.toArray(symbols);
    String[] oldVars = NewVariables.newVariables(symbols, "o", used);
    UsedVariables.add(oldVars, used);
    String[] newVars = NewVariables.newVariables(symbols, "n", used);
    
    // construct sequence of bound variables
    ValueSymbol[] oldSymbols = new ValueSymbol[vn];
    ValueSymbol[] newSymbols = new ValueSymbol[vn];
    TypedIdentifier[] tids = new TypedIdentifier[vn*2];
    for (int i=0; i<vn; i++)
    {
      Type type = symbols[i].getLogicType();
      Identifier oldId = Identifier.construct(oldVars[i]);
      Identifier newId = Identifier.construct(newVars[i]);
      TypedIdentifier oldTid = TypedIdentifier.construct(oldId, type);
      TypedIdentifier newTid = TypedIdentifier.construct(newId, type);
      oldSymbols[i] = ValueSymbol.construct(oldId, type, oldTid);
      newSymbols[i] = ValueSymbol.construct(newId, type, newTid);
      tids[2*i] = oldTid;
      tids[2*i+1] = newTid;
    }
    
    // perform substitution for program references
    Map<VariableSymbol,Value> vars = new MapClass<VariableSymbol,Value>();
    for (int i=0; i<vn; i++)
    {
      vars.put(symbols[i], 
          new Substitution.Value(
              Substitution.Action.term, 
              reference(oldSymbols[i].getName()),
              Substitution.Action.term, 
              reference(newSymbols[i].getName())
              ));
    } 
    Formula formula0 = (Formula)Substitution.perform(formula, null, null, vars);
    
    // construct specification
    Set<VariableSymbol> Vs2 = new SetClass<VariableSymbol>(spec.variables);
    Vs2.removeAll(Vs0);
    return new StatJudgement(
        null,
        existsFormula(tids, formula0),
        spec.executes, spec.continues, spec.breaks, spec.returns,
        spec.exceptions, Vs2);
  }
  
  /****************************************************************************
   * Add to collection Vs all local variables declared by statements Cs.
   * @param Vs a collection of variable symbols.
   * @param Cs a sequence of statements.
   ***************************************************************************/
  private void addLocal(Set<VariableSymbol> Vs,
    fmrisc.ProgramExplorer.Syntax.Program.Statement[] Cs)
  {
    for (fmrisc.ProgramExplorer.Syntax.Program.Statement C : Cs)
    {
      if (!(C instanceof fmrisc.ProgramExplorer.Syntax.Program.VariableDeclaration))
        continue;
      fmrisc.ProgramExplorer.Syntax.Program.VariableDeclaration C0 =
        (fmrisc.ProgramExplorer.Syntax.Program.VariableDeclaration)C;
      VariableSymbol symbol = (VariableSymbol)C0.getName().getSymbol();
      Vs.add(symbol);
    }
  }
  
  /***************************************************************************
   * Combine specifications of two consecutively executed commands "C1;C2"
   * @param spec1 specification of first command C1
   * @param spec2 specification of second command C2
   * @return the combined specification
   **************************************************************************/
  private StatJudgement combineSeq(StatJudgement spec1, StatJudgement spec2)
  {
    // combine specification formulas to common variable frame
    SpecCombination comb = combineSpec(spec1, spec2, true);
      
    // unused names for mathematical variables substituting program variables
    Collection<String> used = UsedVariables.compute(comb.formula1);
    UsedVariables.add(comb.formula2, used);
    
    // unused names for mathematical variables substituting program variables
    // one extra variable for the intermediate state
    int vn = comb.variables.size();
    VariableSymbol[] symbols = new VariableSymbol[vn];
    comb.variables.toArray(symbols);
    String[] vars = NewVariables.newVariables(symbols, "", used);
    UsedVariables.add(vars, used);
    String state = NewVariables.newVariable("state", used);
    
    // construct sequence of bound variables
    TypedIdentifier[] tids = new TypedIdentifier[vn+1];
    ValueSymbol[] symbols0 = new ValueSymbol[vn+1];
    for (int i=0; i<vn; i++)
    {
      Type type = symbols[i].getLogicType();
      Identifier ident = Identifier.construct(vars[i]);
      tids[i] = TypedIdentifier.construct(ident, type);
      symbols0[i] = ValueSymbol.construct(ident, type, tids[i]);
    }
    Identifier sident = Identifier.construct(state);
    tids[vn] = TypedIdentifier.construct(sident, stype);
    symbols0[vn] = ValueSymbol.construct(sident, stype, tids[vn]);
    
    // perform substitutions for program references
    Map<VariableSymbol,Value> vars1 = new MapClass<VariableSymbol,Value>();
    for (int i=0; i<vn; i++)
    {
      vars1.put(symbols[i], 
          new Substitution.Value(
              Substitution.Action.none, null, 
              Substitution.Action.term, 
              reference(symbols0[i].getName())));
    } 
    Formula formula1 = (Formula)Substitution.perform(comb.formula1, 
        null, reference(sident), vars1);
    Map<VariableSymbol,Value> vars2 = new MapClass<VariableSymbol,Value>();
    for (int i=0; i<vn; i++)
    {
      vars2.put(symbols[i], 
          new Substitution.Value(
              Substitution.Action.term, 
              reference(symbols0[i].getName()),
              Substitution.Action.none, null));
    } 
    Formula formula2 = (Formula)Substitution.perform(comb.formula2, 
        reference(sident), null, vars2);
    
    // first command finishes execution normally
    if (!spec1.continues && !spec1.breaks && !spec1.returns && 
        spec1.exceptions.isEmpty())
    {
      return new StatJudgement(
          null,
          existsFormula(tids,
              AndFormula.construct(formula1, formula2)),
          spec2.executes, spec2.continues, spec2.breaks, spec2.returns, 
          spec2.exceptions, comb.variables);
    }
    
    // first command may not finish execution normally
    Set<ClassSymbol> exceptions = new SetClass<ClassSymbol>(spec1.exceptions);
    exceptions.addAll(spec2.exceptions);
    Formula elseFormula = 
      EqualsFormula.construct(stateLiteral(false), reference(sident));
    for (int i=0; i<vn; i++)
    {
      elseFormula = AndFormula.construct(elseFormula,
          EqualsFormula.construct(
              postVariable(getReference(symbols[i])),
              reference(tids[i].getName())));
    }
    Formula result =
      existsFormula(tids,
          AndFormula.construct(
              formula1, 
              IfThenElseFormula.construct(
                  StateExecutes.construct(reference(sident)),
                  formula2,
                  elseFormula)));
    return new StatJudgement(
        null, result,
        spec1.executes && spec2.executes,
        spec1.continues || spec2.continues, 
        spec1.breaks || spec2.breaks, 
        spec1.returns || spec2.returns, 
        exceptions, comb.variables);
  }
  
  /****************************************************************************
   * The statement "break"
   ***************************************************************************/
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.BreakStatement C)
  {
    return new StatJudgement(
        null,
        StateBreaks.construct(stateLiteral(false)),
        false, false, true, false, null, null);
  }
  
  /****************************************************************************
   * The conditional statement "if (E) C1 [ else C2 ]"
   ***************************************************************************/
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.ConditionalStatement C)
  {
    Vector<Formula> pre = new Vector<Formula>();
    Formula cond = logic.deriveFormula(C.cond, pre);
    StatJudgement s1 = derive(C.thenstat);
    if (C.elsestat == null) 
    {
      return new StatJudgement(
          andFormula(pre),
          IfThenElseFormula.construct(cond, s1.formula, 
              AndFormula.construct(
                  readsonly(s1.variables),
                  StateExecutes.construct(stateLiteral(false)))),
          true, s1.continues, s1.breaks, s1.returns, 
          s1.exceptions, s1.variables);
    }
    StatJudgement s2 = derive(C.elsestat);
    Set<ClassSymbol> exceptions = new SetClass<ClassSymbol>(s1.exceptions);
    exceptions.addAll(s2.exceptions);
    SpecCombination comb = combineSpec(s1, s2, false);
    return new StatJudgement(
        andFormula(pre),
        IfThenElseFormula.construct(cond, comb.formula1, comb.formula2),
        s1.executes || s2.executes,
        s1.continues || s2.continues, 
        s1.breaks || s2.breaks, 
        s1.returns || s2.returns, 
        exceptions,
        comb.variables);
  }

  /****************************************************************************
   * The statement "continue"
   ***************************************************************************/
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.ContinueStatement C)
  {
    // in for-loops, continue performs the loop update
    Formula formula = StateContinues.construct(stateLiteral(false));
    Set<VariableSymbol> vars = null;
    int n = forLoopUpdates.size();
    if (n >= 2)
    {
      fmrisc.ProgramExplorer.Syntax.Program.Statement update =
          forLoopUpdates.get(n-2);
      if (update != null)
      {
        StatJudgement spec = derive(update);
        vars = spec.variables;
        formula = AndFormula.construct(spec.formula, formula);
      }
    }   
    return new StatJudgement(
        null,
        formula,
        false, true, false, false, null, vars);
  }
  
  /****************************************************************************
   * The empty statement ";"
   ***************************************************************************/
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.EmptyStatement C)
  {
    // statement has no effect
    return new StatJudgement(
        null,
        StateExecutes.construct(stateLiteral(false)),
        true, false, false, false, null, null); 
  }
  
  /****************************************************************************
   * The statement "return [E]"
   ***************************************************************************/
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.ReturnStatement C)
  {
    Formula formula = StateReturns.construct(stateLiteral(false));
    Vector<Formula> pre = new Vector<Formula>();
    if (C.value != null)
    {
      Term term = logic.deriveTerm(C.value, pre);
      formula = AndFormula.construct(formula,
          EqualsFormula.construct(
              stateValue(stateLiteral(false)),
              term));
    }
    return new StatJudgement(
        andFormula(pre),
        formula,
        false, false, false, true, null, null);
  }
  
  /****************************************************************************
   * The statement "throw new I(E)"
   ***************************************************************************/
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.ThrowStatement C)
  {
    Formula formula = 
      StateThrowsException.construct(stateLiteral(false), C.method);
    Vector<Formula> pre = new Vector<Formula>();
    Term term = logic.deriveTerm(C.getValue(), pre);
    formula = AndFormula.construct(formula,
        EqualsFormula.construct(
            stateMessage(stateLiteral(false)),
            term));
    ClassSymbol symbol = (ClassSymbol)C.method.getBase().getSymbol();
    Set<ClassSymbol> exceptions = new SetClass<ClassSymbol>();
    exceptions.add(symbol);
    return new StatJudgement(
        andFormula(pre),
        formula,
        false, false, false, false, exceptions, null);
  }
  
  /****************************************************************************
   * The protected block statement "try { C1 } ( catch(T I) C2 )+"
   ***************************************************************************/
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.TryCatchStatement C)
  {
    // derive specifications and bring them to common variable frame
    int n = C.handlers.length;
    StatJudgement specs[] = new StatJudgement[n+1];
    for (int i=0; i<n; i++)
    {
      VariableSymbol symbol = 
        (VariableSymbol)C.handlers[i].param.name.getSymbol();
      specs[i] = derive(C.handlers[i].body);
      // we remove handler parameter such that it does not appear in common frame
      specs[i].variables.remove(symbol);
    }
    specs[n] = derive(C.body);
    SpecCombinations comb = combineSpec(specs);
    
    // unused names for mathematical variables substituting program variables
    Collection<String> used = new SetClass<String>();
    for (int i=0; i<n+1; i++)
      UsedVariables.add(comb.formulas[i], used);
    int vn = comb.variables.size();
    VariableSymbol[] symbols = new VariableSymbol[vn];
    comb.variables.toArray(symbols);
    String[] vars = NewVariables.newVariables(symbols, "", used);
    UsedVariables.add(vars, used);
    String svar = NewVariables.newVariable("state", used);
    used.add(svar);
    
    // construct sequence of bound variables
    TypedIdentifier[] tids = new TypedIdentifier[vn+1];
    ValueSymbol[] symbols0 = new ValueSymbol[vn+1];
    for (int i=0; i<vn; i++)
    {
      Type type = symbols[i].getLogicType();
      Identifier ident = Identifier.construct(vars[i]);
      tids[i] = TypedIdentifier.construct(ident, type);
      symbols0[i] = ValueSymbol.construct(ident, type, tids[i]);
    }
    Identifier sident = Identifier.construct(svar);
    tids[vn] = TypedIdentifier.construct(sident, stype);
    symbols0[vn] = ValueSymbol.construct(sident, stype, tids[vn]);
    
    // perform substitutions for program references
    Map<VariableSymbol,Value> vars1 = new MapClass<VariableSymbol,Value>();
    for (int i=0; i<vn; i++)
    {
      vars1.put(symbols[i], 
          new Substitution.Value(
              Substitution.Action.none, null, 
              Substitution.Action.term, 
              reference(symbols0[i].getName())));
    } 
    Formula formula1 = (Formula)Substitution.perform(comb.formulas[n], 
        null, reference(sident), vars1);
            
    // construct existentially quantified formulas for handlers
    Formula[] formulas = new Formula[n];
    for (int i=0; i<n; i++)
    {
      // handler parameter
      Identifier param = C.handlers[i].param.name;
      
      // determine unused identifier names
      Collection<String> used0 = new SetClass<String>(used);
      UsedVariables.add(comb.formulas[i], used0);
      String pname = param.getString();
      String opname = NewVariables.newVariable(pname + "o", used0);
      used0.add(opname);
      String npname = NewVariables.newVariable(pname + "n", used0);
      used0.add(npname);
      String tvar = NewVariables.newVariable("state", used0);
      
      // construct typed identifiers
      VariableSymbol psymbol = (VariableSymbol)param.getSymbol();
      Type ptype = psymbol.getLogicType();
      if (ptype != stringLogicType)
        Main.internalError("exception object has not string type");
      Identifier opident = Identifier.construct(opname);
      TypedIdentifier otid = TypedIdentifier.construct(opident, ptype);
      ValueSymbol opsymbol = ValueSymbol.construct(opident, ptype, otid);
      Identifier npident = Identifier.construct(npname);
      TypedIdentifier ntid = TypedIdentifier.construct(npident, ptype);
      ValueSymbol npsymbol = ValueSymbol.construct(npident, ptype, ntid);
      Identifier tident = Identifier.construct(tvar);
      TypedIdentifier ttid = TypedIdentifier.construct(tident, stype);
      ValueSymbol.construct(tident, stype, ttid);
      
      // perform substitution in handler formula
      Map<VariableSymbol,Value> vars2 = new MapClass<VariableSymbol,Value>();
      for (int j=0; j<vn; j++)
      {
        vars2.put(symbols[j], 
            new Substitution.Value(
                Substitution.Action.term, 
                reference(symbols0[i].getName()),
                Substitution.Action.none, null));
      } 
      vars2.put(psymbol,
          new Substitution.Value(
              Substitution.Action.term, 
              reference(opsymbol.getName()),
              Substitution.Action.term, 
              reference(npsymbol.getName())));
      Formula formula0 = 
        (Formula)Substitution.perform(comb.formulas[i],
            reference(tident), null, vars2);
      
      // construct existential formula
      TypedIdentifier[] tids0 = new TypedIdentifier[]{ otid, ntid, ttid };
      formulas[i] = 
        existsFormula(tids0,
            AndFormula.construct(
                EqualsFormula.construct(
                    reference(opident),
                    stateValue(reference(sident))),
                AndFormula.construct(
                    StateExecutes.construct(reference(tident)),
                    formula0)));
    }
    
    // construct equality formula as base of conditional formula
    Formula eqFormula = 
      EqualsFormula.construct(
          stateLiteral(false), reference(sident));
    for (int i=vn-1; i>=0; i--)
    {
      Formula eq =
        EqualsFormula.construct(
            postVariable(getReference(symbols[i])),
            reference(tids[i].getName()));
      eqFormula = AndFormula.construct(eq, eqFormula);
    }
    
    // construct conditional formula
    Formula ifFormula = eqFormula;
    for (int i=n-1; i>=0; i--)
    {
      Name cname = 
        ((fmrisc.ProgramExplorer.Syntax.Program.NamedType)
            C.handlers[i].param.type).name;
      ifFormula =
        IfThenElseFormula.construct(
            StateThrowsException.construct(
                reference(sident), cname),
                formulas[i],
                ifFormula);
    }   
    
    // construct the specification
    boolean executes = false;
    for (StatJudgement s : specs)
      executes = executes || s.executes;
    boolean continues = false;
    for (StatJudgement s : specs)
      continues = continues || s.continues;
    boolean breaks = false;
    for (StatJudgement s : specs)
      breaks = breaks || s.breaks;
    boolean returns = false;
    for (StatJudgement s : specs)
      returns = returns || s.returns;
    Set<ClassSymbol> exceptions = 
      new SetClass<ClassSymbol>(specs[n].exceptions);
    for (int i=0; i<n; i++)
    {
      Name cname = 
        ((fmrisc.ProgramExplorer.Syntax.Program.NamedType)
            C.handlers[i].param.type).name;
      ClassSymbol csymbol = (ClassSymbol)cname.getSymbol();
      exceptions.remove(csymbol);
    }
    for (int i=0; i<n; i++)
      exceptions.addAll(specs[i].exceptions);
    return new StatJudgement(
        null,
       existsFormula(tids, 
           AndFormula.construct(formula1, ifFormula)),
       executes, continues, breaks, returns, exceptions,
       comb.variables);
  }
  
  /****************************************************************************
   * The variable declaration statement "Type var;"
   ***************************************************************************/
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.VariableStatement C)
  {
    // method is only called if declaration appears not as a block element
    // has no effect then
    return new StatJudgement(
        null,
        StateExecutes.construct(stateLiteral(false)),
        true, false, false, false, null, null); 
  }
  
  /****************************************************************************
   * A while loop "while (E) C"
   ***************************************************************************/
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.WhileLoopStatement C)
  {
    // determine body specification
    StatJudgement cSpec = derive(C.body);
    
    // determine invariant, if any
    Formula invariant = null;
    if (C.getLoopAnnotation() != null)
    {
      LoopSpec loopSpec = (LoopSpec)C.getLoopAnnotation();
      invariant = loopSpec.getLoopInvariant();
      if (invariant != null)
        invariant = Normalization.perform(invariant, cSpec.variables);
    }
    
    // no invariant may be given and loop body may break
    if (invariant == null && cSpec.breaks)
    {
      return new StatJudgement(
          null,
          /* implied from frame
          AndFormula.construct(
              NotFormula.construct(
                  StateContinues.construct(stateLiteral(false))),
              NotFormula.construct(
                  StateBreaks.construct(stateLiteral(false)))), */
          ExpressionLogic.booleanFormula(true),
          cSpec.executes, false, false, cSpec.returns, 
          cSpec.exceptions, cSpec.variables);
    }
    
    // loop condition and its preconditions
    Vector<Formula> pre = new Vector<Formula>();
    Formula eFormula = logic.deriveFormula(C.cond, pre);

    // if body does not break we have additional information on post-state
    Formula optFormula = null;
    if (!cSpec.breaks)
    {
      Substitution.Value substitution = 
        new Substitution.Value(
            Substitution.Action.flip, null,
            Substitution.Action.none, null);
      Map<VariableSymbol, Substitution.Value> vsubst = 
        new MapClass<VariableSymbol, Substitution.Value>();
      for (VariableSymbol v : cSpec.variables)
        vsubst.put(v, substitution);
      Formula eFormula0 = 
        (Formula)Substitution.perform(
            eFormula,
            stateLiteral(false), null, vsubst);
      if (!cSpec.returns && cSpec.exceptions.isEmpty())
        optFormula = NotFormula.construct(eFormula0);
      else
        optFormula = 
          ImpliesFormula.construct(
              StateExecutes.construct(stateLiteral(false)),
              NotFormula.construct(eFormula0));
    }
    
    // no invariant may be given, we then know "optFormula != null"
    if (invariant == null)
    {
      return new StatJudgement(
          andFormula(pre),
          /* implied by frame
          AndFormula.construct(
              NotFormula.construct(
                  StateContinues.construct(stateLiteral(false))), */
                  AndFormula.construct(
                      NotFormula.construct(
                          StateBreaks.construct(stateLiteral(false))),
                      optFormula),
          cSpec.executes, false, false, cSpec.returns, 
          cSpec.exceptions, cSpec.variables); 
    }
    
    // -----------------------------------------------------------------------
    // start invariant handling
    // -----------------------------------------------------------------------
    
    // invariant is given, normalize it with respect to modified variables
    invariant = Normalization.perform(invariant, cSpec.variables);
    
    // unused names for logical variables in invariance goal
    Collection<String> used = UsedVariables.compute(invariant);
    UsedVariables.add(eFormula, used);
    UsedVariables.add(cSpec.formula, used);
    int n = cSpec.variables.size();
    VariableSymbol[] vsymbols = new VariableSymbol[n];
    cSpec.variables.toArray(vsymbols);
    String[] names = NewVariables.newVariables(vsymbols, "", used);
    UsedVariables.add(names, used);
    
    // determine additional state variables needed 
    boolean stateOld = cSpec.breaks || cSpec.returns || !cSpec.exceptions.isEmpty();
    boolean stateNew = cSpec.continues || stateOld;
    int n0=n;
    if (stateOld) n0 = n0+1;
    if (stateNew) n0 = n0+1;
    
    // logical variables for quantification in invariance goal
    TypedIdentifier[] vars = new TypedIdentifier[n0];
    ValueSymbol[] symbols = new ValueSymbol[n0];
    for (int i=0; i<n; i++)
    {
      Type type = vsymbols[i].getLogicType();
      Identifier ident = Identifier.construct(names[i]);
      TypedIdentifier var = TypedIdentifier.construct(ident, type);
      symbols[i] = ValueSymbol.construct(ident, type, var);
      vars[i] = var;
    }
    Identifier oldState = null;
    Identifier newState = null;
    int n1=n;
    if (stateOld)
    {
      String name = NewVariables.newVariable("so", used);
      used.add(name);
      oldState = Identifier.construct(name);
      TypedIdentifier var = TypedIdentifier.construct(oldState, stype);
      symbols[n1] = ValueSymbol.construct(oldState, stype, var);
      vars[n1] = var;
      n1 = n1+1;
    }
    if (stateNew)
    {
      String name = NewVariables.newVariable("sn", used);
      used.add(name);
      newState = Identifier.construct(name);
      TypedIdentifier var = TypedIdentifier.construct(newState, stype);
      symbols[n1] = ValueSymbol.construct(newState, stype, var);
      vars[n1] = var;
      n1 = n1+1;
    }
    
    // construct antecedent of implication
    Map<VariableSymbol,Value> map1 = new MapClass<VariableSymbol,Value>();
    for (int i=0; i<n; i++)
    {
      map1.put(vsymbols[i], 
          new Substitution.Value(
              Substitution.Action.none, null, 
              Substitution.Action.term, reference(vars[i].getName())
          ));
    } 
    Term termOld = null;
    if (stateOld) termOld = reference(oldState);
    Formula invariantGoal = 
      (Formula)Substitution.perform(invariant, null, termOld, map1);
    if (stateOld)
    {
      if (cSpec.continues)
      {
        invariantGoal = 
          AndFormula.construct(invariantGoal,
              OrFormula.construct(
                  StateExecutes.construct(reference(oldState)),
                  StateContinues.construct(reference(oldState))));
      }
      else
      {
        // special case where loop body does not continue
        invariantGoal = 
          AndFormula.construct(invariantGoal,
              StateExecutes.construct(reference(oldState)));
      }
    }
    if (stateNew)
    {
      invariantGoal =
        AndFormula.construct(invariantGoal,
            StateExecutes.construct(reference(newState)));
    }
    Map<VariableSymbol,Value> map2 = new MapClass<VariableSymbol,Value>();
    for (int i=0; i<n; i++)
    {
      map2.put(vsymbols[i], 
          new Substitution.Value(
              Substitution.Action.term, reference(vars[i].getName()), 
              Substitution.Action.none, null
          ));
    } 
    Term termNew = null;
    if (stateNew) termNew = reference(newState);
    Formula cond = 
      (Formula)Substitution.perform(eFormula, termNew, null, map2);
    invariantGoal = AndFormula.construct(invariantGoal, cond);
    Formula body =
      (Formula)Substitution.perform(cSpec.formula, termNew, null, map2);
    invariantGoal = AndFormula.construct(invariantGoal, body);
    invariantGoal = ImpliesFormula.construct(invariantGoal, invariant);
    invariantGoal = forallFormula(vars, invariantGoal);
    
    // create verification task
    invariants.add(new InvariantTask(
        this.method, C, invariantGoal,
        invariant, C.getPosition(),
        eFormula, cSpec.formula, 
        invariants.getDirectory(), stype,
        cSpec.variables, cSpec.exceptions,
        theories, decls, out));
    
    // -----------------------------------------------------------------------
    // end invariant handling
    // -----------------------------------------------------------------------
    
    // first invariant substitution
    Substitution.Value substitution = 
      new Substitution.Value(
          Substitution.Action.none, null,
          Substitution.Action.flip, null);
    Map<VariableSymbol, Substitution.Value> vsubst = 
      new MapClass<VariableSymbol, Substitution.Value>();
    for (VariableSymbol v : cSpec.variables)
      vsubst.put(v, substitution);
    Formula invariant0 = (Formula)Substitution.perform(invariant,
        null, stateLiteral(true), vsubst);
    
    // second invariant substitution
    Collection<String> ivars = UsedVariables.compute(invariant);
    String stateVar = NewVariables.newVariable("state", ivars);
    Identifier stateId = Identifier.construct(stateVar);
    TypedIdentifier tid = TypedIdentifier.construct(stateId, stype);
    ValueSymbol.construct(stateId, stype, tid);
    Formula invariant1 = (Formula)Substitution.perform(invariant, 
        null, reference(stateId), 
        new MapClass<VariableSymbol, Value>());
    
    // the "invariant step" 
    Formula invariantStep = null;
    Formula invariantCond = null;
    if (cSpec.continues) 
      invariantCond = StateContinues.construct(reference(stateId));
    if (cSpec.breaks)
      invariantCond = orFormula(invariantCond, 
          StateBreaks.construct(reference(stateId)));
    if (invariantCond == null)
    {
      invariantStep =
        AndFormula.construct(
            invariant1,
            EqualsFormula.construct(
                stateLiteral(false), reference(stateId)));
    }
    else
    {
      invariantStep =
        AndFormula.construct(
            invariant1,
            IfThenElseFormula.construct(
                invariantCond,
                StateExecutes.construct(stateLiteral(false)),
                EqualsFormula.construct(
                    stateLiteral(false), reference(stateId))));
    }
    if (cSpec.breaks)
    {
      Formula terminatesCond = null;
      if (cSpec.executes)
        terminatesCond = StateExecutes.construct(reference(stateId));
      if (cSpec.continues)
        terminatesCond = orFormula(terminatesCond,
            StateContinues.construct(reference(stateId)));
      if (terminatesCond != null)
      {
        Substitution.Value substitution0 = 
          new Substitution.Value(
              Substitution.Action.flip, null,
              Substitution.Action.none, null);
        Map<VariableSymbol, Substitution.Value> vsubst0 = 
          new MapClass<VariableSymbol, Substitution.Value>();
        for (VariableSymbol v : cSpec.variables)
          vsubst0.put(v, substitution0);
        Formula eFormula0 = 
          (Formula)Substitution.perform(
              eFormula,
              reference(tid.getName()), null, vsubst0);
        Formula terminatesFormula = 
          ImpliesFormula.construct(terminatesCond, NotFormula.construct(eFormula0));
        invariantStep = AndFormula.construct(invariantStep, terminatesFormula);
      }
    }
    invariantStep = 
      /* we make invariant0 a statement precondition
      ImpliesFormula.construct(
          invariant0,
          */
      existsFormula(new TypedIdentifier[] { tid }, invariantStep);
    
    // the specification formula
    Formula specFormula = invariantStep;
    if (optFormula != null) 
      specFormula = AndFormula.construct(optFormula, specFormula);
    /* implied by frame
    specFormula = AndFormula.construct(
        NotFormula.construct(StateContinues.construct(stateLiteral(false))),
        AndFormula.construct(
          NotFormula.construct(StateBreaks.construct(stateLiteral(false))),
          specFormula)); */
    
    // the specification
    return new StatJudgement(
        andFormula(andFormula(pre), invariant0),
        specFormula, cSpec.executes, false, false, 
        cSpec.returns, cSpec.exceptions, cSpec.variables);
  }
  
  /****************************************************************************
   * A for loop "for(init; cond; update) C"
   ***************************************************************************/
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.ForLoopStatement C)
  {
    // make sure that also while replacement receives annotation
    StatJudgement result = derive((fmrisc.ProgramExplorer.Syntax.Program.Statement)C.whileReplacement);
    
    // for presentation, propagate while-loop precondition to for-loop
    fmrisc.ProgramExplorer.Syntax.Program.Statement init = 
      C.whileReplacement.stats[0];
    fmrisc.ProgramExplorer.Syntax.Program.Statement loop = 
      C.whileReplacement.stats[1];
    
    // determine precondition of initialization statement and while loop
    Formula precondition0 = init.getJudgement().getPrecondition();
    Formula precondition1 = loop.getJudgement().getPrecondition();
    if (precondition1 != null)
      precondition1 = new Pre(stype).derive(init, precondition1);
    Formula precondition = andFormula(precondition0, precondition1);
    
    // if there is no precondition, we are done
    if (precondition == null) return result;
    
    // if init introduces new variable, universal quantification is needed 
    if (init instanceof Declaration)
    {
      Declaration init0 = (Declaration)init;
      Identifier ident = init0.getName();
      VariableSymbol symbol = (VariableSymbol)ident.getSymbol();
      Collection<String> used = UsedVariables.compute(precondition);
      String var = NewVariables.newVariable(ident.getString(), used);
      Identifier ident0 = Identifier.construct(var);
      Type type = symbol.getLogicType();
      TypedIdentifier tid = TypedIdentifier.construct(ident0, type);
      ValueSymbol.construct(ident0, type, tid);
      Map<VariableSymbol, Substitution.Value> subst = 
        new MapClass<VariableSymbol, Substitution.Value>();
      Term term = Reference.construct(ident0);
      Substitution.Value value = new Substitution.Value(
          Substitution.Action.term, term, 
          Substitution.Action.term, term);
      subst.put(symbol, value);
      precondition = (Formula)Substitution.perform(precondition, 
          null, null, subst);
      precondition = ForallFormula.construct(new TypedIdentifier[] { tid },
          precondition);
    }
    
    // set precondition and return judgement
    C.setPrecondition(precondition);
    return result;
  }
  
  /****************************************************************************
   * A call statement "[ ... = ] p(...) "
   ***************************************************************************/
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.CallStatement C)
  {
    if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.AssignCallStatement) 
      return derive((fmrisc.ProgramExplorer.Syntax.Program.AssignCallStatement)C);
    if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.DeclCallStatement) 
      return derive((fmrisc.ProgramExplorer.Syntax.Program.DeclCallStatement)C);
    if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.ThrowStatement) 
      return derive((fmrisc.ProgramExplorer.Syntax.Program.ThrowStatement)C);
    if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.VoidCallStatement) 
      return derive((fmrisc.ProgramExplorer.Syntax.Program.VoidCallStatement)C);
    Main.internalError("unknown kind of method call");
    return null;
  }
  
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.AssignCallStatement C)
  {
    if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.AssignNewStatement) 
      return derive((fmrisc.ProgramExplorer.Syntax.Program.AssignNewStatement)C);
    Vector<Formula> pre = new Vector<Formula>();
    Name name = getVariableName(C.lval, pre);
    return deriveCallStatement(name, selectors, pre, C);
  }
    
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.AssignNewStatement C)
  {
    Vector<Formula> pre = new Vector<Formula>();
    Name name = getVariableName(C.lval, pre);
    return deriveCallStatement(name, selectors, pre, C);
  }
  
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.DeclCallStatement C)
  {
    if (C instanceof fmrisc.ProgramExplorer.Syntax.Program.DeclNewStatement) 
      return derive((fmrisc.ProgramExplorer.Syntax.Program.DeclNewStatement)C);
    
    // called only if statement appears outside block, wrap it into block
    return derive(
      fmrisc.ProgramExplorer.Syntax.Program.BlockStatement.construct(
          new fmrisc.ProgramExplorer.Syntax.Program.Statement[] { C }));
  }
    
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.DeclNewStatement C)
  {
    // called only if statement appears outside block, wrap it into block
    return derive(
      fmrisc.ProgramExplorer.Syntax.Program.BlockStatement.construct(
          new fmrisc.ProgramExplorer.Syntax.Program.Statement[] { C }));
  }
  
  private StatJudgement derive(
    fmrisc.ProgramExplorer.Syntax.Program.VoidCallStatement C)
  {
    return deriveCallStatement(null, null, null, C);
  }
  
  /***************************************************************************
   * Derive specification from call statement
   * @param var variable to which result of the call is assigned (may be null)
   * @param selectors the selectors of the variable (may be null)   
   * @param preconditions preconditions of the statement (may be null)
   * @param C the statement
   * @return the specification
   **************************************************************************/
  private StatJudgement deriveCallStatement(Name var, Selector[] sels,
    Vector<Formula> preconditions,
    fmrisc.ProgramExplorer.Syntax.Program.CallStatement C)
  {    
    // we may allocate our own preconditions
    if (preconditions == null) preconditions = new Vector<Formula>();
    
    // determine method and its specification
    ParamSymbol method = C.getMethodSymbol();
    if (method == null) Main.internalError("no method symbol found");
    MethodSpec spec = method.getSpecification();
    
    // resolve variable references in method name to generate non-null pre-conditions
    Term base = null;
    Vector<Identifier> idents = new Vector<Identifier>();
    for (Identifier ident : C.method.idents)
    {
      Symbol symbol = ident.getSymbol();
      if (symbol == method) break;
      idents.add(ident);
      if (symbol instanceof VariableSymbol)
      {
        VariableSymbol symbol0 = (VariableSymbol)symbol;
        if (base == null)
        {
          Name name = NameUtils.uniqueName(Name.construct(idents));
          base = PreVariable.construct(name);
          base.setType(symbol0.getLogicType());
        }
        else
        {
          fmrisc.ProgramExplorer.Syntax.Program.Selector sel =
            fmrisc.ProgramExplorer.Syntax.Program.IdentifierSelector.construct(ident);
          Selector[] sel0 = logic.deriveSelector(base, sel, preconditions);
          for (Selector s : sel0)
          {
            Type btype = base.getType();
            base = SelectionTerm.construct(base, s);
            if (s instanceof SelectorIdentifier)
            {
              ValueSymbol symbol1 = 
                (ValueSymbol)((SelectorIdentifier)s).getIdentifier().getSymbol();
              base.setType(symbol1.getType());
            }
            else if (s instanceof SelectorIndex)
            {
              ArrayType type = (ArrayType)btype;
              base.setType(type.getBase());
            }
            else if (s instanceof SelectorNumber)
            {
              IntLiteral number = ((SelectorNumber)s).getNumber();
              int i = Integer.parseInt(number.getName());
              TupleType type = (TupleType)btype;
              base.setType(type.getBase()[i]);
            }
            else
              Main.internalError("unknown selector: " + sel);
          }
        }
      }
    }
    if (base != null) 
      preconditions.add(logic.notNullFormula(base));
    
    // determine exceptions/variables frame
    Set<ClassSymbol> exceptions = new SetClass<ClassSymbol>();
    Set<VariableSymbol> variables = new SetClass<VariableSymbol>();
    if (spec != null)
    {
      Name[] exceptions0 = spec.getExceptions();
      Name[] variables0 = spec.getVariables();
      if (exceptions0 != null)
      {
        for (Name ex : exceptions0)
          exceptions.add((ClassSymbol)ex.getSymbol());
      }
      if (variables0 != null)
      {
        Normalization.modifiableVariables(variables0, variables);
      }
    }

    // if method has access to "this" symbol, 
    // determine the symbol and whether method can modify it
    ThisSymbol thisSymbol = null;
    boolean modifiesThis = false;
    if (method instanceof ObjectMethodSymbol || method instanceof ConstructorSymbol)
    {
      thisSymbol = method.getClassSymbol().getThisSymbol();
      modifiesThis = variables.contains(thisSymbol);
    }
    
    // determine variable symbol and whether it can be modified by method
    // also construct symbol for result value of method/constructor
    VariableSymbol varSymbol = null;
    VariableSymbol resultSymbol = null;
    boolean modifiesVariable = false;
    if (var != null) 
    {
      varSymbol = (VariableSymbol)var.getSymbol();
      modifiesVariable = variables.contains(varSymbol);
      TypeExpression resultType = null;
      if (method instanceof ConstructorSymbol)
        resultType = thisSymbol.getType();
      else
        resultType = ((MethodSymbol)method).getReturnType();
      resultSymbol = BodySymbol.construct(Identifier.construct("result"),
          resultType, null);
      Type returnType = null;
      if (method instanceof ConstructorSymbol)
        returnType = thisSymbol.getLogicType();
      else
        returnType = ((MethodSymbol)method).getLogicReturnType(); 
      resultSymbol.setLogicType(returnType);
    }
    
    // determine formula and normalize it such that all parameters and
    // non-assignable global variables occur in pre-state form
    Formula pre = null;
    Formula formula = null;
    if (spec != null)
    {
      pre = Normalization.preCondition(method, false); // spec.getPreCondition();
      formula = Normalization.postCondition(method);   // spec.getPostCondition();
      if (method instanceof ConstructorSymbol)
      {
        Term this0 = PostVariable.construct(Name.construct(thisSymbol.getName()));
        this0.setType(thisSymbol.getLogicType());
        formula = 
          AndFormula.construct(
              logic.notNullFormula(this0), 
              formula);
      }
      /*
      Name[] variables0 = spec.getVariables();
      if (variables0 != null)
      {
        Vector<Formula> equalities = new Vector<Formula>();;
        Normalization.objectEqualities(variables0, variables, equalities);
        for (Formula eq : equalities)
          formula = andFormula(formula, eq);
      }
      */
      // make precondition of call precondition of statement
      /* 
      if (formula != null)
      {
        if (pre != null)
          formula = ImpliesFormula.construct(pre, formula);
      }
      */
      if (formula != null)
        formula = Normalization.perform(formula, variables);
    }
    
    // determine method parameters
    ParameterSymbol[] params = method.getParameterArray();
    int paramN = params.length;
    
    // determine modifiable reference parameters of method 
    Vector<Integer> refs = new Vector<Integer>();
    for (int i=0; i<paramN; i++)
    {
      ParameterSymbol param = params[i];
      Identifier ident = param.getName();
      ParameterSymbol symbol = (ParameterSymbol)ident.getSymbol();
      if (!variables.contains(symbol)) continue;
      fmrisc.ProgramExplorer.Syntax.Program.TypeExpression type = param.getType();
      if (type == stringProgramType) continue;
      if ((type instanceof fmrisc.ProgramExplorer.Syntax.Program.NamedType) ||
          (type instanceof fmrisc.ProgramExplorer.Syntax.Program.ArrayType))
      {
        refs.add(i);
      }
    }
    int refsN = refs.size();
    
    // determine the number of existentially quantified variables:
    // * one (pre-state) for each parameter
    // * one (pre-state) for "this" (object methods only)
    // * one (post-state) for each reference parameter
    // * one (post-state) for "this" (object methods, this must be in frame)
    // * one (post-state) for the result variable (if any, and 
    //     method is constructor or 
    //     variable is in frame and method throws exception)
    int preN = paramN;
    int postN = refs.size(); 
    if (thisSymbol != null) preN = preN+1;
    if (modifiesThis) postN = postN+1;
    boolean needResultVar =
      (method instanceof ConstructorSymbol || 
        (modifiesVariable && !exceptions.isEmpty()));
    if (needResultVar) postN = postN+1;
    int varN = preN+postN;
    TypedIdentifier[] tids = new TypedIdentifier[varN+2];
    
    // determine the symbols from which to derive the parameter names
    int paramPos = -1;
    int thisPosPre = -1;
    int refPos = -1;
    int thisPosPost = -1;
    int varPos = -1;
    VariableSymbol[] symbols = new VariableSymbol[varN];
    int pos = 0;
    paramPos = pos;
    for (int i=0; i<paramN; i++) // parameters (pre-state)
    {
      symbols[pos] = (VariableSymbol)params[i].getName().getSymbol();
      pos = pos+1;
    }
    if (thisSymbol != null)     // "this" (pre-state)
    {
      thisPosPre = pos;
      symbols[pos] = thisSymbol;
      pos = pos+1;
    }
    refPos = pos;
    for (int i=0; i<refsN; i++) // reference parameters (post-state)
    {
      symbols[pos] = (VariableSymbol)params[refs.get(i)].getName().getSymbol();
      pos = pos+1;
    }
    if (modifiesThis)      // "this" (post-state)
    {
      thisPosPost = pos;
      symbols[pos] = thisSymbol;
      pos = pos+1;
    }
    if (needResultVar)  // variable (post-state)
    {
      varPos = pos;
      symbols[pos] = resultSymbol;
      pos = pos+1;
    }
    if (pos != varN) Main.internalError("error in typed variable symbols");
    
    // unused names for logical variables substituting parameters
    Collection<String> used = new SetClass<String>();
    if (pre != null) UsedVariables.add(pre, used);
    if (formula != null) UsedVariables.add(formula, used);
    String[] newVars = NewVariables.newVariables(symbols, "", used);
    UsedVariables.add(newVars, used);
    String s1 = NewVariables.newVariable("s1", used);
    used.add(s1);
    String s2 = NewVariables.newVariable("s2", used);
    used.add(s2);
    
    // logical variables for quantification
    ValueSymbol[] varSymbols = new ValueSymbol[varN];
    for (int i=0; i<varN; i++)
    {
      Type type = symbols[i].getLogicType();
      Identifier varId = Identifier.construct(newVars[i]);
      TypedIdentifier tid = TypedIdentifier.construct(varId, type);
      varSymbols[i] = ValueSymbol.construct(varId, type, tid);
      tids[i] = tid;
    }
    Type stype = null;
    if (method instanceof MethodSymbol)
      stype = ((MethodSymbol)method).getLogicReturnType();
    stype = StateType.construct(stype);
    Identifier s1Id = Identifier.construct(s1);
    TypedIdentifier s1Tid = TypedIdentifier.construct(s1Id, stype);
    ValueSymbol.construct(s1Id, stype, s1Tid);
    tids[varN] = s1Tid;
    Identifier s2Id = Identifier.construct(s2);
    TypedIdentifier s2Tid = TypedIdentifier.construct(s2Id, stype);
    ValueSymbol.construct(s2Id, stype, s2Tid);
    tids[varN+1] = s2Tid;
    
    // perform substitution in formula
    if (pre != null || formula != null)
    {
      // construct substitution
      Map<VariableSymbol,Value> vars = new MapClass<VariableSymbol,Value>();
      
      // pre-state substitution for parameters
      for (int i=0; i<paramN; i++)
      {
        vars.put(params[i], 
            new Substitution.Value(
                Substitution.Action.term, 
                reference(varSymbols[paramPos+i].getName()),
                Substitution.Action.none, 
                null
            ));
      } 
      
      // pre/post-state substitution for reference parameters
      // (may overrride above substitution)
      for (int i=0; i<refsN; i++)
      {
        vars.put(params[refs.get(i)], 
            new Substitution.Value(
                Substitution.Action.term, 
                reference(varSymbols[refs.get(i)].getName()),
                Substitution.Action.term,
                reference(varSymbols[refPos+i].getName())
            ));
      } 
      
      // pre/post-state substitution for this
      if (thisSymbol != null)
      {
        if (modifiesThis)
        {
          if (method instanceof ObjectMethodSymbol)
          {
            vars.put(thisSymbol, 
                new Substitution.Value(
                    Substitution.Action.term, 
                    reference(varSymbols[thisPosPre].getName()),
                    Substitution.Action.term, 
                    reference(varSymbols[thisPosPost].getName())
                ));
          }
          else // constructor
          {
            vars.put(thisSymbol, 
                new Substitution.Value(
                    Substitution.Action.term, 
                    reference(varSymbols[thisPosPre].getName()),
                    Substitution.Action.term, 
                    reference(varSymbols[varPos].getName())
                ));
          }
        }
        else
          vars.put(thisSymbol, 
              new Substitution.Value(
                  Substitution.Action.term, 
                  reference(varSymbols[thisPosPre].getName()),
                  Substitution.Action.none, 
                  null
              ));
      }
      
      // post-state substitution for modifiable result variable
      if (modifiesVariable)
      {
        vars.put(varSymbol, 
            new Substitution.Value(
                Substitution.Action.none, 
                null,
                Substitution.Action.term, 
                reference(varSymbols[varPos].getName())
            ));
      }
      
      // perform substitution
      if (pre != null)
        pre = (Formula)Substitution.perform(pre, 
            reference(s1Id), reference(s1Id), vars);
      if (formula != null)
        formula = (Formula)Substitution.perform(formula, 
            reference(s1Id), reference(s2Id), vars);
    }
    
    // determine caller object
    VariableSymbol object = null;
    Name oname = null;
    if (method instanceof ObjectMethodSymbol)
    {
      oname = NameUtils.getVariable(C.method);
      if (oname == null)
      {
        // object is "this"
        ObjectMethodSymbol method0 = (ObjectMethodSymbol)method;
        object = method0.getClassSymbol().getThisSymbol();
        oname = Name.construct(object.getName());
      }
      else
      {
        oname = NameUtils.uniqueName(oname);
        object = (VariableSymbol)oname.getSymbol();
      }
    }
    
    // set initial values for pre-state variables 
    // 1. parameters
    Formula equalities = null;
    Term[] arguments = new Term[paramN];
    for (int i=0; i<paramN; i++)
    {
      arguments[i] = logic.deriveTerm(C.args[i], preconditions);
      EqualsFormula eq = EqualsFormula.construct(
          reference(varSymbols[paramPos+i].getName()),
          arguments[i]);
      equalities = andFormula(equalities, eq);
    }
    // 2. "this" symbol (only for object methods, not for constructors)
    if (thisSymbol != null && method instanceof ObjectMethodSymbol)
    {
      EqualsFormula eq = EqualsFormula.construct(
          reference(varSymbols[thisPosPre].getName()),
          preVariable(oname));
      equalities = andFormula(equalities, eq);
    }
    pre = andFormula(equalities, pre);
    formula = andFormula(equalities, formula);
    // 3. "old" and "new" state
    SimilarFormula sim1 = SimilarFormula.construct(
        stateLiteral(true), reference(s1Id));
    SimilarFormula sim2 = SimilarFormula.construct(
        stateLiteral(false), reference(s2Id));
    pre = andFormula(pre, sim1);
    formula = andFormula(andFormula(formula, sim1), sim2);
    
    // set final value(s) for result variable(s)
    if (varSymbol != null)
    {
      Term resultValue;
      if (method instanceof ConstructorSymbol)
        resultValue = reference(varSymbols[varPos].getName());
      else
      {
        resultValue = stateValue(reference(s2Id));
        resultValue.setType(resultSymbol.getLogicType());
      }
      if (sels != null)
      {
        resultValue = 
          UpdateTerm.construct(preVariable(var), sels, resultValue);
        resultValue.setType(varSymbol.getLogicType());
      }
      Formula executesFormula =
        EqualsFormula.construct(postVariable(var), resultValue);
      if (exceptions.isEmpty())
      {
        // no exceptions may be thrown
        // therefore value is method result
        formula = andFormula(formula, executesFormula);
      }
      else if (!modifiesVariable)
      {
        // exception may be thrown but 
        // result variable cannot be modified as global variable
        // therefore value is method result or unchanged
        formula = 
          andFormula(
              formula,
              IfThenElseFormula.construct(
                  StateExecutes.construct(stateLiteral(false)),
                  executesFormula,
                  EqualsFormula.construct(
                      postVariable(var),
                      preVariable(var))));
      }
      else
      {
        // exception may be thrown and
        // result variable can be modified as global variable 
        // therefore value is method result or determined by method postcondition
        formula = 
          andFormula(
              formula,
              IfThenElseFormula.construct(
                  StateExecutes.construct(stateLiteral(false)),
                  executesFormula,
                  EqualsFormula.construct(
                      postVariable(var),
                      reference(varSymbols[varPos].getName()))));
      }
    }
    if (modifiesThis)
    {
      // method may modify the state of this object
      formula = andFormula(
          formula,
          EqualsFormula.construct(
              postVariable(Name.construct(thisSymbol.getName())),
              reference(varSymbols[thisPosPost].getName())));
    }
    for (int i=0; i<refsN; i++)
    {
      // new value of reference may be determined from postcondition
      formula = andFormula(
          formula,
          EqualsFormula.construct(
              (Term)PostStateCloner.clone(arguments[refs.get(i)]),
              reference(varSymbols[refPos+i].getName())));
    }
    
    // remove "this" from frame but add result variables
    if (modifiesThis) 
    { 
      variables.remove(thisSymbol);
      if (method instanceof ObjectMethodSymbol)
        variables.add(object);
    }
    
    // remover modifiable parameters from frame but add argument variables
    for (int i=0; i<refsN; i++)
    {
      int p = refs.get(i);
      variables.remove(params[p]);
      Name var0 = getBaseVariable(C.args[p]);
      if (var0 == null) continue;
      variables.add((VariableSymbol)var0.getSymbol());
    }
    
    // add result symbol to frame
    if (varSymbol != null) variables.add(varSymbol);
    
    // result specification
    if (pre != null)
      pre = existsFormula(tids, pre);
    if (formula == null)
      formula = ExpressionLogic.booleanFormula(true);
    else
      formula = existsFormula(tids, formula);
    return new StatJudgement(
        andFormula(andFormula(preconditions), pre),
        formula, true, false, false, false, exceptions, variables);
  }
  
  /****************************************************************************
   * Get base variable of value expression.
   * @param exp the value expression.
   * @return its base variable (null, if there is none).
   ***************************************************************************/
  private Name getBaseVariable(fmrisc.ProgramExplorer.Syntax.Program.ValueExpression exp)
  {
    Name name = getBaseName(exp);
    if (name == null) return null;
    Symbol symbol = name.getSymbol();
    if (!(symbol instanceof VariableSymbol)) return null;
    return name;
  }
  
  /****************************************************************************
   * Get base name of value expression.
   * @param exp the value expression.
   * @return its base name (null, if there is none).
   ***************************************************************************/
  private Name getBaseName(fmrisc.ProgramExplorer.Syntax.Program.ValueExpression exp)
  {
    if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.VariableValue)
    {
      fmrisc.ProgramExplorer.Syntax.Program.VariableValue exp0 =
        (fmrisc.ProgramExplorer.Syntax.Program.VariableValue)exp;
      return Name.construct(exp0.ident);
    }
    
    if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.SelectorExpression)
    {
      fmrisc.ProgramExplorer.Syntax.Program.SelectorExpression exp0 =
        (fmrisc.ProgramExplorer.Syntax.Program.SelectorExpression)exp;
      Name base = getBaseName(exp0.base);
      if (base == null) return null;
      if (base.getSymbol() instanceof VariableSymbol) return base;
      Vector<Identifier> idents0 = new Vector<Identifier>();
      int n = exp0.sels.length;
      int pos = 0;
      while (pos < n)
      { 
        if (!(exp0.sels[pos] instanceof fmrisc.ProgramExplorer.Syntax.Program.IdentifierSelector))
          Main.internalError("unexpected selector");
        fmrisc.ProgramExplorer.Syntax.Program.IdentifierSelector sel =
          (fmrisc.ProgramExplorer.Syntax.Program.IdentifierSelector)exp0.sels[pos];
        Symbol symbol = sel.name.getSymbol();
        idents0.add(sel.name);
        pos = pos+1;
        if (symbol instanceof VariableSymbol) break;
      }
      int n0 = base.idents.length;
      int n1 = n0+pos;
      Identifier[] idents = new Identifier[n1];
      for (int i=0; i<n0; i++)
        idents[i] = base.idents[i];
      for (int i=n0; i<n1; i++)
        idents[i] = idents0.elementAt(i-n0);
      return Name.construct(idents);
    }
    
    // no base variable encountered
    return null;
  }
  
  // --------------------------------------------------------------------------
  //
  // auxiliaries
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Construct a reference from an identifier setting also its type.
   * @param ident an identifier denoting a value.
   * @return a reference to the value with the type information attached.
   ***************************************************************************/
  public static Reference reference(Identifier ident)
  {
    Symbol symbol = ident.getSymbol();
    if (symbol == null) 
      Main.internalError("no symbol for identifier");
    if (!(symbol instanceof ValueSymbol))
      Main.internalError("identifier does not denote a value");
    ValueSymbol symbol0 = (ValueSymbol)symbol;
    Reference ref =  Reference.construct(Name.construct(ident));
    Type type = symbol0.getType();
    if (type == null) Main.internalError("no type: " + ident);
    ref.setType(type);
    return ref;
  }
  
  /****************************************************************************
   * Construct a state literal giving it the current state type.
   * @param old true if it is the old state (the new state, otherwise)
   * @return the literal with the state type attached.
   ***************************************************************************/
  public StateLiteral stateLiteral(boolean old)
  {
    StateLiteral literal = old ? OldState.construct() : NewState.construct();
    if (stype == null) Main.internalError("no state type");
    literal.setType(stype);
    return literal;
  }
  
  /***************************************************************************
   * Construct a state value term.
   * @param state the state.
   * @return the corresponding state value term.
   **************************************************************************/
  public StateValue stateValue(Term state)
  {
    StateValue result = StateValue.construct(state);
    result.setType(stype.getReturn());
    return result;
  }
  
  /***************************************************************************
   * Construct a state message term.
   * @param state the state.
   * @return the corresponding state message term.
   **************************************************************************/
  public StateMessage stateMessage(Term state)
  {
    StateMessage result = StateMessage.construct(state);
    result.setType(stringLogicType);
    return result;
  }
  
  /***************************************************************************
   * Construct a pre-variable.
   * @param name the name of the variable.
   * @return the variable.
   **************************************************************************/
  public static PreVariable preVariable(Name name)
  {
    PreVariable result = PreVariable.construct(name);
    VariableSymbol symbol = (VariableSymbol)name.getSymbol();
    Type type = symbol.getLogicType();
    // symbol.setLogicType(type);
    result.setType(type);
    return result;
  }
  
  /***************************************************************************
   * Construct a post-variable.
   * @param name the name of the variable.
   * @return the variable.
   **************************************************************************/
  public static PostVariable postVariable(Name name)
  {
    PostVariable result = PostVariable.construct(name);
    VariableSymbol symbol = (VariableSymbol)name.getSymbol();
    Type type = symbol.getLogicType();
    // symbol.setLogicType(type);
    result.setType(type);
    return result;
  }
  
  /***************************************************************************
   * Construct formula that describes that program variables are unchanged.
   * @param vars the program variables.
   * @return the formula.
   **************************************************************************/
  private Formula readsonly(Set<VariableSymbol> vars)
  {
    Formula formula = null;
    for (VariableSymbol var : vars)
    {
      Name name = (var instanceof GlobalSymbol) ?
          ((GlobalSymbol)var).getLongName() : Name.construct(var.getName());
      Formula eq =
        EqualsFormula.construct(postVariable(name), preVariable(name));
      if (formula == null)
        formula = eq;
      else
        formula = AndFormula.construct(eq, formula);
    }
    if (formula == null) return ExpressionLogic.booleanFormula(true);
    return formula;
  }
  
  /***************************************************************************
   * Bind variables in formula by universal quantification.
   * @param vars the variables.
   * @param formula the formula.
   * @return the formula with all variables universally quantified.
   **************************************************************************/
  public static Formula forallFormula(TypedIdentifier[] vars, Formula formula)
  {
    if (vars.length == 0) return formula;
    return ForallFormula.construct(vars, formula);
  }
  
  /***************************************************************************
   * Bind variables in formula by existential quantification.
   * @param vars the variables.
   * @param formula the formula.
   * @return the formula with all variables existentially quantified.
   **************************************************************************/
  public static Formula existsFormula(TypedIdentifier[] vars, Formula formula)
  {
    if (vars.length == 0) return formula;
    return ExistsFormula.construct(vars, formula);
  }
  
  /****************************************************************************
   * Get name by which variable can be referenced.
   * @param symbol a symbol denoting a variable.
   * @return the name for referencing the variable.
   ***************************************************************************/
  private static Name getReference(VariableSymbol symbol)
  {
    if (symbol instanceof GlobalSymbol)
    {
      GlobalSymbol symbol0 = (GlobalSymbol)symbol;
      return symbol0.getLongName();
    }
    return Name.construct(symbol.getName());
  }
  
  /***************************************************************************
   * The result of a combination of two specifications.
   **************************************************************************/
  private final static class SpecCombination
  {
    public final Formula formula1;
    public final Formula formula2;
    public final Set<VariableSymbol> variables;
    
    /*************************************************************************
     * Construct result of a combination of two specifications.
     * @param formula1 the new first specification formula.
     * @param formula2 the new second specification formula.
     * @param variables the common variable frame of these formulas.
     ************************************************************************/
    public SpecCombination(Formula formula1, Formula formula2,
      Set<VariableSymbol> variables)
    {
      this.formula1 = formula1;
      this.formula2 = formula2;
      this.variables = variables;
    }
  }
  
  /****************************************************************************
   * Combine the formulas of two specifications.
   * @param spec1 the fist specification.
   * @param spec2 the second specification.
   * @param seq true if the combination is sequential (parallel otherwise).
   * @return the result of the combination
   ***************************************************************************/
  private SpecCombination 
  combineSpec(StatJudgement spec1, StatJudgement spec2, boolean seq)
  {
    Formula formula1 = spec1.formula;
    Formula formula2 = spec2.formula;
    
    // create common variable frame
    Set<VariableSymbol> variables = 
      new SetClass<VariableSymbol>(spec1.variables);
    variables.addAll(spec2.variables);
    for (VariableSymbol v : variables)
    {
      if (!spec1.variables.contains(v))
      {
        Name name = getReference(v);
        Formula eq = 
          EqualsFormula.construct(
              postVariable(name),
              preVariable(name));
        formula1 = AndFormula.construct(formula1, eq);
      }
      else if (!spec2.variables.contains(v))
      {
        Name name = getReference(v);
        Formula eq = 
          EqualsFormula.construct(
              postVariable(name),
              preVariable(name));
        formula2 = AndFormula.construct(formula2, eq);
      }
    }
    
    // give mode of poststate of first specification
    // (existentially quantified in sequential combination)
    if (seq)
    {
      Formula f = null; 
      if (spec1.executes) 
        f = orFormula(f, StateExecutes.construct(stateLiteral(false)));
      if (spec1.continues)
        f = orFormula(f, StateContinues.construct(stateLiteral(false)));
      if (spec1.breaks)
        f = orFormula(f, StateBreaks.construct(stateLiteral(false)));
      if (spec1.returns)
        f = orFormula(f, StateReturns.construct(stateLiteral(false)));
      for (ClassSymbol ex : spec1.exceptions)
        f = orFormula(f, StateThrowsException.construct(stateLiteral(false), 
            ex.getLongName()));
      formula1 = andFormula(formula1, f);
    }
    
    // create common execution frame
    if (!seq && spec1.executes && !spec2.executes)
      formula2 = AndFormula.construct(formula2,
          NotFormula.construct(StateExecutes.construct(stateLiteral(false))));
    if (!seq && spec2.executes && !spec1.executes)
      formula1 = AndFormula.construct(formula1,
          NotFormula.construct(StateExecutes.construct(stateLiteral(false))));
    if (spec1.continues && !spec2.continues)
      formula2 = AndFormula.construct(formula2,
          NotFormula.construct(StateContinues.construct(stateLiteral(false))));
    if (spec2.continues && !spec1.continues)
      formula1 = AndFormula.construct(formula1,
          NotFormula.construct(StateContinues.construct(stateLiteral(false))));
    if (spec1.breaks && !spec2.breaks)
      formula2 = AndFormula.construct(formula2,
          NotFormula.construct(StateBreaks.construct(stateLiteral(false))));
    if (spec2.breaks && !spec1.breaks)
      formula1 = AndFormula.construct(formula1,
          NotFormula.construct(StateBreaks.construct(stateLiteral(false))));
    if (spec1.returns && !spec2.returns)
      formula2 = AndFormula.construct(formula2,
          NotFormula.construct(StateReturns.construct(stateLiteral(false))));
    if (spec2.returns && !spec1.returns)
      formula1 = AndFormula.construct(formula1,
          NotFormula.construct(StateReturns.construct(stateLiteral(false))));
    for (ClassSymbol ex : spec1.exceptions)
    {
      if (!spec2.exceptions.contains(ex))
        formula2 = AndFormula.construct(formula2,
            NotFormula.construct(StateThrowsException.construct(
                stateLiteral(false), ex.getLongName())));
    }
    for (ClassSymbol ex : spec2.exceptions)
    {
      if (!spec1.exceptions.contains(ex)) 
        formula1 = AndFormula.construct(formula1,
            NotFormula.construct(StateThrowsException.construct(
                stateLiteral(false), ex.getLongName())));
    }
    
    // the combination result
    return new SpecCombination(formula1, formula2, variables);
  }
  
  /***************************************************************************
   * The result of a combination of a sequence of specifications specifications.
   **************************************************************************/
  private final static class SpecCombinations
  {
    public final Formula[] formulas;
    public final Set<VariableSymbol> variables;
    
    /*************************************************************************
     * Construct result of a combination of specifications.
     * @param formulas the new specification formulas.
     * @param variables the common variable frame of these formulas.
     ************************************************************************/
    public SpecCombinations(Formula[] formulas,
      Set<VariableSymbol> variables)
    {
      this.formulas = formulas;
      this.variables = variables;
    }
  }
  
  /****************************************************************************
   * Combine the formulas of a sequence of specifications.
   * @param spec the specification sequence
   * @return the result of the combination
   ***************************************************************************/
  private SpecCombinations combineSpec(StatJudgement[] spec)
  {
    Set<VariableSymbol> variables = new SetClass<VariableSymbol>();
    for (StatJudgement s : spec)
      variables.addAll(s.variables);
    int n = spec.length;
    Formula[] formulas = new Formula[n];
    for (int i=0; i<n; i++)
      formulas[i] = spec[i].formula;
    for (VariableSymbol v : variables)
    {
      for (int i=0; i<n; i++)
      {      
        if (!spec[i].variables.contains(v))
        {
          Name name = getReference(v);
          Formula eq = 
            EqualsFormula.construct(
                postVariable(name),
                preVariable(name));
          formulas[i] = AndFormula.construct(formulas[i], eq);
        }
      }
    }
    return new SpecCombinations(formulas, variables);
  }
  
  /***************************************************************************
   * Combine formulas by conjunction.
   * @param formula1 the first formula (may be null, representing TRUE).
   * @param formula2 the second formula (may be null, representing TRUE).
   * @return the conjunction of both formulas (may be null, representing TRUE).
   **************************************************************************/
  public static Formula andFormula(Formula formula1, Formula formula2)
  {
    if (formula1 == null) return formula2;
    if (formula2 == null) return formula1;
    return AndFormula.construct(formula1, formula2);
  }
  
  /***************************************************************************
   * Combine formulas by conjunction.
   * @param formulas the formulas (each may be null, representing TRUE).
   * @return the conjunction of the formulas (may be null, representing TRUE).
   **************************************************************************/
  public static Formula andFormula(Vector<Formula> formulas)
  {
    Formula result = null;
    for (Formula formula : formulas)
      result = andFormula(result, formula);
    return result;
  }
  
  /***************************************************************************
   * Combine formulas by implication.
   * @param formula1 the first formula (may be null, representing TRUE).
   * @param formula2 the second formula (may be null, representing TRUE).
   * @return the conjunction of both formulas (may be null, representing TRUE).
   **************************************************************************/
  public static Formula impliesFormula(Formula formula1, Formula formula2)
  {
    if (formula1 == null) return formula2;
    if (formula2 == null) return null;
    return ImpliesFormula.construct(formula1, formula2);
  }
  
  /***************************************************************************
   * Combine formulas by disjunction.
   * @param formula1 the first formula (may be null, representing FALSE).
   * @param formula2 the second formula (may be null, representing FALSE).
   * @return the conjunction of both formulas (may be null, representing FALSE).
   **************************************************************************/
  public static Formula orFormula(Formula formula1, Formula formula2)
  {
    if (formula1 == null) return formula2;
    if (formula2 == null) return formula1;
    return OrFormula.construct(formula1, formula2);
  }
  
  /***************************************************************************
   * Negate formula.
   * @param formula the first formula (may be null, representing TRUE).
   * @return the negation of the formula (may be null, representing TRUE).
   **************************************************************************/
  public static Formula notFormula(Formula formula)
  {
    if (formula == null) return ExpressionLogic.booleanFormula(false);
    return NotFormula.construct(formula);
  }
  
  /***************************************************************************
   * Combine formulas by conditional selection.
   * @param formula0 the selection formula (may be null, representing TRUE).
   * @param formula1 the first formula (may be null, representing TRUE).
   * @param formula2 the second formula (may be null, representing TRUE).
   * @return the conjunction of both formulas (may be null, representing TRUE).
   **************************************************************************/
  public static Formula ifThenElseFormula(Formula formula0,
    Formula formula1, Formula formula2)
  {
    if (formula0 == null) return formula1;
    if (formula1 == null) return impliesFormula(notFormula(formula0), formula2);
    if (formula2 == null) return impliesFormula(formula0, formula1);
    return IfThenElseFormula.construct(formula0, formula1, formula2);
  }
}
