// ----------------------------------------------------------------------------
// StateProblem.java
// A proving task in state logic.
// $Id: StateProblem.java,v 1.35 2012/10/18 16:12:58 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.ewuni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Proving;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Judgements.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;

import java.util.*;
import java.io.*;

/**
 * 
 * A proving problem in state logic.
 *
 */
public final class StateProblem
{
  // names of types to be visited
  private final String stateTypeName = StateType.stateKeyword;;
  
  // the base name of program variables and exceptions
  private final static String oldVarBaseName = "_old";
  private final static String newVarBaseName = "_new";
  private final static String exceptionBaseName = "exception_";
  
  // the name of state fields
  private final static String modeFieldName = "mode";
  private final static String valueFieldName = "val";
  private final static String exceptionFieldName = "exception";
  private final static String messageFieldName = "message";
  
  // the names of the state constants, functions, and predicates
  // (must create names not used elsewhere)
  private final static String newStateName = "now_";
  private final static String oldStateName = "next_";
  private final static String stateValueName = "value_";
  private final static String stateMessageName = "message_";
  private final static String stateBreaksName = "breaks_";
  private final static String stateContinuesName = "continues_";
  private final static String stateExecutesName = "executes_";
  private final static String stateReturnsName = "returns_";
  private final static String stateThrowsName = "throws_";
  private final static String stateThrowsExceptionName = "throwsException_";
  private final static String stateBreaksFlagName = "BREAKS_";
  private final static String stateContinuesFlagName = "CONTINUES_";
  private final static String stateExecutesFlagName = "EXECUTES_";
  private final static String stateReturnsFlagName = "RETURNS_";
  private final static String stateThrowsFlagName = "THROWS_";
  private final static String stateSimilarName = "similar_";
  
  // the core information
  private Declaration[] decls; // the declarations
  private Formula goal;        // the goal to be proved
  
  // the current state information 
  private Map<Symbol, Symbol> map;               // a symbol map
  private StateType stateType;                   // result type (may be null)
  private Type stringType;                       // the string type
  private Collection<VariableSymbol> variables;  // modifiable variables
  private Collection<ClassSymbol> exceptions;    // throwable exceptions
  
  // auxiliary information
  private ErrorStream out;      // a stream for printing messages
  
  /****************************************************************************
   * Construct a proving problem.
   * @param decls the declarations of entities referenced in the goal.
   * @param goal the goal to be proved.
   * @param map a map of type symbols to new type symbols
   * @param stateType the current state type (null, if no state)
   * @param stringType the string type
   * @param variables the modifiable variables 
   * @param exceptions the throwable exceptions 
   * @param out a stream for printing messages
   ***************************************************************************/
  public StateProblem(Declaration[] decls, Formula goal,
    Map<Symbol, Symbol> map, StateType stateType, Type stringType,
    Collection<VariableSymbol> variables, Collection<ClassSymbol> exceptions, 
    ErrorStream out)
  {
    this.decls = decls;
    this.goal = goal;
    this.map = map;
    this.stateType = stateType;
    this.stringType = stringType;
    this.variables = variables;
    this.exceptions = exceptions;
    this.out = out;
  }
  
  /***************************************************************************
   * Print task to output stream.
   * @param out the stream to print to
   **************************************************************************/
  public void print(PrintWriter out)
  {
    out.println("Declarations:");
    for (Declaration d : decls)
    {
      d.print(out);
      out.println(";");
    }
    out.print("Goal: ");
    goal.print(out);
    out.println("");
    if (stateType != null) out.println("State type: " + stateType);
    if (!variables.isEmpty())
    {
      out.println("Assignable variables:");
      int n = variables.size();
      int i=0;
      for (VariableSymbol var : variables)
      {
        if (var instanceof GlobalSymbol)
          ((GlobalSymbol)var).getName().print(out);
        else
          var.getName().print(out);
        if (i < n-1) out.print(", ");
        i++;
      }
      out.println("");
    }
    if (!exceptions.isEmpty())
    {
      out.println("Throwable exceptions");
      int n = exceptions.size();
      int i=0;
      for (ClassSymbol exception : exceptions)
      {
        exception.getLongName().print(out);
        if (i < n-1) out.print(", ");
        i++;
      }
      out.println("");
    }
  }
  
  /***************************************************************************
   * Translate the task into a predicate logic proving task.
   * @return the corresponding predicate logic task.
   **************************************************************************/
  public ClassicalProblem classicalProblem()
  {
    Translator translator = new Translator();
    Declaration[] decls = translator.getDeclarations();
    Formula goal = translator.getGoal();
    LogicChecking checker = LogicChecking.createChecker(out);

    /*
    System.out.println("---");
    for (Declaration d : decls)
      System.out.println(d);
    System.out.println(goal);
    System.out.println("/--");
    System.out.println(this.goal);
    */
    
    boolean okay = checker.process(decls);
    if (!okay) Main.internalError("error in declarations");
    okay = checker.process(goal);
    if (!okay) Main.internalError("error in goal: " + goal);
    return new ClassicalProblem(decls, goal, out);
  }
  
  /****************************************************************************
   * visit name of program variable to name of constant.
   * @param symbol the variable symbol
   * @param old true iff constant shall refer to old state of the variable.
   * @return the translation.
   ***************************************************************************/
  private static String constantName(VariableSymbol symbol, boolean old)
  {
    Name name = null;
    if (symbol instanceof ClassVariableSymbol)
      name = ((ClassVariableSymbol)symbol).getLongName();
    else if (symbol instanceof ObjectVariableSymbol)
      Main.internalError("unexpected object variable " + symbol.getName());
    else
      name = Name.construct(symbol.getName());
    StringBuffer b = new StringBuffer();
    append(b, name);
    if (old)
      b.append(oldVarBaseName);
    else
      b.append(newVarBaseName);
    return b.toString();
  }
  
  /***************************************************************************
   * Append identifier representation of name to buffer.
   * @param b the buffer to append to.
   * @param name the name whose representation is appended.
   **************************************************************************/
  private static void append(StringBuffer b, Name name)
  {
    Identifier[] idents = name.idents;
    int n = idents.length;
    for (int i=0; i<n; i++)
    {
      b.append(idents[i].getString());
      if (i+1<n) b.append('_');
    }
  }
  
  /*************************************************************************
   * A translator for this extended proving task.
   ************************************************************************/
  private final class Translator extends ASTCloner
  {
    // the current state information (may be all null)
    private StateType stateType; // result type
    
    // the generated declarations 
    private Vector<Declaration> decls;
    
    // the generated goal
    private Formula goal;
    
    // the mapping of textual state type representations to unique state types
    private final Map<String, StateType> stateNames;
    
    // the mapping of *unique* state types to numbers, and 
    // the number of state types generated so far
    private final Map<StateType, Integer> stateTypes;
    private int stateTypesCounter;
    
    // a mapping of pairs of state types s1,s2 to index i of s1 ~i s2
    private final Vector<StateType> simType1;
    private final Vector<StateType> simType2;
    
    // the symbols of the modifiable variables
    private final Set<VariableSymbol> stateVarSymbols;

    /*************************************************************************
     * Construct a translator for this extended proving task.
     ************************************************************************/
    public Translator()
    {
      // initialize table
      this.decls = new Vector<Declaration>();
      this.stateNames = new MapClass<String, StateType>();
      this.stateTypes = new MapClass<StateType, Integer>();
      this.stateTypesCounter = 0;
      this.simType1 = new Vector<StateType>();
      this.simType2 = new Vector<StateType>();
      this.stateVarSymbols = new SetClass<VariableSymbol>();
      
      // set state info for translation
      this.stateType = StateProblem.this.stateType;
      
      // perform translation of declarations
      for (Declaration d : StateProblem.this.decls)
        translate(d);
      
      // generate declarations for state type
      if (stateType != null) 
      {
        stateType = uniqueType(stateType);
        register(stateType);
      }
      
      // constants for referenced program variables
      ReferencedVars refvars = new ReferencedVars();
      for (Declaration d : StateProblem.this.decls)
        refvars.add(d);
      refvars.add(StateProblem.this.goal);
      Collection<VariableSymbol> vars = refvars.getResult();
      if (variables != null) 
      { 
        stateVarSymbols.addAll(variables);
        vars.addAll(variables);
      }
      for (VariableSymbol var : vars)
      {
        // object variables are translated to references of base objects
        if (var instanceof ObjectVariableSymbol) continue;
        
        // the type of the variable
        Type type = var.getLogicType();
        if (type == null) 
          Main.internalError("no logic type: " + var.getDeclaration());
        type = (Type)ReferenceCloner.clone(type, map);
        
        // constant for old value of program variable
        add(ValueDeclarationClass.construct(
            Identifier.construct(constantName(var, true)), 
            translate(type)));

        // modifiable variables also have a "new" value
        if (stateVarSymbols.contains(var))
        {
          // constant for new value of program variable
          add(ValueDeclarationClass.construct(
              Identifier.construct(constantName(var, false)), 
              translate(type)));
        }
      }

      // constants for referenced exceptions
      ReferencedExcs refexs = new ReferencedExcs();
      for (Declaration d : StateProblem.this.decls)
        refexs.add(d);
      refexs.add(StateProblem.this.goal);
      Collection<ClassSymbol> excs = refexs.getResult();
      if (exceptions != null) excs.addAll(exceptions);
      int i = 0;
      for (ClassSymbol exception : excs)
      {
        Name name = exception.getLongName();
        add(ValueDeclarationInitialized.construct(
            Identifier.construct(exceptionName(name)), 
            IntType.construct(),
            IntLiteral.construct(Integer.toString(i))));
        i = i+1;
      }
      
      // translate goal
      goal = translate(StateProblem.this.goal);
    }
    
    /***************************************************************************
     * Add declaration to sequence of generated declarations.
     * @param decl the declaration to be added.
     **************************************************************************/
    private void add(Declaration decl)
    {
      decls.add(decl);
    }
    
    /****************************************************************************
     * visit name of exception to name of constant.
     * @param name the name of the exception.
     * @return the translation.
     ***************************************************************************/
    private String exceptionName(Name name)
    {
      StringBuffer b = new StringBuffer();
      b.append(exceptionBaseName);
      append(b, NameUtils.uniqueName(name));
      return b.toString();
    }
    
    /**************************************************************************
     * visit a declaration and store the result.
     * @param d the declaration to be visited.
     *************************************************************************/
    public void translate(Declaration d)
    {
      decls.add((Declaration)clone0(d));
    }
    
    /**************************************************************************
     * Translate a type.
     * @param type the type to be translated.
     * @return the translation result.
     *************************************************************************/
    public Type translate(Type t)
    {
      return (Type)clone0(t);
    }
    
    /**************************************************************************
     * Translate a formula.
     * @param formula the formula to be translated.
     * @return the result of the translation.
     *************************************************************************/
    public Formula translate(Formula formula)
    {
      return (Formula)clone0(formula);
    }
    
    /**************************************************************************
     * Translate a term.
     * @param term the term to be translated.
     * @return the result of the translation.
     *************************************************************************/
    public Term translate(Term term)
    {
      return (Term)clone0(term);
    }
    
    /**************************************************************************
     * Get the declarations accumulated in the translation process.
     * @return all the declarations generated so far.
     *************************************************************************/
    public Declaration[] getDeclarations()
    {
      int n = decls.size();
      Declaration[] result = new Declaration[n];
      decls.toArray(result);
      return result;
    }
    
    /**************************************************************************
     * Get the visitd goal.
     * @return the visitd goal.
     *************************************************************************/
    public Formula getGoal()
    {
      return goal;
    }
    
    /****************************************************************************
     * Get unique string extension for state type number i
     * @param i the number of the state type
     * @return the extension.
     ***************************************************************************/
    private String getStateTypeExtension(int i)
    {
      if (i == 0) return "";
      return "_" + i;
    }
    
    /***************************************************************************
     * Register new state type
     * @param stateType the state type to be registered.
     * @return the name extension for the state type.
     **************************************************************************/
    private String register(StateType stateType)
    {
      String ext = getStateTypeExtension(stateTypesCounter);
      stateTypes.put(stateType, stateTypesCounter);
      stateTypesCounter++;
      generateStateTypeDecls(stateType);
      return ext;
    }
    
    /**************************************************************************
     * Construct unique version of state type.
     * @param type a state type.
     * @return its unique version.
     ************************************************************************/
    private StateType uniqueType(StateType type)
    {
      type = (StateType)TypeCloner.clone(type, map);
      String text = type.toString();
      StateType result = stateNames.get(text);
      if (result != null) return result;
      stateNames.put(text, type);
      return type;
    }
    
    /***************************************************************************
     * Get extension for denoted state type.
     * @param stateType a state type.
     * @return an extension unique for the type
     * (if the type has not yet been registered, it is now).
     **************************************************************************/
    private String getExtension(StateType stateType)
    {
      // make state type unique before using it for table lookup
      stateType = uniqueType(stateType);
      Integer i = stateTypes.get(stateType);
      if (i == null) return register(stateType);
      return getStateTypeExtension(i.intValue());
    }

    /***************************************************************************
     * Get similarity relation index for denoted pair of state types.
     * @param type1 the first type
     * @param type2 the second type
     * @return if positive, the index  i+1 of relation "type1 ~i type2"
     *         if negative, the index -i-1 of relation "type2 ~i type1"
     * (if the similarity relation has not yet been registered, it is now).
     **************************************************************************/
    private int getSimIndex(StateType type1, StateType type2)
    {
      // make state type unique before using it for table lookup
      type1 = uniqueType(type1);
      type2 = uniqueType(type2);
      int n = simType1.size();
      for (int i=0; i<n; i++)
      {
        Type t1 = simType1.get(i);
        Type t2 = simType2.get(i);
        if (t1 == type1 && t2 == type2) return i+1;
        if (t2 == type1 && t1 == type2) return -i-1;
      }
      String ext = getStateTypeExtension(n);
      generateSimDeclaration(ext, type1, type2);
      simType1.add(type1);
      simType2.add(type2);
      return n+1;
    }
    
    /***************************************************************************
     * Generate state type declarations with denoted extension.
     * @param stateType the state type for which to generate the declarations
     **************************************************************************/
    private void generateStateTypeDecls(StateType stateType)
    {
      // get state type extension
      String ext = getExtension(stateType);
      String typeName = stateTypeName + ext;
      
      // state type 
      Type returnType = stateType.getReturn();
      if (returnType == null)
      {
        add(TypeDefinition.construct(
            Identifier.construct(typeName),
            RecordType.construct(
                new TypedIdentifier[] {
                    TypedIdentifier.construct(
                        Identifier.construct(modeFieldName),
                        IntType.construct()),
                    TypedIdentifier.construct(
                        Identifier.construct(exceptionFieldName),
                        IntType.construct()),
                    TypedIdentifier.construct(
                        Identifier.construct(messageFieldName),
                        stringType)          
                }
            )));
      }
      else
      {
        add(TypeDefinition.construct(
            Identifier.construct(typeName),
            RecordType.construct(
                new TypedIdentifier[] {
                    TypedIdentifier.construct(
                        Identifier.construct(modeFieldName),
                        IntType.construct()),
                    TypedIdentifier.construct(
                        Identifier.construct(valueFieldName),
                        returnType),
                    TypedIdentifier.construct(
                        Identifier.construct(exceptionFieldName),
                        IntType.construct()),
                    TypedIdentifier.construct(
                        Identifier.construct(messageFieldName),
                        stringType)                        
                }
            )));
      }
      
      // constant for old state
      add(ValueDeclarationClass.construct(
          Identifier.construct(oldStateName + ext),
          Reference.construct(Identifier.construct(typeName))));
      
      // constant for new state
      add(ValueDeclarationClass.construct(
          Identifier.construct(newStateName + ext),
          Reference.construct(Identifier.construct(typeName))));
      
      // associated value function
      if (returnType != null)
      {
        add(ValueDeclarationInitialized.construct(
            Identifier.construct(stateValueName + ext),
            FunctionType.construct(
                new Type[] {Reference.construct(Identifier.construct(typeName))},
                translate(returnType)),
            LambdaTerm.construct(
                new TypedIdentifier[] {
                    TypedIdentifier.construct(
                        Identifier.construct("s"),
                        Reference.construct(Identifier.construct(typeName)))
                },
                SelectionTerm.construct(
                    Reference.construct(Identifier.construct("s")),
                    SelectorIdentifier.construct(
                        Identifier.construct(valueFieldName)))
            )));
      }
      
      // associated message function
      add(ValueDeclarationInitialized.construct(
          Identifier.construct(stateMessageName + ext),
          FunctionType.construct(
              new Type[] {Reference.construct(Identifier.construct(typeName))},
              stringType),
              LambdaTerm.construct(
                  new TypedIdentifier[] {
                      TypedIdentifier.construct(
                          Identifier.construct("s"),
                          Reference.construct(Identifier.construct(typeName)))
                  },
                  SelectionTerm.construct(
                      Reference.construct(Identifier.construct("s")),
                      SelectorIdentifier.construct(
                          Identifier.construct(messageFieldName)))
              )));
      
      // associated executes predicate
      add(ValueDeclarationInitialized.construct(
          Identifier.construct(stateExecutesFlagName + ext),
          IntType.construct(),
          IntLiteral.construct("0"))); 
      add(ValueDeclarationInitialized.construct(
          Identifier.construct(stateExecutesName + ext),
          FunctionType.construct(
              new Type[] {Reference.construct(Identifier.construct(typeName))},
              BooleanType.construct()),
              LambdaFormula.construct(
                  new TypedIdentifier[] {
                      TypedIdentifier.construct(
                          Identifier.construct("s"),
                          Reference.construct(Identifier.construct(typeName)))
                  },
                  EqualsFormula.construct(
                      SelectionTerm.construct(
                          Reference.construct(Identifier.construct("s")),
                          SelectorIdentifier.construct(
                              Identifier.construct(modeFieldName))),
                      Reference.construct(Identifier.construct(stateExecutesFlagName + ext))))));
      
      // associated continues predicate
      add(ValueDeclarationInitialized.construct(
          Identifier.construct(stateContinuesFlagName + ext),
          IntType.construct(),
          IntLiteral.construct("1"))); 
      add(ValueDeclarationInitialized.construct(
          Identifier.construct(stateContinuesName + ext),
          FunctionType.construct(
              new Type[] {Reference.construct(Identifier.construct(typeName))},
              BooleanType.construct()),
              LambdaFormula.construct(
                  new TypedIdentifier[] {
                      TypedIdentifier.construct(
                          Identifier.construct("s"),
                          Reference.construct(Identifier.construct(typeName)))
                  },
                  EqualsFormula.construct(
                      SelectionTerm.construct(
                          Reference.construct(Identifier.construct("s")),
                          SelectorIdentifier.construct(
                              Identifier.construct(modeFieldName))),
                      Reference.construct(Identifier.construct(stateContinuesFlagName + ext))))));
      
      // associated breaks predicate
      add(ValueDeclarationInitialized.construct(
          Identifier.construct(stateBreaksFlagName + ext),
          IntType.construct(),
          IntLiteral.construct("2")));  
      add(ValueDeclarationInitialized.construct(
          Identifier.construct(stateBreaksName + ext),
          FunctionType.construct(
              new Type[] {Reference.construct(Identifier.construct(typeName))},
              BooleanType.construct()),
              LambdaFormula.construct(
                  new TypedIdentifier[] {
                      TypedIdentifier.construct(
                          Identifier.construct("s"),
                          Reference.construct(Identifier.construct(typeName)))
                  },
                  EqualsFormula.construct(
                      SelectionTerm.construct(
                          Reference.construct(Identifier.construct("s")),
                          SelectorIdentifier.construct(
                              Identifier.construct(modeFieldName))),
                      Reference.construct(Identifier.construct(stateBreaksFlagName + ext)))
              )
      ));
      
      // associated returns predicate
      add(ValueDeclarationInitialized.construct(
          Identifier.construct(stateReturnsFlagName + ext),
          IntType.construct(),
          IntLiteral.construct("3"))); 
      add(ValueDeclarationInitialized.construct(
          Identifier.construct(stateReturnsName + ext),
          FunctionType.construct(
              new Type[] {Reference.construct(Identifier.construct(typeName))},
              BooleanType.construct()),
              LambdaFormula.construct(
                  new TypedIdentifier[] {
                      TypedIdentifier.construct(
                          Identifier.construct("s"),
                          Reference.construct(Identifier.construct(typeName)))
                  },
                  EqualsFormula.construct(
                      SelectionTerm.construct(
                          Reference.construct(Identifier.construct("s")),
                          SelectorIdentifier.construct(
                              Identifier.construct(modeFieldName))),
                      Reference.construct(Identifier.construct(stateReturnsFlagName + ext)))
              )));
      
      // associated throws predicate
      add(ValueDeclarationInitialized.construct(
          Identifier.construct(stateThrowsFlagName + ext),
          IntType.construct(),
          IntLiteral.construct("4"))); 
      add(ValueDeclarationInitialized.construct(
          Identifier.construct(stateThrowsName + ext),
          FunctionType.construct(
              new Type[] {Reference.construct(Identifier.construct(typeName))},
              BooleanType.construct()),
              LambdaFormula.construct(
                  new TypedIdentifier[] {
                      TypedIdentifier.construct(
                          Identifier.construct("s"),
                          Reference.construct(Identifier.construct(typeName)))
                  },
                  EqualsFormula.construct(
                      SelectionTerm.construct(
                          Reference.construct(Identifier.construct("s")),
                          SelectorIdentifier.construct(
                              Identifier.construct(modeFieldName))),
                      Reference.construct(Identifier.construct(stateThrowsFlagName + ext)))
              )));
      
      // associated throws exception predicate
      add(ValueDeclarationInitialized.construct(
          Identifier.construct(stateThrowsExceptionName + ext),
          FunctionType.construct(
              new Type[] {
                  Reference.construct(Identifier.construct(typeName)),
                  IntType.construct()},
              BooleanType.construct()),
              LambdaFormula.construct(
                  new TypedIdentifier[] {
                      TypedIdentifier.construct(
                          Identifier.construct("s"),
                          Reference.construct(Identifier.construct(typeName))),
                      TypedIdentifier.construct(
                          Identifier.construct("e"),
                          IntType.construct())   
                  },
                  AndFormula.construct(
                      EqualsFormula.construct(
                          SelectionTerm.construct(
                              Reference.construct(Identifier.construct("s")),
                              SelectorIdentifier.construct(
                                  Identifier.construct(modeFieldName))),
                                  Reference.construct(Identifier.construct(stateThrowsFlagName + ext))),
                      EqualsFormula.construct(
                          SelectionTerm.construct(
                              Reference.construct(Identifier.construct("s")),
                              SelectorIdentifier.construct(
                                  Identifier.construct(exceptionFieldName))),
                                  Reference.construct(Identifier.construct("e"))))                                 
              )));
    }
    
    /*************************************************************************
     * Generate declaration of similarity predicate with denoted extension
     * for denoted types.
     * @param ext the extension of the predicate name.
     * @param type1 the first type.
     * @param type2 the second type.
     ************************************************************************/
    private void generateSimDeclaration(String ext, 
      StateType type1, StateType type2)
    {
      // get state type extensions
      String ext1 = getExtension(type1);
      String name1 = stateTypeName + ext1;
      String ext2 = getExtension(type2);
      String name2 = stateTypeName + ext2;
      
      // associated similar predicate
      add(ValueDeclarationInitialized.construct(
          Identifier.construct(stateSimilarName + ext),
          FunctionType.construct(
              new Type[] {
                  Reference.construct(Identifier.construct(name1)),
                  Reference.construct(Identifier.construct(name2))},
              BooleanType.construct()),
              LambdaFormula.construct(
                  new TypedIdentifier[] {
                      TypedIdentifier.construct(
                          Identifier.construct("s1"),
                          Reference.construct(Identifier.construct(name1))),
                      TypedIdentifier.construct(
                          Identifier.construct("s2"),
                          Reference.construct(Identifier.construct(name2)))
                  },
                  AndFormula.construct(
                    EqualsFormula.construct(
                      SelectionTerm.construct(
                          Reference.construct(Identifier.construct("s1")),
                          SelectorIdentifier.construct(
                              Identifier.construct(modeFieldName))),
                      SelectionTerm.construct(
                          Reference.construct(Identifier.construct("s2")),
                          SelectorIdentifier.construct(
                              Identifier.construct(modeFieldName)))),
                    AndFormula.construct(
                        EqualsFormula.construct(
                            SelectionTerm.construct(
                                Reference.construct(Identifier.construct("s1")),
                                SelectorIdentifier.construct(
                                    Identifier.construct(exceptionFieldName))),
                            SelectionTerm.construct(
                                Reference.construct(Identifier.construct("s2")),
                                SelectorIdentifier.construct(
                                    Identifier.construct(exceptionFieldName)))),
                         EqualsFormula.construct(
                            SelectionTerm.construct(
                                Reference.construct(Identifier.construct("s1")),
                                SelectorIdentifier.construct(
                                  Identifier.construct(messageFieldName))),
                            SelectionTerm.construct(
                                Reference.construct(Identifier.construct("s2")),
                                SelectorIdentifier.construct(
                                  Identifier.construct(messageFieldName)))))))));
    }
    
    /****************************************************************************
     * visit formula.
     * @param formula the formula to be visitd.
     * @return the translation
     ***************************************************************************/
    public void visit(WritesOnlyFormula formula)
    {
      result = unchangedExcept(Normalization.modifiableVariables(formula.getVariables()));
    }
    
    /****************************************************************************
     * visit formula.
     * @param formula the formula to be visitd.
     * @return the translation
     ***************************************************************************/
    public void visit(ReadsOnlyFormula formula)
    {
      result = unchangedExcept(null);
    }
    
    /****************************************************************************
     * Generate formula that asserts that all modifiable program variables
     * remain unchanged except those occurring in the given set.
     * @param vars the variables that may be changed (may be null)
     * @return a formula stating the assertion.
     ***************************************************************************/
    private Formula unchangedExcept(Collection<VariableSymbol> vars)
    {
      Formula result = null;
      for (VariableSymbol var: stateVarSymbols)
      {
        if (vars != null && vars.contains(var)) continue;
        Formula equality =
          EqualsFormula.construct(
              Reference.construct(
                  Identifier.construct(constantName(var, true))),
                  Reference.construct(
                      Identifier.construct(constantName(var, false))));
        if (result == null)
          result = equality;
        else
          result = AndFormula.construct(result, equality);
      }
      if (result == null) result = ExpressionLogic.booleanFormula(true);
      return result;
    }
    
    /****************************************************************************
     * Translate formula.
     * @param formula the formula to be translated.
     * @return the translation
     ***************************************************************************/
    public void visit(StateBreaks formula)
    {
      Term arg = formula.getArgument();
      Type type = arg.getType();
      if (!(type instanceof StateType)) Main.internalError("not a state type");
      String ext = getExtension((StateType)type);
      result = AtomicFormula.construct(
          Reference.construct(Identifier.construct(stateBreaksName + ext)),
          new Term[] {
            translate(arg)
          });
    }
    
    /****************************************************************************
     * Translate formula.
     * @param formula the formula to be translated.
     * @return the translation
     ***************************************************************************/
    public void visit(StateContinues formula)
    {
      Term arg = formula.getArgument();
      Type type = arg.getType();
      if (!(type instanceof StateType)) Main.internalError("not a state type");
      String ext = getExtension((StateType)type);
      result = AtomicFormula.construct(
          Reference.construct(Identifier.construct(stateContinuesName + ext)),
          new Term[] {
            translate(arg)
          });
    }
    
    /****************************************************************************
     * Translate formula.
     * @param formula the formula to be translated.
     * @return the translation
     ***************************************************************************/
    public void visit(StateExecutes formula)
    {
      Term arg = formula.getArgument();
      Type type = arg.getType();
      if (!(type instanceof StateType)) 
        Main.internalError("not a state type: " + formula + " " + type);
      String ext = getExtension((StateType)type);
      result = AtomicFormula.construct(
          Reference.construct(Identifier.construct(stateExecutesName + ext)),
          new Term[] {
            translate(arg)
          });
    }
    
    /****************************************************************************
     * Translate formula.
     * @param formula the formula to be translated.
     * @return the translation
     ***************************************************************************/
    public void visit(StateReturns formula)
    {
      Term arg = formula.getArgument();
      Type type = arg.getType();
      if (!(type instanceof StateType)) Main.internalError("not a state type");
      String ext = getExtension((StateType)type);
      result = AtomicFormula.construct(
          Reference.construct(Identifier.construct(stateReturnsName + ext)),
          new Term[] {
            translate(arg)
          });
    }
    
    /****************************************************************************
     * Translate formula.
     * @param formula the formula to be translated.
     * @return the translation
     ***************************************************************************/
    public void visit(StateThrows formula)
    {
      Term arg = formula.getArgument();
      Type type = arg.getType();
      if (!(type instanceof StateType)) Main.internalError("not a state type");
      String ext = getExtension((StateType)type);
      result = AtomicFormula.construct(
          Reference.construct(Identifier.construct(stateThrowsName + ext)),
          new Term[] {
            translate(arg)
          });
    }
    
    /****************************************************************************
     * Translate formula.
     * @param formula the formula to be translated.
     * @return the translation
     ***************************************************************************/
    public void visit(StateThrowsException formula)
    {
      Term arg = formula.getArgument();
      Name name = formula.getException();
      Type type = arg.getType();
      if (!(type instanceof StateType)) Main.internalError("not a state type");
      String ext = getExtension((StateType)type);
      result = AtomicFormula.construct(
          Reference.construct(Identifier.construct(stateThrowsExceptionName + ext)),
          new Term[] {
            translate(arg),
            Reference.construct(Identifier.construct(exceptionName(name)))
          });
    }
    
    /****************************************************************************
     * Translate formula.
     * @param formula the formula to be translated.
     * @return the translation
     ***************************************************************************/
    public void visit(SimilarFormula formula)
    {
      Term first = formula.getFirst();
      Term second = formula.getSecond();
      StateType type1 = (StateType)first.getType();
      StateType type2 = (StateType)second.getType();
      int index = getSimIndex(type1, type2);
      if (index > 0)
        index = index-1;
      else 
      {
        index = -index-1;
        Term term = first;
        first = second;
        second = term;
      }
      String ext = getStateTypeExtension(index);
      result = AtomicFormula.construct(
          Reference.construct(Identifier.construct(stateSimilarName + ext)),
          new Term[] { translate(first), translate(second) });
    }
    
    /***************************************************************************
     * Translate program variable.
     * @param var the program variable to be translated.
     * @return the translation.
     **************************************************************************/
    private void visitProgramVariable(ProgramVariableBase var)
    {
      Name name = var.getName();
      int pos = NameUtils.getVariablePosition(name);
      VariableSymbol symbol = (VariableSymbol)name.idents[pos].getSymbol();
      boolean prestate = 
        (var instanceof PreVariable) ||
        (!stateVarSymbols.contains(symbol));
      String cname = constantName(symbol, prestate);
      result = Reference.construct(Identifier.construct(cname));
      int n = name.idents.length;
      for (int i=pos+1; i<n; i++)
      {
        SelectorIdentifier sel = SelectorIdentifier.construct(name.idents[i]);
        result = SelectionTerm.construct((Term)result, sel);
      }
    }
    
    /***************************************************************************
     * Translate program variable.
     * @param var the program variable to be translated.
     * @return the translation.
     **************************************************************************/
    public void visit(PostVariable var)
    {
      visitProgramVariable(var);
    }
    
    /***************************************************************************
     * Translate program variable.
     * @param var the program variable to be translated.
     * @return the translation.
     **************************************************************************/
    public void visit(PreVariable var)
    {
      visitProgramVariable(var);
    }
    
    /****************************************************************************
     * Translate term.
     * @param term the term to be translated.
     * @return the translation
     ***************************************************************************/
    public void visit(StateValue term)
    {
      Term arg = term.getArgument();
      Type type = arg.getType();
      if (!(type instanceof StateType)) 
        Main.internalError("not a state type");
      String ext = getExtension((StateType)type);
      result = FunctionApplication.construct(
          Reference.construct(Identifier.construct(stateValueName + ext)),
          new Term[] { translate(arg) });
    }
    
    /****************************************************************************
     * Translate term.
     * @param term the term to be translated.
     * @return the translation
     ***************************************************************************/
    public void visit(StateMessage term)
    {
      Term arg = term.getArgument();
      Type type = arg.getType();
      if (!(type instanceof StateType)) 
        Main.internalError("not a state type");
      String ext = getExtension((StateType)type);
      result = FunctionApplication.construct(
          Reference.construct(Identifier.construct(stateMessageName + ext)),
          new Term[] { translate(arg) });
    }
    
    /****************************************************************************
     * Translate term.
     * @param term the term to be translated.
     * @return the translation
     ***************************************************************************/
    public void visit(NewState term)
    {
      String ext = getExtension(stateType);
      result = Reference.construct(Identifier.construct(newStateName + ext));
    }
    
    /****************************************************************************
     * Translate term.
     * @param term the term to be translated.
     * @return the translation
     ***************************************************************************/
    public void visit(OldState term)
    {
      String ext = getExtension(stateType);
      result = Reference.construct(Identifier.construct(oldStateName + ext));
    }
    
    /****************************************************************************
     * Translate type.
     * @param type the type to be translated.
     * @return the translation
     ***************************************************************************/
    public void visit(StateType type)
    {
      String ext = getExtension(type);
      result = Reference.construct(Identifier.construct(stateTypeName + ext));
    }
    
    /***************************************************************************
     * Create clone of formula and make sure that result is a formula
     * (not just a boolean-valued term).
     * @param tree a formula.
     * @return the clone of the formula.
     **************************************************************************/
    protected Formula cloneFormula(Formula tree)
    {
      AST result = clone0(tree);
      if (result instanceof Formula) return (Formula)result;
      Term result0 = (Term)result;
      Term t = ExpressionLogic.booleanFormula(true);
      return EqualsFormula.construct(result0, t);
    }
  }
}

