// ----------------------------------------------------------------------------
// FileTree.java
// A tree displaying directory structures.
// $Id: FileTree.java,v 1.19 2017/10/24 08:28:35 schreine Exp schreine $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Fileic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.SWT;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;

import java.io.*;
import java.util.*;

import org.eclipse.swt.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;

/**
 * 
 * A tree displaying directory structures.
 *
 */
public final class FileTree
{  
  // the tree (subclassing not allowed)
  private Tree tree;
  
  // the images used for tree nodes
  private Image sourceImage;
  private Image fileImage;
  private Image folderImage;
  private Image refreshImage;
  private Image deleteImage;
  
  // all items in the tree and the associated files
  private Map<TreeItem,File> itemSet;
  private Map<File,FileItem> fileSet;
  
  /***************************************************************************
   * A tree entry.
   **************************************************************************/
  private final class FileItem
  {
    TreeItem node; // the tree item
    
    /*************************************************************************
     * Create a node linked to parent node.
     * @param parent the parent (object of type Tree or FileItem)
     * @param file the file represented by the node.
     ************************************************************************/
    public FileItem(Object parent, File file)
    {
      // create tree node
      if (parent instanceof Tree)
        node = new TreeItem((Tree)parent, SWT.NULL);
      else if (parent instanceof FileItem)
        node = new TreeItem(((FileItem)parent).node, SWT.NULL);
      
      // register file item
      itemSet.put(node, file);
      fileSet.put(file, this);
      
      // set text and image
      node.setText(file.getName());
      if (!file.isDirectory()) 
      {
        if (Source.isSource(file))
          node.setImage(sourceImage);
        else
          node.setImage(fileImage);
        return;
      }
      node.setImage(folderImage);
    }
  }
  
  /****************************************************************************
   * Popup information on a file
   ***************************************************************************/
  private static class FileTip 
  {
    private Shell shell;
    private Label label;
    
    // only one instance of file tip at a time
    private static FileTip tip = null;
    
    /**************************************************************************
     * Create popup window displaying information on given file.
     * @param parent the parent window.
     * @param pos the position of the window.
     * @param file the file.
     *************************************************************************/
    public FileTip(Shell parent, Point pos, File file)
    {     
      close();
      tip = this;
      
      shell = new Shell(parent, SWT.ON_TOP | SWT.TOOL); 
      GridLayout gridLayout = new GridLayout();
      gridLayout.numColumns = 1;
      gridLayout.marginWidth = 2;
      gridLayout.marginHeight = 2;
      shell.setLayout(gridLayout);
      Color color = shell.getDisplay().getSystemColor(SWT.COLOR_INFO_BACKGROUND);
      Color tcolor = shell.getDisplay().getSystemColor(SWT.COLOR_INFO_FOREGROUND);
      shell.setBackground(color);
      
      // determine text
      String text;
      if (file.isDirectory())
        text = "Directory " + file.getAbsolutePath();
      else
        text = "File " + file.getAbsolutePath();
      
      // set label
      label = new Label(shell, SWT.NONE);
      label.setLayoutData(new GridData(GridData.FILL_HORIZONTAL
          | GridData.VERTICAL_ALIGN_CENTER));
      label.setText(text);
      label.setBackground(color);
      label.setForeground(tcolor);
      
      shell.addMouseTrackListener(new MouseTrackAdapter() { 
        public void mouseEnter(MouseEvent e)
        {
          close();
        }
      });
      
      // open window
      shell.setLocation(pos.x+10, pos.y+10);
      shell.pack();
      shell.open();
    }
    
    /*************************************************************************
     * Close tip instance.
     ************************************************************************/
    public static void close()
    {
      if (tip != null) tip.shell.close();
      tip = null;
    }
  }
  
  /***************************************************************************
   * Create a tree displaying directory structures.
   * @param parent the parent widget.
   * @param top the top window.
   **************************************************************************/
  public FileTree(Composite parent, final TopWindow top)
  {    
    // set images
    sourceImage = top.getSourceImage();
    fileImage = top.getFileImage();
    folderImage = top.getFolderImage();
    refreshImage = top.getRefreshImage();
    deleteImage = top.getDeleteImage();
    
    // initialize tree
    tree = new Tree(parent, SWT.SINGLE);
    
    // initialize node/file sets
    initSets();
    
    // set menu
    final Menu menu = new Menu(tree.getShell(), SWT.POP_UP);
    tree.setMenu(menu);
    menu.addListener (SWT.Show, new Listener () {
      public void handleEvent (Event event) {
        MenuItem [] menuItems = menu.getItems ();
        for (int i=0; i<menuItems.length; i++) {
          menuItems [i].dispose ();
        }
        TreeItem [] treeItems = tree.getSelection ();
        if (treeItems.length != 1) return;
        TreeItem item = treeItems[0];
        final File file = itemSet.get(item);
        if (file.isDirectory()) 
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText ("Refresh");
          menuItem.setImage(refreshImage);
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              refreshDirectory(file);
            }
          });
        }
        if (Source.isSource(file))
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText ("Open");
          menuItem.setImage(sourceImage);
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              top.createEditor(file);
            }
          });
        }
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText ("Delete");
          menuItem.setImage(deleteImage);
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              top.delete(file);
            }
          });
        }
      }
    });

    // set click handlers
    tree.addSelectionListener(new SelectionAdapter() {
      
      // item is double clicked
      public void widgetDefaultSelected(SelectionEvent e) 
      { 
        TreeItem item = (TreeItem)e.item;
        File file = itemSet.get(item);
        if (file == null) return;
        if (!Source.isSource(file)) return;
        top.createEditor(file);
      }
      
      // item is selected by single click
      public void widgetSelected(SelectionEvent e) {
      }
    });
    
    // set hover handler
    final Composite parent0 = parent;
    tree.addMouseTrackListener(new MouseTrackAdapter() {
      public void mouseHover(MouseEvent e) {
        Tree tree = (Tree)e.widget;
        File file = find(tree.getItems(), e.x, e.y);
        if (file == null) return;
        new FileTip(parent0.getShell(), tree.toDisplay(e.x, e.y), file);
      }
      public void mouseExit(MouseEvent e) {
        FileTip.close();
      }
    });
  }

  /****************************************************************************
   * Find file associated to item in set of items denoted by mouse position.
   * @param x horizontal coordinate of mouse position.
   * @param y vertical coordinate of mouse position.
   * @return the denoted file (or null, if none was found)
   ***************************************************************************/
  private File find(TreeItem[] items, int x, int y)
  {
    for (TreeItem item : items)
    {
      TreeItem[] items0 = item.getItems();
      File file = find(items0, x, y);
      if (file != null) return file;
      file = itemSet.get(item);
      if (file == null) continue;
      Rectangle b = item.getBounds();
      if (x < b.x || x > b.x + b.width) continue;
      if (y < b.y || y > b.y + b.height) continue;
      return file;
    }
    return null;
  }
  
  /****************************************************************************
   * Initizialize node/file sets.
   ***************************************************************************/
  private void initSets()
  {
    itemSet = new MapClass<TreeItem,File>();
    fileSet = new MapClass<File,FileItem>();
  }
  
  /****************************************************************************
   * Get visual representation of tree.
   * @return the tree.
   ***************************************************************************/
  public Tree getTree()
  {
    return tree;
  }
  
  /****************************************************************************
   * Redraw directory structures.
   * @param roots the root files/directories for display.
   ***************************************************************************/
  public void redraw(File[] roots)
  {
    tree.removeAll();
    initSets();
    for (File root : roots)
    {
      redraw(tree, root);
    }
  }
  
  private static class FileComparator implements Comparator<File>
  {
    public int compare(File file1, File file2)
    {
      boolean dir1 = file1.isDirectory();
      boolean dir2 = file2.isDirectory();
      if (dir1 && !dir2) return -1;
      if (dir2 && !dir1) return +1;
      return file1.getName().compareTo(file2.getName());
    }
  }
  
  /***************************************************************************
   * Redraw file linked into parent component.
   * @param parent the parent component (type Tree or FileItem).
   * @param file the file/directory to be displayed.
   **************************************************************************/
  private void redraw(Object parent, File file)
  {
    // improve performance by avoiding the auxiliary directories
    if (file.getName().startsWith(".")) return;
    
    // display file
    FileItem node = new FileItem(parent, file);
    
    // file may be directory or plain file
    if (!file.isDirectory()) return;
    
    // display files in directory
    File[] files = file.listFiles();
    if (files == null) return;
    Arrays.sort(files, new FileComparator());
    for (File f : files)
    {
      redraw(node, f);
    }
  }

  /***************************************************************************
   * A new file may have been created, update view.
   * @param file potential new file.
   **************************************************************************/
  public void update(File file)
  {
    if (fileSet.containsKey(file)) return;
    File parent = file.getParentFile();
    if (parent == null) return;
    update(parent);
    FileItem node = fileSet.get(parent);
    if (node == null) return;
    redraw(node, file);
  }
  
  /****************************************************************************
   * Refresh directory item.
   * @param refresh item.
   ***************************************************************************/
  private void refreshDirectory(File dir)
  {
    FileItem item = fileSet.get(dir);
    TreeItem node = item.node;
    boolean expanded = node.getExpanded();
    node.removeAll();
    removeKeys(dir);
    File[] files = dir.listFiles();
    Arrays.sort(files, new FileComparator());
    for (File file : files)
      redraw(item, file);
    node.setExpanded(expanded);
  }
  
  /****************************************************************************
   * Remove keys of files below directory.
   * @param dir the directory.
   ***************************************************************************/
  private void removeKeys(File dir)
  {
    String dpath = dir.getAbsolutePath();
    @SuppressWarnings("unchecked")
    Set<File> files = 
      ((Map<File,FileItem>)((MapClass<File,FileItem>)fileSet).clone()).keySet();
    for (File f : files)
    {
      String fpath = f.getAbsolutePath();
      if (fpath.equals(dpath)) continue;
      if (!fpath.startsWith(dpath)) continue;
      FileItem item = fileSet.remove(f);
      itemSet.remove(item.node);
    }
  }
  
  /****************************************************************************
   * Delete file from tree.
   * @param file the file to be deleted.
   ***************************************************************************/
  public void delete(File file)
  {
    FileItem item = fileSet.get(file);
    item.node.dispose();
    itemSet.remove(item.node);
    fileSet.remove(file);
  } 
}
