// ----------------------------------------------------------------------------
// MainSWT.java
// The SWT-based GUI.
// $Id: MainSWT.java,v 1.23 2017/07/04 08:04:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.SWT;

import java.util.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.widgets.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;

/**
 * 
 * The SWT-based GUI.
 * 
 */
public final class MainSWT 
{ 
  // the single instance of this class
  private static MainSWT mainSWT;
  
  // the application icon
  public static final String riscImageName = "risc-logo-transparent.png";
  
  // the display associated to the GUI
  private Display display;
  
  // the top-level window
  private TopWindow top;
  
  // the thread executing some application request
  private Thread thread;
  
  // various colors
  private Color backgroundColor; // background color
  
  // the virtual directories, their number and their server
  private int dirCounter;
  private Map<String, VirtualDirectory> directories;
  private VirtualDirectoryServer server;

  // the command server
  private CommandServer cserver;
  
  /***************************************************************************
   * Run GUI.
   **************************************************************************/
  public static void main()
  {
    mainSWT = new MainSWT();
    mainSWT.run();
  }
  
  /***************************************************************************
   * Create an instance of the GUI.
   **************************************************************************/
  private MainSWT()
  { 
  }
  
  /***************************************************************************
   * Run the GUI in this thread.
   **************************************************************************/
  private void run()
  {  
    // set local state
    dirCounter = 0;
    directories = new MapClass<String, VirtualDirectory>();
    server = new VirtualDirectoryServer(directories);
    
    // create display
    display = new Display(); 

    // create colors
    backgroundColor = new Color(display, new RGB(255, 255, 255));
    
    // create top-level window 
    top = TopWindow.construct(this);
    
    // create command server
    cserver = new CommandServer(top);
    
    // open window
    top.resize(); // avoid visible resize of window
    top.open();
    top.resize(); // fix problem with areas cut off at bottom
    
    // perform initial application processing
    execute(new Runnable() 
    {
      public void run() 
      {
        Main.init(top);
        top.processPackages();
        top.closeTaskFolders();
        terminate();
      }
    });
    
    //open window and process events until window is closed
    try
    {
    Shell shell = top.getShell();
    while (!shell.isDisposed())
    {
      if (!display.readAndDispatch())
        display.sleep();
    }
    }
    catch(org.eclipse.swt.SWTException e)
    {
      System.out.println(e.getMessage());
    }
  }
  
  /***************************************************************************
   * Execute application request.
   * @param request the request.
   **************************************************************************/
  public void execute(Runnable request)
  {
    if (thread != null) return;
    thread = new Thread(request);
    thread.start();
  }
  
  /***************************************************************************
   * Execute display update request on behalf of application.
   * @param request the request.
   **************************************************************************/
  public void update(Runnable request)
  {
    display.syncExec(request);
  }
  
  /***************************************************************************
   * Indicate termination of application request.
   * Must be only called by the thread that has executed the request.
   **************************************************************************/
  public void terminate()
  {
    if (thread != Thread.currentThread())
      Main.internalError("foreign thread indicated termination");
    thread = null;
  }
  
  /***************************************************************************
   * Get the display of this GUI.
   * @return the display.
   **************************************************************************/
  public Display getDisplay()
  {
    return display;
  }
  
  /***************************************************************************
   * Get background color.
   * @return the color
   **************************************************************************/
  public Color getBackgroundColor()
  {
    return backgroundColor;
  }
  
  /***************************************************************************
   * Display analyze view.
   **************************************************************************/
  public static void displayAnalyzeView()
  {
    mainSWT.top.getDisplay().syncExec(new Runnable() {
      public void run() {
        mainSWT.top.displayAnalyzeView();
      }});
  }
  
  /***************************************************************************
   * Display verify view.
   **************************************************************************/
  public static void displayVerifyView()
  {
    mainSWT.top.getDisplay().syncExec(new Runnable() {
      public void run() {
        mainSWT.top.displayVerifyView();
      }});
  }
  
  /*************************************************************************
   * Set location of shell at pos but ensure that content is fully visible.
   * @param shell the shell.
   * @param x first coordinate of suggested position.
   * @param y second coordinate of suggested position.
   ************************************************************************/
  public static void setLocation(Shell shell, int x, int y)
  {
    Point size = shell.getSize();
    Display display = shell.getDisplay();
    Rectangle r = display.getClientArea();
    if (x+size.x+5 > r.width)
    {
      x = r.width-size.x-5;
      if (x < 0) x=0;
    }
    if (y+size.y+5 > r.height)
    {
      y = r.height-size.y-5;
      if (y < 0) y=0;
    }
    shell.setLocation(x, y);
  }
  
  /***************************************************************************
   * Enter a new virtual directory.
   * @param directory the directory.
   * @return the unique identifier for the directory.
   **************************************************************************/
  public String enterVirtualDirectory(VirtualDirectory directory)
  {
    String name = String.valueOf("dir" + dirCounter);
    dirCounter = dirCounter+1;
    directories.put(name, directory);
    return name;
  }
  
  /***************************************************************************
   * Delete a virtual directory.
   * @param name the name of the directory to be deleted.
   **************************************************************************/
  public void deleteVirtualDirectory(String name)
  {
    directories.remove(name);
  }
  
  /***************************************************************************
   * Get port number of virtual directory server.
   * @return the number.
   **************************************************************************/
  public int getVirtualDirectoryPort()
  {
    return server.getPortNumber();
  }
  
  /***************************************************************************
   * Get port of command server.
   * @return the port of the command server.
   **************************************************************************/
  public int getCommandServerPort()
  {
    return cserver.getPortNumber();
  }
}
