// ----------------------------------------------------------------------------
// TaskList.java
// A list displaying not done verification tasks.
// $Id: TaskList.java,v 1.12 2017/10/24 08:28:35 schreine Exp schreine $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.SWT;

import java.util.*;
import java.io.*;

import org.eclipse.swt.*;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Tasks.*;
import fmrisc.ProgramExplorer.Proving.*;

/**
 * 
 * A list displaying not done verification tasks.
 *
 */
public final class TaskList
{  
  // the underlying tree (subclassing not allowed)
  private Tree tree;
  
  // the top window
  private TopWindow top;
  
  // mapping of tree items to tasks
  private Map<TreeItem,Task> itemSet;
  
  /****************************************************************************
   * Popup information on a task
   ***************************************************************************/
  private static class TaskTip 
  {
    private Shell shell;
    private Label label;
    
    // only one instance of task tip at a time
    private static TaskTip tip = null;
    
    /**************************************************************************
     * Create popup window displaying information on given task.
     * @param parent the parent window.
     * @param pos the position of the window.
     * @param task the task.
     *************************************************************************/
    public TaskTip(Shell parent, Point pos, Task task)
    {     
      close();
      tip = this;
      
      shell = new Shell(parent, SWT.ON_TOP | SWT.TOOL); 
      GridLayout gridLayout = new GridLayout();
      gridLayout.numColumns = 1;
      gridLayout.marginWidth = 2;
      gridLayout.marginHeight = 2;
      shell.setLayout(gridLayout);
      Color color = shell.getDisplay().getSystemColor(SWT.COLOR_INFO_BACKGROUND);
      Color tcolor = shell.getDisplay().getSystemColor(SWT.COLOR_INFO_FOREGROUND);
      shell.setBackground(color);
      
      // determine text
      StringWriter out = new StringWriter();
      task.print(new PrintWriter(out));
      String text = out.toString();
      text = text.substring(0, text.length()-1); // remove trailing newline
      
      // set label
      label = new Label(shell, SWT.NONE);
      label.setLayoutData(new GridData(GridData.FILL_HORIZONTAL
          | GridData.VERTICAL_ALIGN_CENTER));
      label.setText(text);
      label.setBackground(color);
      label.setForeground(tcolor);
      
      shell.addMouseTrackListener(new MouseTrackAdapter() { 
        public void mouseEnter(MouseEvent e)
        {
          close();
        }
      });
      
      // open window
      shell.pack();
      MainSWT.setLocation(shell, pos.x+10, pos.y+10);
      shell.open();
    }
      
    /*************************************************************************
     * Close tip instance.
     ************************************************************************/
    public static void close()
    {
      if (tip != null) tip.shell.close();
      tip = null;
    }
  }
  
  /***************************************************************************
   * Create a list displaying all not done tasks from a folder (recursively).
   * @param parent the parent widget.
   * @param top the top window.
   **************************************************************************/
  public TaskList(Composite parent, final TopWindow top)
  {    
    this.top = top;
    this.tree = new Tree(parent, SWT.SINGLE);
    reset();
    
    // create menu
    createMenu();
    
    // set click handlers
    tree.addSelectionListener(new SelectionAdapter() {
      
      // item is double clicked
      public void widgetDefaultSelected(SelectionEvent e) 
      { 
        TreeItem item = (TreeItem)e.item;
        Task task = itemSet.get(item);
        if (task == null) return;
        SourcePosition pos = task.getPosition();
        if (pos == null) return;
        top.displaySource(pos, true);
      }
      
      // item is selected by single click
      public void widgetSelected(SelectionEvent e) {
      }
    });
    
    // set hover handler
    final Composite parent0 = parent;
    tree.addMouseTrackListener(new MouseTrackAdapter() {
      public void mouseHover(MouseEvent e) {
        Tree tree = (Tree)e.widget;
        Task task = find(tree.getItems(), e.x, e.y);
        if (task == null) return;
        new TaskTip(parent0.getShell(), tree.toDisplay(e.x, e.y), task);
      }
      public void mouseExit(MouseEvent e) {
        TaskTip.close();
      }
    });
  }

  /****************************************************************************
   * Find task associated to item in set of items denoted by mouse position.
   * @param items the set of tree items to be recursively searched.
   * @param x horizontal coordinate of mouse position.
   * @param y vertical coordinate of mouse position.
   * @return the denoted task (or null, if none was found)
   ***************************************************************************/
  private Task find(TreeItem[] items, int x, int y)
  {
    for (TreeItem item : items)
    {
      TreeItem[] items0 = item.getItems();
      Task task = find(items0, x, y);
      if (task != null) return task;
      task = itemSet.get(item);
      if (task == null || task instanceof TaskFolder) continue;
      Rectangle b = item.getBounds();
      if (x < b.x || x > b.x + b.width) continue;
      if (y < b.y || y > b.y + b.height) continue;
      return task;
    }
    return null;
  }
  

  
  /****************************************************************************
   * Execute task associated to tree item.
   * @param node the tree item denoting a task.
   ***************************************************************************/
  private void executeTask(TreeItem node)
  {
    final Task task = itemSet.get(node);
    
    // create thread executing the task
    new Thread(
        new Runnable() { public void run() { task.resume(); } }
        ).start();
  }
  
  /****************************************************************************
   * Redraw all information;
   ***************************************************************************/
  private void reset()
  {
    itemSet = new MapClass<TreeItem,Task>();
    tree.removeAll();
  }
  
  /****************************************************************************
   * Get the underlying tree representation.
   * @return the tree representation.
   ***************************************************************************/
  public Tree getTree()
  {
    return tree;
  }
  
  /****************************************************************************
   * Redraw all not done tasks in folder.
   * @param folder the task folder.
   ***************************************************************************/
  public void redraw(TaskFolder folder)
  {
    reset();
    redrawFolder(folder);
  }
  
  /***************************************************************************
   * Create menu for tree
   **************************************************************************/
  private void createMenu()
  {
    final Menu menu = new Menu(tree.getShell(), SWT.POP_UP);
    tree.setMenu(menu);
    menu.addListener (SWT.Show, new Listener () {
      public void handleEvent (Event event) {
        MenuItem [] menuItems = menu.getItems ();
        for (int i=0; i<menuItems.length; i++) {
          menuItems [i].dispose ();
        }
        TreeItem [] treeItems = tree.getSelection ();
        if (treeItems.length != 1) return;
        final TreeItem item = treeItems[0];
        final Task task = itemSet.get(item);
        if (!task.isFinal())
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText ("Execute Task");
          menuItem.setImage(top.getRunImage());
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              executeTask(item);
            }
          });
        }
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText ("Print Task");
          menuItem.setImage(top.getAnalyzeImage());
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              task.print(top.getConsole());
            }
          });
        }
        if (task instanceof StateTask)
        {
          final StateTask task0 = (StateTask)task;
          {
            MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
            menuItem.setText("Print State Proving Problem");
            menuItem.setImage(top.getAnalyzeImage());
            menuItem.addSelectionListener(new SelectionAdapter() {
              public void widgetSelected(SelectionEvent e) {
                PrintWriter out = top.getConsole();
                StateProblem task1 = task0.stateProblem();
                task1.print(out);
              }
            });
          }
        }
        if (task instanceof ClassicalTask)
        {
          final ClassicalTask task0 = (ClassicalTask)task;
          {
            MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
            menuItem.setText("Print Classical Proving Problem");
            menuItem.setImage(top.getAnalyzeImage());
            menuItem.addSelectionListener(new SelectionAdapter() {
              public void widgetSelected(SelectionEvent e) {
                PrintWriter out = top.getConsole();
                ClassicalProblem task1 = task0.classicalProblem();
                task1.print(out);
              }
            });
          }
        }
        if (task instanceof ProofNavigatorTask)
        {
          final ProofNavigatorTask task0 = (ProofNavigatorTask)task;
          {
            MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
            menuItem.setText("Print ProofNavigator Proving Problem");
            menuItem.setImage(top.getAnalyzeImage());
            menuItem.addSelectionListener(new SelectionAdapter() {
              public void widgetSelected(SelectionEvent e) {
                PrintWriter out = top.getConsole();
                ProofNavigatorProblem task1 = task0.proofNavigatorProblem();
                task1.print(out);
              }
            });
          }
        final Object info = task.getInfo();
        if (info != null)
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText("Show Proof");
          menuItem.setImage(top.getAnalyzeImage());
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              Main.displayInfo(task);
            }
          });
        }
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText("Reset Task");
          menuItem.setImage(top.getAnalyzeImage());
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              task0.reset();
            }
          });
        }
      }
      }
    });
  }
  
  /****************************************************************************
   * Redraw all not done tasks in current folder.
   * @param folder.
   ***************************************************************************/
  private void redrawFolder(TaskFolder folder)
  {
    Task[] tasks = folder.getTasks();
    for (Task task : tasks)
    {
      if (task.isOptional()) continue;
      Task.Status status = task.getStatus();
      if (status != Task.Status.DONE && status != Task.Status.ALMOST) 
        createNode(task);
    }
    
    TaskFolder[] folders = folder.getFolders();
    for (TaskFolder subfolder : folders)
      redrawFolder(subfolder);
  }
  
  /****************************************************************************
   * Create tree node for given task.
   * @param task the task.
   ***************************************************************************/
  private void createNode(Task task)
  {
    TreeItem node = new TreeItem(tree, SWT.NULL);
    node.setText(task.getName());
    itemSet.put(node, task);
    top.redraw(node, task, true);
  }
}
