// ----------------------------------------------------------------------------
// TaskTree.java
// A tree displaying verification tasks.
// $Id: TaskTree.java,v 1.28 2017/10/24 08:28:35 schreine Exp schreine $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.SWT;

import java.util.*;
import java.io.*;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.layout.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Tasks.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Proving.*;
import fmrisc.ProgramExplorer.Semantics.*;

/**
 * 
 * A tree displaying verification tasks.
 *
 */
public final class TaskTree
{
  // the tree (subclassing not allowed)
  private Tree tree;
  
  // the top window
  private TopWindow top;
  
  // all items in the tree and the associated tasks
  private Map<TreeItem,Task> itemSet;
  private Map<Task,TaskItem> taskSet;
  
  /***************************************************************************
   * A tree entry denoting a task
   **************************************************************************/
  private final class TaskItem
  {
    final TreeItem node; // the tree item
    final Task task;     // the associated task;
    
    /*************************************************************************
     * Create a node linked to parent node.
     * @param parent the parent (object of type Tree or TaskItem)
     * @param task the task represented by the node.
     ************************************************************************/
    public TaskItem(Object parent, Task task)
    {
      // create tree node
      if (parent instanceof Tree)
        node = new TreeItem((Tree)parent, SWT.NULL);
      else if (parent instanceof TaskItem)
        node = new TreeItem(((TaskItem)parent).node, SWT.NULL);
      else
      {
        Main.internalError("Unknown tree node");
        node = null;
      }
      
      // register task
      this.task = task;
      
      // register file item
      itemSet.put(node, task);
      taskSet.put(task, this);
      
      // set text 
      node.setText(task.getName());
      
      // set image
      redraw();
      
      // set observer that is notified when task changes its status
      task.setObserver(new Task.Observer() {
        public void sendNotification(Task task)
        {
          top.getDisplay().syncExec(new Runnable() {
            public void run()
            {              
              // redraw node status
              if (!node.isDisposed()) redraw();
            }
          });
        }
      }
      );
    }
   
    /*************************************************************************
     * Redraw associated tree node.
     ************************************************************************/
    private void redraw()
    {
      // redraw node
      top.redraw(node, task, true);
      
      // change explorer status
      if (!(TaskItem.this.task instanceof TaskFolder))
        top.redrawOpenExplorer();
      
      // move selection on (to show changed status)
      tree.deselect(node);
      /*
      TreeItem parent = node.getParentItem();
      if (parent == null) return;
      TreeItem[] children = parent.getItems();
      int n = children.length;
      int i = parent.indexOf(node);
      if (i < n-1)
        tree.setSelection(children[i+1]);
      else
        tree.setSelection(parent);
      */
      
      // close folder, if done
      // TaskFolder folder = TaskItem.this.task.getParent();
      // if (folder != null) close(folder);
    }
  }
  
  /****************************************************************************
   * Popup information on a task
   ***************************************************************************/
  private final static class TaskTip 
  {
    private Shell shell;
    private Label label;
    
    // only one instance of task tip at a time
    private static TaskTip tip = null;
    
    /**************************************************************************
     * Create popup window displaying information on given task.
     * @param parent the parent window.
     * @param pos the position of the window.
     * @param task the task.
     *************************************************************************/
    public TaskTip(Shell parent, Point pos, Task task)
    {     
      close();
      tip = this;
      
      shell = new Shell(parent, SWT.ON_TOP | SWT.TOOL); 
      GridLayout gridLayout = new GridLayout();
      gridLayout.numColumns = 1;
      gridLayout.marginWidth = 2;
      gridLayout.marginHeight = 2;
      shell.setLayout(gridLayout);
      Color color = shell.getDisplay().getSystemColor(SWT.COLOR_INFO_BACKGROUND);
      Color tcolor = shell.getDisplay().getSystemColor(SWT.COLOR_INFO_FOREGROUND);
      shell.setBackground(color);
      
      // determine text
      StringWriter out = new StringWriter();
      task.print(new PrintWriter(out));
      String text = out.toString();
      text = text.substring(0, text.length()-1); // remove trailing newline
      
      // set label
      label = new Label(shell, SWT.NONE);
      label.setLayoutData(new GridData(GridData.FILL_HORIZONTAL
          | GridData.VERTICAL_ALIGN_CENTER));
      label.setText(text);
      label.setBackground(color);
      label.setForeground(tcolor);
      
      shell.addMouseTrackListener(new MouseTrackAdapter() { 
        public void mouseEnter(MouseEvent e)
        {
          close();
        }
      });
      
      // open window
      shell.pack();
      MainSWT.setLocation(shell, pos.x+10, pos.y+10);
      shell.open();
    }
    
    /*************************************************************************
     * Close tip instance.
     ************************************************************************/
    public static void close()
    {
      if (tip != null) tip.shell.close();
      tip = null;
    }
  }
  
  /***************************************************************************
   * Create a tree displaying task structures.
   * @param parent the parent widget.
   * @param top the top window.
   **************************************************************************/
  public TaskTree(Composite parent, final TopWindow top)
  {    
    // set top window
    this.top = top;
    
    // initialize tree
    tree = new Tree(parent, SWT.SINGLE);

    // initialize node/tasks sets
    initSets();
    
    // set menu
    createMenu();
    
    // set expansion handler
    tree.addTreeListener(new TreeListener() {
      public void treeExpanded(TreeEvent e) {
        TreeItem item = (TreeItem)e.item;
        Task task = itemSet.get(item);
        if (task == null || !(task instanceof TaskFolder)) return;
        expand((TaskFolder)task);
        e.doit = false;
      }

      public void treeCollapsed(TreeEvent e) {
      }
    });
    
    // set click handlers
    tree.addSelectionListener(new SelectionAdapter() {
      
      // item is double clicked
      public void widgetDefaultSelected(SelectionEvent e) 
      { 
        TreeItem item = (TreeItem)e.item;
        Task task = itemSet.get(item);
        if (task == null) return;
        SourcePosition pos = task.getPosition();
        if (pos == null) return;
        top.displaySource(pos, true);
      }
      
      // item is selected by single click
      public void widgetSelected(SelectionEvent e) 
      {
      }
    });
    
    // set hover handler
    final Composite parent0 = parent;
    tree.addMouseTrackListener(new MouseTrackAdapter() {
      public void mouseHover(MouseEvent e) {
        Tree tree = (Tree)e.widget;
        Task task = find(tree.getItems(), e.x, e.y);
        if (task == null) return;
        new TaskTip(parent0.getShell(), tree.toDisplay(e.x, e.y), task);
      }
      public void mouseExit(MouseEvent e) {
        TaskTip.close();
      }
    });
  }
  
  /****************************************************************************
   * Find task associated to item in set of items denoted by mouse position.
   * @param items the set of tree items to be recursively searched.
   * @param x horizontal coordinate of mouse position.
   * @param y vertical coordinate of mouse position.
   * @return the denoted task (or null, if none was found)
   ***************************************************************************/
  private Task find(TreeItem[] items, int x, int y)
  {
    for (TreeItem item : items)
    {
      TreeItem[] items0 = item.getItems();
      Task task = find(items0, x, y);
      if (task != null) return task;
      task = itemSet.get(item);
      if (task == null || task instanceof TaskFolder) continue;
      Rectangle b = item.getBounds();
      if (x < b.x || x > b.x + b.width) continue;
      if (y < b.y || y > b.y + b.height) continue;
      return task;
    }
    return null;
  }
  
  /****************************************************************************
   * Initizialize node/task sets.
   ***************************************************************************/
  private void initSets()
  {
    itemSet = new MapClass<TreeItem,Task>();
    taskSet = new MapClass<Task,TaskItem>();
  }
  
  /***************************************************************************
   * Create menu for tree
   **************************************************************************/
  private void createMenu()
  {
    final Menu menu = new Menu(tree.getShell(), SWT.POP_UP);
    tree.setMenu(menu);
    menu.addListener (SWT.Show, new Listener () {
      public void handleEvent (Event event) {
        MenuItem [] menuItems = menu.getItems ();
        for (int i=0; i<menuItems.length; i++) {
          menuItems [i].dispose ();
        }
        TreeItem [] treeItems = tree.getSelection ();
        if (treeItems.length != 1) return;
        final TreeItem item = treeItems[0];
        final Task task = itemSet.get(item);
        if (!task.isFinal())
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText ("Execute Task");
          menuItem.setImage(top.getRunImage());
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              executeTask(item);
            }
          });
        }
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText ("Print Task");
          menuItem.setImage(top.getAnalyzeImage());
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              task.print(top.getConsole());
            }
          });
        }
        if (task instanceof StateTask)
        {
          final StateTask task0 = (StateTask)task;
          {
            MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
            menuItem.setText("Print State Proving Problem");
            menuItem.setImage(top.getAnalyzeImage());
            menuItem.addSelectionListener(new SelectionAdapter() {
              public void widgetSelected(SelectionEvent e) {
                PrintWriter out = top.getConsole();
                StateProblem task1 = task0.stateProblem();
                task1.print(out);
              }
            });
          }
        }
        if (task instanceof ClassicalTask)
        {
          final ClassicalTask task0 = (ClassicalTask)task;
          {
            MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
            menuItem.setText("Print Classical Proving Problem");
            menuItem.setImage(top.getAnalyzeImage());
            menuItem.addSelectionListener(new SelectionAdapter() {
              public void widgetSelected(SelectionEvent e) {
                PrintWriter out = top.getConsole();
                ClassicalProblem task1 = task0.classicalProblem();
                task1.print(out);
              }
            });
          }
        }
        if (task instanceof ProofNavigatorTask)
        {
          final ProofNavigatorTask task0 = (ProofNavigatorTask)task;
          {
            MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
            menuItem.setText("Print ProofNavigator Proving Problem");
            menuItem.setImage(top.getAnalyzeImage());
            menuItem.addSelectionListener(new SelectionAdapter() {
              public void widgetSelected(SelectionEvent e) {
                PrintWriter out = top.getConsole();
                ProofNavigatorProblem task1 = task0.proofNavigatorProblem();
                task1.print(out);
              }
            });
          }
        final Object info = task.getInfo();
        if (info != null)
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText("Show Proof");
          menuItem.setImage(top.getAnalyzeImage());
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              Main.displayInfo(task);
            }
          });
        }
        {
          MenuItem menuItem = new MenuItem (menu, SWT.PUSH);
          menuItem.setText("Reset Task");
          menuItem.setImage(top.getAnalyzeImage());
          menuItem.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
              task0.reset();
            }
          });
        }
      }
      }
    });
  }
  
  /****************************************************************************
   * Execute a task associated to the tree item.
   * @param node the tree item to which the task is associated.
   ***************************************************************************/
  private void executeTask(TreeItem node)
  {
    final Task task = itemSet.get(node);
    
    // create thread executing the task
    new Thread(
        new Runnable() { 
          public void run() { 
            task.resume(); 
          } 
        }
    ).start();
  }
  
  /****************************************************************************
   * Get visual representation of tree.
   * @return the tree.
   ***************************************************************************/
  public Tree getTree()
  {
    return tree;
  }
  
  /****************************************************************************
   * Redraw task folder structures.
   * @param root the root folder for display.
   ***************************************************************************/
  public void redraw(TaskFolder root)
  {
    tree.removeAll();
    initSets();
    Task[] tasks = root.getTasks();
    for (Task t : tasks)
    {
      redraw(tree, t);
    }
    TaskFolder[] folders = root.getFolders();
    for (TaskFolder f : folders)
    {
      redraw(tree, f);
      expand(f);
    }
  }
  
  /***************************************************************************
   * Redraw task linked into parent component.
   * @param parent the parent component (type Tree or TaskItem).
   * @param task the task to be displayed.
   **************************************************************************/
  private void redraw(Object parent, Task task)
  {
    // display task
    TaskItem node = new TaskItem(parent, task);
    
    // file may be task folder
    if (!(task instanceof TaskFolder)) return;
    TaskFolder folder = (TaskFolder)task;
      
    // display tasks
    Task[] tasks = folder.getTasks();
    for (Task t : tasks)
    {
      redraw(node, t);
    }
    
    // display subfolders
    TaskFolder[] folders = folder.getFolders();
    for (Task t : folders)
    {
      redraw(node, t);
    }
  }

  /****************************************************************************
   * Expand all task folders.
   * @param root the root of the folder hierarchy.
   ***************************************************************************/
  private void expand(TaskFolder root)
  {
    // expand only folders with tasks that are not done
    // if (root.getStatus() == Task.Status.DONE) return;
    TaskItem item = taskSet.get(root);
    if (isEmpty(root))
    {
      item.node.setExpanded(false);
      return;
    }
    item.node.setExpanded(true);
    TaskFolder[] folders = root.getFolders();
    for (TaskFolder f :folders) expand(f);
  }
    
  /***************************************************************************
   * Determine whether task folder has no (or only optional) tasks.
   * @param folder the folder.
   * @return true if the folder and all its subfolders have only optional tasks.
   *************************************************************************/
  private boolean isEmpty(TaskFolder folder)
  {
    for (Task task : folder.getTasks())
    {
      if (!task.isOptional()) return false;
    }
    for (TaskFolder f : folder.getFolders())
    {
      if (!isEmpty(f)) return false;
    }
    return true;
  }
  
  /****************************************************************************
   * Close task folder, if it is done.
   * @param folder the folder
   ***************************************************************************/
  public void close(TaskFolder folder)
  {
    if (folder.getStatus() != Task.Status.DONE) return;
    TaskItem item = taskSet.get(folder);
    if (item == null) return;
    item.node.setExpanded(false);
    TaskFolder parent = folder.getParent();
    if (parent != null) close(parent);
  }
  
  /***************************************************************************
   * Close all task folders.
   **************************************************************************/
  public void hide()
  {
    for (TreeItem item : tree.getItems())
      item.setExpanded(false);
  }
  
  /***************************************************************************
   * Close all task folders that are done.
   **************************************************************************/
  public void close()
  {
    for (TreeItem item : tree.getItems())
      close(item);
  }
  
  /***************************************************************************
   * Close task folder, if it is done.
   * @param item an item denoting a task folder or task.
   **************************************************************************/
  private void close(TreeItem item)
  {
    // close children recursively
    for (TreeItem item0 : item.getItems())
      close(item0);
    
    // close folder, if done
    Task task = itemSet.get(item);
    if (task == null || !(task instanceof TaskFolder)) return;
    TaskFolder folder = (TaskFolder)task;
    if (folder.getStatus() == Task.Status.DONE)
      item.setExpanded(false);
  }
  
  /***************************************************************************
   * Display task folder associated to symbol (if any).
   * @param symbol a symbol whose tasks are to be displayed.
   **************************************************************************/
  public void display(Symbol symbol)
  {
    Set<Map.Entry<Task, TaskItem>> entries = taskSet.entrySet();
    for (Map.Entry<Task, TaskItem> entry : entries)
    {
      Task task = entry.getKey();
      if (!(task instanceof TaskFolder)) continue;
      TaskFolder folder = (TaskFolder)task;
      if (symbol == folder.getSymbol())
      {
        TaskItem item = entry.getValue();
        tree.setSelection(item.node);
        item.node.setExpanded(true);
        return;
      }
    }
  }
}
