// ----------------------------------------------------------------------------
// UsedVariables.java
// Compute the (logic) variables/constants used in an expression.
// $Id: UsedVariables.java,v 1.7 2011/09/30 15:47:30 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is used software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;

import java.util.*;

/**
 * 
 * Compute the (logic) variables/constants used in an expression.
 *
 */
public final class UsedVariables extends ASTVisitorBase
{
  /***************************************************************************
   * Compute used variables of expression.
   * @param exp an expression.
   * @return a collection of the names of the used variables of the expression
   **************************************************************************/
  public static Collection<String> compute(Expression exp)
  {
    UsedVariables bv = new UsedVariables();
    exp.accept(bv);
    return bv.getVariables();
  }
  
  /***************************************************************************
   * Compute used variables of expression.
   * @param exp an expression.
   * @param result a collection to which the names of the used variables 
   *        of the expression are added.
   **************************************************************************/
  public static void add(Expression exp, Collection<String> result)
  {
    UsedVariables bv = new UsedVariables(result);
    exp.accept(bv);
  }
  
  /***************************************************************************
   * Add variables to variable set.
   * @param vars an array of variables
   * @param result a collection to which the given variables are added
   **************************************************************************/
  public static void add(String[] vars, Collection<String> result)
  {
    for (String v : vars)
      result.add(v);
  }
  
  // the collection holding the used variable names computed up to now
  private Collection<String> used = null;
  
  /***************************************************************************
   * Create visitor with empty result collection.
   **************************************************************************/
  private UsedVariables()
  {
    this(new SetClass<String>());
  }
  
  /****************************************************************************
   * Create visitor with denoted result collection.
   * @param result the collection to which the used variables are added.
   ***************************************************************************/
  private UsedVariables(Collection<String> result)
  {
    used = result;
  }
  
  /****************************************************************************
   * Get used variables computed by the expression visitor.
   * @return a collection of the names of the used variables.
   ***************************************************************************/
  private Collection<String> getVariables()
  {
    return used;
  }
  
  /****************************************************************************
   * Process reference.
   * @param ident a reference.
   ***************************************************************************/
  public void visit(Reference ident) 
  { 

    String var = ident.getName().getBase().getString();
    used.add(var);
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param exp a quantified expression.
   ***************************************************************************/
  public void visit(QuantifiedExpression exp)
  {    
    TypedIdentifier[] variables = exp.getVariables();
    int n = variables.length;
    for (int i=0; i<n; i++)
      used.add(variables[i].getName().getString());
    exp.acceptChildren(this);
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param exp a quantified expression.
   ***************************************************************************/
  public void visit(LetExpression exp)
  { 
    Definition[] definitions = exp.getDefinitions();
    int n = definitions.length;
    for (int i=0; i<n; i++)
      used.add(definitions[i].getName().getString());
    exp.acceptChildren(this);
  }
}

