// ----------------------------------------------------------------------------
// ASTCloner.java
// A cloner for abstract syntax trees.
// $Id: ASTCloner.java,v 1.24 2012/10/18 16:12:57 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;

/**
 * 
 * A cloner for abstract syntax trees.
 *
 */
public class ASTCloner extends ASTVisitorBase
{
  /****************************************************************************
   * Create a clone (deep copy) of a tree denoting a logic entity.
   * @param tree the tree to be cloned 
   * @return the clone of the tree 
   ***************************************************************************/
  public static AST clone(AST tree)
  {
    ASTCloner cloner = new ASTCloner();
    AST result = cloner.clone0(tree);
    if (result == null) Main.internalError("error in cloning");
    return result;
  }
  
  /****************************************************************************
   * Create a clone (deep copy) of a tree denoting a logic entity.
   * @param tree the tree to be cloned 
   * @return the clone of the tree 
   ***************************************************************************/
  protected AST clone0(AST tree)
  {
    tree.accept(this);
    
    // clone also type annotation of terms
    if (tree instanceof Term)
    {
      Type type = ((Term)tree).getType();
      if (type != null) ((Term)result).setType(type);
    }
    return result;
  }
  
  /***************************************************************************
   * Create clone of formula.
   * @param tree a formula.
   * @return the clone of the formula.
   **************************************************************************/
  protected Formula cloneFormula(Formula tree)
  {
    return (Formula)clone0(tree);
  }
  
  /****************************************************************************
   * Clone a sequence of trees.
   * @param trees the trees to be cloned.
   * @return the list of clones
   ***************************************************************************/
  protected void clone0(AST[] trees, AST[] results)
  {
    int n = trees.length;
    for (int i=0; i<n; i++)
      results[i] = clone0(trees[i]);
  }
  
  /****************************************************************************
   * Clone a sequence of identifiers.
   * @param trees the trees to be cloned.
   * @return the list of clones
   ***************************************************************************/
  protected Identifier[] clone0(Identifier[] trees)
  {
    Identifier[] results = new Identifier[trees.length];
    clone0(trees, results);
    return results;
  }
    
  /****************************************************************************
   * Clone a sequence of typed identifiers.
   * @param trees the trees to be cloned.
   * @return the list of clones
   ***************************************************************************/
  protected TypedIdentifier[] clone0(TypedIdentifier[] trees)
  {
    TypedIdentifier[] results = new TypedIdentifier[trees.length];
    clone0(trees, results);
    return results;
  }
  
  /****************************************************************************
   * Clone a sequence of terms
   * @param trees the trees to be cloned.
   * @return the list of clones
   ***************************************************************************/
  protected Term[] clone0(Term[] trees)
  {
    Term[] results = new Term[trees.length];
    clone0(trees, results);
    return results;
  }
  
  /****************************************************************************
   * Clone a sequence of names
   * @param trees the trees to be cloned.
   * @return the list of clones
   ***************************************************************************/
  protected Name[] clone0(Name[] trees)
  {
    Name[] results = new Name[trees.length];
    clone0(trees, results);
    return results;
  }
  
  /****************************************************************************
   * Clone a sequence of value definitions
   * @param trees the trees to be cloned.
   * @return the list of clones
   ***************************************************************************/
  protected ValueDefinition[] clone0(ValueDefinition[] trees)
  {
    ValueDefinition[] results = new ValueDefinition[trees.length];
    clone0(trees, results);
    return results;
  }
  
  /****************************************************************************
   * Clone a sequence of valued identifiers
   * @param trees the trees to be cloned.
   * @return the list of clones
   ***************************************************************************/
  protected ValuedIdentifier[] clone0(ValuedIdentifier[] trees)
  {
    ValuedIdentifier[] results = new ValuedIdentifier[trees.length];
    clone0(trees, results);
    return results;
  }
  
  /****************************************************************************
   * Clone a sequence of selectors
   * @param trees the trees to be cloned.
   * @return the list of clones
   ***************************************************************************/
  protected Selector[] clone0(Selector[] trees)
  {
    Selector[] results = new Selector[trees.length];
    clone0(trees, results);
    return results;
  }
  
  /****************************************************************************
   * Clone a sequence of types
   * @param trees the trees to be cloned.
   * @return the list of clones
   ***************************************************************************/
  protected Type[] clone0(Type[] trees)
  {
    Type[] results = new Type[trees.length];
    clone0(trees, results);
    return results;
  }
  
  // --------------------------------------------------------------------------
  //
  // We only implement the cloning of logic entities.
  //
  // --------------------------------------------------------------------------
  
  // the clone result
  protected AST result; 
   
  /***************************************************************************
   * Hook to setResult result.
   * If returns with non-null result, the visit is abandoned.
   * @param tree the tree to be visited.
   **************************************************************************/
  protected void setResult(AST tree)
  {
    result = null;
  }
  
  public void visit(Name tree) 
  { 
    setResult(tree); if (result != null) return;
    result = Name.construct((Identifier[])clone0(tree.idents));
    // return true; 
  }
  
  public void visit(Identifier tree) 
  { 
    setResult(tree); if (result != null) return;
    // clone inherits reference to symbol
    Identifier ident = Identifier.construct(tree.getString());
    ident.setSymbol(tree.getSymbol());
    result = ident;
    // return true; 
  }
  
  public void visit(TypedIdentifier tree) 
  { 
    setResult(tree); if (result != null) return;
    result = TypedIdentifier.construct(
        (Identifier)clone0(tree.getName()),
        (Type)clone0(tree.getType()));
    // return true; 
  }
  
  public void visit(ValuedIdentifier tree) 
  { 
    setResult(tree); if (result != null) return;
    result = ValuedIdentifier.construct(
        (Identifier)clone0(tree.getIdentifier()),
        (Term)clone0(tree.getValue()));
    // return true; 
  }
  
  public void visit(ValueDefinitionClass tree) 
  { 
    setResult(tree); if (result != null) return;
    result = ValueDefinitionClass.construct(
        (Identifier)clone0(tree.getName()),
        (Term)clone0(tree.getValue()));
    // return true; 
  }
  
  public void visit(FormulaDefinition tree)
  {
    setResult(tree); if (result != null) return;
    result = FormulaDefinition.construct(
        (Identifier)clone0(tree.getName()),
        cloneFormula(tree.getValue()));  
    // return true;
  }
  
  public void visit(AxiomDefinition tree) 
  { 
    setResult(tree); if (result != null) return;
    result = AxiomDefinition.construct(
        (Identifier)clone0(tree.getName()),
        cloneFormula(tree.getValue()));  
    // return true; 
  }
   
  public void visit(TypeDeclaration tree) 
  { 
    setResult(tree); if (result != null) return;
    result = TypeDeclaration.construct(
        (Identifier)clone0(tree.getName())); 
    // return true; 
  }
  
  public void visit(TypeDefinition tree) 
  { 
    setResult(tree); if (result != null) return;
    result = TypeDefinition.construct(
        (Identifier)clone0(tree.getName()),
        (Type)clone0(tree.getValue())); 
    // return true; 
  }
  
  public void visit(ValueDeclarationClass tree)
  {
    setResult(tree); if (result != null) return;
    result = ValueDeclarationClass.construct(
        (Identifier)clone0(tree.getName()),
        (Type)clone0(tree.getType())); 
    // return true;
  }
  
  public void visit(ValueDeclarationInitialized tree) 
  { 
    setResult(tree); if (result != null) return;
    result = ValueDeclarationInitialized.construct(
        (Identifier)clone0(tree.getName()),
        (Type)clone0(tree.getType()),
        (Term)clone0(tree.getValue())); 
    // return true; 
  }
  
  public void visit(Reference tree) 
  { 
    setResult(tree); if (result != null) return;
    result = Reference.construct((Name)clone0(tree.getName()));
    // return true; 
  }
  
  public void visit(AtomicFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = AtomicFormula.construct(
        (Term)clone0(tree.getFunction()), 
        (Term[])clone0(tree.getArguments()));
    // return true; 
  }
  
  public void visit(FunctionApplication tree) 
  { 
    setResult(tree); if (result != null) return;
    result = FunctionApplication.construct(        
        (Term)clone0(tree.getFunction()), 
        (Term[])clone0(tree.getArguments()));
    // return true; 
  }
  
  public void visit(EqualsFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = EqualsFormula.construct(
        (Term)clone0(tree.getFirst()),
        (Term)clone0(tree.getSecond()));
    // return true; 
  }
  public void visit(GreaterEqualFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = GreaterEqualFormula.construct(
        (Term)clone0(tree.getFirst()),
        (Term)clone0(tree.getSecond()));
    // return true; 
  }
  
  public void visit(GreaterFormula tree) 
  {
    setResult(tree); if (result != null) return;
    result = GreaterFormula.construct(
        (Term)clone0(tree.getFirst()),
        (Term)clone0(tree.getSecond()));
    // return true; 
  }
  
  public void visit(LessEqualFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = LessEqualFormula.construct(
        (Term)clone0(tree.getFirst()),
        (Term)clone0(tree.getSecond()));
    // return true; 
  }
  
  public void visit(LessFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = LessFormula.construct(
        (Term)clone0(tree.getFirst()),
        (Term)clone0(tree.getSecond()));
    // return true; 
  }
  
  public void visit(NotEqualsFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = NotEqualsFormula.construct(
        (Term)clone0(tree.getFirst()),
        (Term)clone0(tree.getSecond()));
    // return true; 
  }
  
  public void visit(SimilarFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = SimilarFormula.construct(
        (Term)clone0(tree.getFirst()),
        (Term)clone0(tree.getSecond()));
    // return true; 
  }
  
  public void visit(AndFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = AndFormula.construct(
        cloneFormula(tree.getFirst()),
        cloneFormula(tree.getSecond()));
    // return true; 
  }
  
  public void visit(EquivalentFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = EquivalentFormula.construct(
        cloneFormula(tree.getFirst()),
        cloneFormula(tree.getSecond()));
    // return true; 
  }
  
  public void visit(ImpliesFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = ImpliesFormula.construct(
        cloneFormula(tree.getFirst()),
        cloneFormula(tree.getSecond()));
    // return true; 
  }
  
  public void visit(NotEquivalentFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = NotEquivalentFormula.construct(
        cloneFormula(tree.getFirst()),
        cloneFormula(tree.getSecond()));
    // return true; 
  }
  
  public void visit(OrFormula tree)
  {     
    setResult(tree); if (result != null) return;
    result = OrFormula.construct(
      cloneFormula(tree.getFirst()),
      cloneFormula(tree.getSecond()));
    // return true; 
  }
  
  public void visit(DividesTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = DividesTerm.construct(
        (Term)clone0(tree.getFirst()),
        (Term)clone0(tree.getSecond()));
    // return true; 
  }
  
  public void visit(MinusTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = MinusTerm.construct(
        (Term)clone0(tree.getFirst()),
        (Term)clone0(tree.getSecond()));
    // return true; 
  }
  
  public void visit(PlusTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = PlusTerm.construct(
        (Term)clone0(tree.getFirst()),
        (Term)clone0(tree.getSecond()));
    // return true; 
  }
  
  public void visit(BitOrTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = BitOrTerm.construct(
        (Term)clone0(tree.getFirst()),
        (Term)clone0(tree.getSecond()));
    // return true; 
  }
  
  public void visit(TimesTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = TimesTerm.construct(
        (Term)clone0(tree.getFirst()),
        (Term)clone0(tree.getSecond()));
    // return true; 
  }
  
  public void visit(BitAndTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = BitAndTerm.construct(
        (Term)clone0(tree.getFirst()),
        (Term)clone0(tree.getSecond()));
    // return true; 
  }
  
  public void visit(IntLiteral tree) 
  { 
    setResult(tree); if (result != null) return;
    result = IntLiteral.construct(tree.getName());
    // return true; 
  }
  
  public void visit(True tree) 
  { 
    setResult(tree); if (result != null) return;
    result = True.construct();
    // return true; 
  }
  
  public void visit(BitTrue tree) 
  { 
    setResult(tree); if (result != null) return;
    result = BitTrue.construct();
    // return true; 
  }
  
  public void visit(False tree) 
  { 
    setResult(tree); if (result != null) return;
    result = False.construct();
    // return true; 
  }
  
  public void visit(BitFalse tree) 
  { 
    setResult(tree); if (result != null) return;
    result = BitFalse.construct();
    // return true; 
  }
  
  public void visit(OldState tree) 
  { 
    setResult(tree); if (result != null) return;
    result = OldState.construct();
    // return true; 
  }
  
  public void visit(NewState tree) 
  { 
    setResult(tree); if (result != null) return;
    result = NewState.construct();
    // return true; 
  }
  
  public void visit(StringLiteral tree) 
  { 
    setResult(tree); if (result != null) return;
    result = StringLiteral.construct(tree.getName());
    // return true; 
  }
  
  public void visit(WritesOnlyFormula tree)
  {
    setResult(tree); if (result != null) return;
    result = WritesOnlyFormula.construct((Name[])clone0(tree.getVariables()));
    // return true;
  }
  
  public void visit(ReadsOnlyFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = ReadsOnlyFormula.construct();
    // return true; 
  }
  
  public void visit(IfThenElseFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = IfThenElseFormula.construct(
        cloneFormula(tree.getCondition()),
        cloneFormula(tree.getThenBranch()),
        cloneFormula(tree.getElseBranch()));
    // return true; 
  }
  
  public void visit(IfThenElseTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = IfThenElseTerm.construct(
        cloneFormula(tree.getCondition()),
        (Term)clone0(tree.getThenBranch()),
        (Term)clone0(tree.getElseBranch()));
    // return true; 
  }
  
  public void visit(LetFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = LetFormula.construct(
        (ValueDefinition[])clone0(tree.getDefinitions()),
        cloneFormula(tree.getBody()));
    // return true; 
  }
  
  public void visit(LetTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = LetTerm.construct(
        (ValueDefinition[])clone0(tree.getDefinitions()),
        (Term)clone0(tree.getBody()));
    // return true; 
  }
  
  public void visit(StateBreaks tree) 
  { 
    setResult(tree); if (result != null) return;
    result = StateBreaks.construct((Term)clone0(tree.getArgument()));
    // return true; 
  }
  
  public void visit(StateContinues tree) 
  { 
    setResult(tree); if (result != null) return;
    result = StateContinues.construct((Term)clone0(tree.getArgument()));
    // return true; 
  }
  
  public void visit(StateExecutes tree) 
  { 
    setResult(tree); if (result != null) return;
    result = StateExecutes.construct((Term)clone0(tree.getArgument()));
    // return true; 
  }
  
  public void visit(StateReturns tree) 
  { 
    setResult(tree); if (result != null) return;
    result = StateReturns.construct((Term)clone0(tree.getArgument()));
    // return true; 
  }
  
  public void visit(StateThrows tree) 
  { 
    setResult(tree); if (result != null) return;
    result = StateThrows.construct((Term)clone0(tree.getArgument()));
    // return true; 
  }
  
  public void visit(StateThrowsException tree) 
  { 
    setResult(tree); if (result != null) return;
    result = StateThrowsException.construct(
        (Term)clone0(tree.getArgument()),
        (Name)clone0(tree.getException()));
    // return true; 
  }
  
  public void visit(StateValue tree) 
  {
    setResult(tree); if (result != null) return;
    result = StateValue.construct((Term)clone0(tree.getArgument()));
    // return true; 
  }
  
  public void visit(StateMessage tree) 
  {
    setResult(tree); if (result != null) return;
    result = StateMessage.construct((Term)clone0(tree.getArgument()));
    // return true; 
  }
  
  public void visit(ForallFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = ForallFormula.construct(
        (TypedIdentifier[])clone0(tree.getVariables()),
        cloneFormula(tree.getBody()));
    // return true; 
  }
  
  public void visit(ExistsFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = ExistsFormula.construct(
        (TypedIdentifier[])clone0(tree.getVariables()),
        cloneFormula(tree.getBody()));
    // return true; 
  }
  
  public void visit(ArrayTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = ArrayTerm.construct(
        (TypedIdentifier[])clone0(tree.getVariables()),
        (Term)clone0(tree.getBody()));
    // return true; 
  }
  
  public void visit(LambdaTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = LambdaTerm.construct(
        (TypedIdentifier[])clone0(tree.getVariables()),
        (Term)clone0(tree.getBody()));
    // return true; 
  }
   
  public void visit(LambdaFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = LambdaFormula.construct(
        (TypedIdentifier[])clone0(tree.getVariables()),
        cloneFormula(tree.getBody()));
    // return true; 
  }
  
  public void visit(RecordTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = RecordTerm.construct((ValuedIdentifier[])clone0(tree.getBase()));
    // return true; 
  }
  
  public void visit(SelectionTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = SelectionTerm.construct(
        (Term)clone0(tree.getBase()),
        (Selector)clone0(tree.getSelector()));
    // return true; 
  }
  
  public void visit(TupleTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = TupleTerm.construct((Term[])clone0(tree.getBase()));
    // return true; 
  }
  
  public void visit(UpdateTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = UpdateTerm.construct(
      (Term)clone0(tree.getBase()),
      (Selector[])clone0(tree.getSelectors()),
      (Term)clone0(tree.getValue()));
    // return true;
  }
  
  public void visit(PostVariable tree) 
  { 
    setResult(tree); if (result != null) return;
    result = PostVariable.construct((Name)clone0(tree.getName()));
    // return true; 
  }
  
  public void visit(PreVariable tree) 
  { 
    setResult(tree); if (result != null) return;
    result = PreVariable.construct((Name)clone0(tree.getName()));
    // return true; 
  }
  
  public void visit(ArrayType tree) 
  { 
    setResult(tree); if (result != null) return;
    result = ArrayType.construct(
        (Type)clone0(tree.getIndex()),
        (Type)clone0(tree.getBase()));
    // return true; 
  }
  
  public void visit(FunctionType tree) 
  { 
    setResult(tree); if (result != null) return;
    result = FunctionType.construct(
        (Type[])clone0(tree.getDomain()),
        (Type)clone0(tree.getRange()));
    // return true; 
  }
  
  public void visit(RecordType tree) 
  { 
    setResult(tree); if (result != null) return;
    result = RecordType.construct((TypedIdentifier[])clone0(tree.getBase()));
    // return true; 
  }
  
  public void visit(StateType tree) 
  { 
    setResult(tree); if (result != null) return;
    Type type0 = tree.getReturn();
    if (type0 == null)
      result = StateType.construct(null);
    else
      result = StateType.construct((Type)clone0(tree.getReturn()));
    // return true; 
  }
  
  public void visit(SubrangeType tree) 
  { 
    setResult(tree); if (result != null) return;
    result = SubrangeType.construct(
        (Term)clone0(tree.getFrom()),
        (Term)clone0(tree.getTo()));
    // return true; 
  }
  
  public void visit(SubType tree) 
  { 
    setResult(tree); if (result != null) return;
    result = SubType.construct((Term)clone0(tree.getPredicate()));
    // return true; 
  }
  
  public void visit(TupleType tree) 
  { 
    setResult(tree); if (result != null) return;
    result = TupleType.construct((Type[])clone0(tree.getBase()));
    // return true; 
  }
  
  public void visit(AtomicType tree) 
  {
    setResult(tree); if (result != null) return;
    result = AtomicType.construct(tree.getName());
    // return true;
  }
  
  public void visit(BooleanType tree) 
  { 
    setResult(tree); if (result != null) return;
    result = BooleanType.construct();
    // return true; 
  }
  
  public void visit(BitType tree) 
  { 
    setResult(tree); if (result != null) return;
    result = BitType.construct();
    // return true; 
  }
  
  public void visit(IntType tree) 
  { 
    setResult(tree); if (result != null) return;
    result = IntType.construct();
    // return true; 
  }
  
  public void visit(NatType tree) 
  { 
    setResult(tree); if (result != null) return;
    result = NatType.construct();
    // return true; 
  }
  
  public void visit(RealType tree) 
  { 
    setResult(tree); if (result != null) return;
    result = RealType.construct();
    // return true; 
  }
  
  public void visit(Real0Type tree) 
  { 
    setResult(tree); if (result != null) return;
    result = RealType.construct();
    // return true; 
  }
  
  public void visit(StringType tree) 
  { 
    setResult(tree); if (result != null) return;
    result = StringType.construct();
    // return true; 
  }
  
  public void visit(NotFormula tree) 
  { 
    setResult(tree); if (result != null) return;
    result = NotFormula.construct(cloneFormula(tree.getBase()));
    // return true; 
  }
  
  public void visit(NegationTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = NegationTerm.construct((Term)clone0(tree.getBase()));
    // return true; 
  }
  
  public void visit(BitNegationTerm tree) 
  { 
    setResult(tree); if (result != null) return;
    result = BitNegationTerm.construct((Term)clone0(tree.getBase()));
    // return true; 
  }
  
  public void visit(SelectorIdentifier tree) 
  { 
    setResult(tree); if (result != null) return;
    result = SelectorIdentifier.construct((Identifier)clone0(tree.getIdentifier()));
    // return true; 
  }
  
  public void visit(SelectorIndex tree) 
  { 
    setResult(tree); if (result != null) return;
    result = SelectorIndex.construct((Term)clone0(tree.getIndex()));
    // return true; 
  }
  
  public void visit(SelectorNumber tree) 
  { 
    setResult(tree); if (result != null) return;
    result = SelectorNumber.construct((IntLiteral)clone0(tree.getNumber()));
    // return true; 
  }
}
