// ----------------------------------------------------------------------------
// ASTVisitorBase.java
// Base class of abstract syntax tree visitors
// $Id: ASTVisitorBase.java,v 1.40 2012/10/18 16:12:58 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General protected License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General protected License for more details.
//
// You should have received a copy of the GNU General protected License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax;

import fmrisc.ProgramExplorer.Syntax.Program.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.Judgements.*;

/**
 * Base class of abstract syntax tree visitors.
 */
public abstract class ASTVisitorBase implements ASTVisitor
{  
  // --------------------------------------------------------------------------
  // Note: we implement the dispatch of visit() by sequences of "instanceof"
  // tests. Alternatively, we might speedup dispatches by proceeding as follows:
  // 1) for every AST type T, we hash T.class to a unique numerical id.
  // 2) for a given tree, we lookup the id of tree.getClass() and use
  //    switch(id) to dispatch the appropriate visit() method.
  // This strategy would be applicable in those cases where the dispatch
  // is performed on the dynamic (actual) type of the object (rather than
  // on any supertype or interface type).
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Visit an abstract syntax tree.
   * @param tree the node of the abstract syntax tree that is visited.
   ***************************************************************************/
  public void visit(AST tree)
  { 
    // default implementation traverses children
    tree.acceptChildren(this);
  }
  
  public void visit(Name tree) { visit((AST)tree); }
  public void visit(StatJudgement tree) { visit((AST)tree); }
  
  // --------------------------------------------------------------------------
  //
  // ASTSourceBase
  //
  // --------------------------------------------------------------------------
  
  public void visit(ASTSourceBase tree)
  {
    visit((ASTBase)tree);
  }
  
  public void visit(Handler tree) { visit((ASTSourceBase)tree); }
  public void visit(Identifier tree) { visit((ASTSourceBase)tree); }
  
  public void visit(fmrisc.ProgramExplorer.Syntax.Program.DeclarationBase tree)
  {
    visit((ASTSourceBase)tree);
  }

  public void visit(CompilationUnitBase tree) 
  { 
    visit((fmrisc.ProgramExplorer.Syntax.Program.DeclarationBase)tree);
  }
  
  public void visit(ClassDeclaration tree) { visit((CompilationUnitBase)tree); }
  public void visit(TheoryDeclaration tree) { visit((CompilationUnitBase)tree); }
  
  public void visit(TopDeclarationBase tree) 
  { 
    visit((fmrisc.ProgramExplorer.Syntax.Program.DeclarationBase)tree);
  }
  
  public void visit(GlobalVariableDeclaration tree) 
  { 
    visit((TopDeclarationBase)tree);
  }
  
  public void visit(ClassVariableDeclaration tree) 
  { visit((GlobalVariableDeclaration)tree);}
  public void visit(ObjectVariableDeclaration tree) 
  { visit((GlobalVariableDeclaration)tree);}
  
  public void visit(ParamDeclarationBase tree) 
  { 
    visit((TopDeclarationBase)tree);
  }
  
  public void visit(ConstructorDeclaration tree) { visit((ParamDeclarationBase)tree); }
  
  public void visit(MethodDeclaration tree) 
  { 
    visit((ParamDeclarationBase)tree);
  }
  
  public void visit(ClassMethodDeclaration tree) { visit((MethodDeclaration)tree); }
  public void visit(ObjectMethodDeclaration tree) { visit((MethodDeclaration)tree); }
  
  public void visit(VariableDeclarationBase tree) 
  { 
    visit((fmrisc.ProgramExplorer.Syntax.Program.DeclarationBase)tree);
  }
  
  public void visit(Parameter tree) { visit((VariableDeclarationBase)tree); }
  public void visit(VariableStatement tree) { visit((VariableDeclarationBase)tree); }
  
  public void visit(StatementBase tree) 
  { 
    visit((ASTSourceBase)tree);
  }
  
  public void visit(AssertionStatement tree) { visit((StatementBase)tree); }
  public void visit(AssignmentStatement tree) { visit((StatementBase)tree); }
  public void visit(BlockStatement tree) { visit((StatementBase)tree); }
  public void visit(BreakStatement tree) { visit((StatementBase)tree); }
  public void visit(ConditionalStatement tree) { visit((StatementBase)tree); }
  public void visit(ContinueStatement tree) { visit((StatementBase)tree); }
  public void visit(EmptyStatement tree) { visit((StatementBase)tree); }
  public void visit(ReturnStatement tree) { visit((StatementBase)tree); }
  public void visit(TryCatchStatement tree) { visit((StatementBase)tree); }
  
  public void visit(LoopStatementBase tree) 
  { 
    visit((StatementBase)tree);
  }
  
  public void visit(ForLoopStatement tree) { visit((LoopStatementBase)tree); }
  public void visit(WhileLoopStatement tree) { visit((LoopStatementBase)tree); }
  
  public void visit(CallStatement tree) 
  { 
    visit((StatementBase)tree);
  }
  
  public void visit(AssignCallStatement tree) 
  { 
    visit((CallStatement)tree);
  }
  public void visit(AssignNewStatement tree) { visit((AssignCallStatement)tree); }
  
  public void visit(DeclCallStatement tree) 
  { 
    visit((CallStatement)tree);
  }
  public void visit(DeclNewStatement tree)  { visit((DeclCallStatement)tree); }
  
  public void visit(ThrowStatement tree) { visit((CallStatement)tree); }
  
  public void visit(VoidCallStatement tree) { visit((CallStatement)tree); }  
  
  // --------------------------------------------------------------------------
  //
  // ASTProgramBase
  //
  // --------------------------------------------------------------------------

  public void visit(ASTProgramBase tree)
  {
    visit((ASTBase)tree);
  }
  
  public void visit(fmrisc.ProgramExplorer.Syntax.Program.ArrayType tree) 
  { visit((ASTProgramBase)tree); }
  public void visit(ASTProgramAnnotatedBase tree) 
  { visit((ASTProgramBase)tree); }
  public void visit(LeftValue tree) 
  { visit((ASTProgramBase)tree); }
  public void visit(NamedType tree) 
  { visit((ASTProgramBase)tree); }
  public void visit(NewType tree) 
  { visit((ASTProgramBase)tree); }
  
  public void visit(ImportBase tree)
  {
    visit((ASTProgramBase)tree);
  }
  
  public void visit(PackageImport tree) { visit((ImportBase)tree); }
  public void visit(ClassImport tree) { visit((ImportBase)tree); }
  
  public void visit(PrimitiveType tree) 
  { 
    visit((ASTProgramBase)tree);
  }
  
  public void visit(fmrisc.ProgramExplorer.Syntax.Program.BooleanType tree) 
  { visit((PrimitiveType)tree); }
  public void visit(CharType tree) 
  { visit((PrimitiveType)tree); }
  public void visit(fmrisc.ProgramExplorer.Syntax.Program.IntType tree) 
  { visit((PrimitiveType)tree); }
  public void visit(NullType tree) 
  { visit((PrimitiveType)tree); }
  public void visit(fmrisc.ProgramExplorer.Syntax.Program.StringType tree) 
  { visit((PrimitiveType)tree); }
  public void visit(VoidType tree) 
  { visit((PrimitiveType)tree); }
  public void visit(ExceptionType tree) 
  { visit((PrimitiveType)tree); }
  
  public void visit(ProgramAnnotationBase tree) 
  { 
    visit((ASTProgramBase)tree);
  }
  
  public void visit(SpecificationBase tree) 
  { 
    visit((ProgramAnnotationBase)tree);
  }
  
  public void visit(MethodSpec tree) { visit((SpecificationBase)tree); }
  public void visit(UnitSpec tree) { visit((SpecificationBase)tree); }
  public void visit(ClassSpec tree) { visit((SpecificationBase)tree); }
  public void visit(LoopSpec tree) { visit((SpecificationBase)tree); }
  public void visit(StatementSpec tree) { visit((SpecificationBase)tree); }
  
  public void visit(fmrisc.ProgramExplorer.Syntax.Program.SelectorBase tree) 
  { visit((ASTProgramBase)tree); }
  public void visit(IdentifierSelector tree) 
  { visit((fmrisc.ProgramExplorer.Syntax.Program.SelectorBase)tree); }
  public void visit(IndexSelector tree) 
  { visit((fmrisc.ProgramExplorer.Syntax.Program.SelectorBase)tree); }
  public void visit(MessageSelector tree) 
  { visit((fmrisc.ProgramExplorer.Syntax.Program.SelectorBase)tree); }
  
  public void visit(ValueExpressionBase tree) 
  { 
    visit((ASTProgramBase)tree);
  }
  
  public void visit(fmrisc.ProgramExplorer.Syntax.Program.BooleanLiteral tree) 
  { visit((ValueExpressionBase)tree); }
  public void visit(CharLiteral tree) 
  { visit((ValueExpressionBase)tree); }
  public void visit(fmrisc.ProgramExplorer.Syntax.Program.IntLiteral tree) 
  { visit((ValueExpressionBase)tree); }
  public void visit(NewArrayExpression tree) 
  { visit((ValueExpressionBase)tree); }
  public void visit(NullLiteral tree) 
  { visit((ValueExpressionBase)tree); }
  public void visit(SelectorExpression tree)
  { visit((ValueExpressionBase)tree); }
  public void visit(fmrisc.ProgramExplorer.Syntax.Program.StringLiteral tree) 
  { visit((ValueExpressionBase)tree); }
  public void visit(VariableValue tree) 
  { visit((ValueExpressionBase)tree); }

  public void visit(BinaryValueExpression tree) 
  { 
    visit((ValueExpressionBase)tree);
  }
  
  public void visit(AndExpression tree) { visit((BinaryValueExpression)tree); }
  public void visit(DividesExpression tree) { visit((BinaryValueExpression)tree); }
  public void visit(EqualsExpression tree) { visit((BinaryValueExpression)tree); }
  public void visit(GreaterEqualExpression tree) { visit((BinaryValueExpression)tree); }
  public void visit(GreaterExpression tree) { visit((BinaryValueExpression)tree); }
  public void visit(LessEqualExpression tree) { visit((BinaryValueExpression)tree); }
  public void visit(LessExpression tree) { visit((BinaryValueExpression)tree); }
  public void visit(MinusExpression tree) { visit((BinaryValueExpression)tree); }
  public void visit(NotEqualsExpression tree) { visit((BinaryValueExpression)tree); }
  public void visit(OrExpression tree) { visit((BinaryValueExpression)tree); }
  public void visit(PercentExpression tree) { visit((BinaryValueExpression)tree); }
  public void visit(PlusExpression tree) { visit((BinaryValueExpression)tree); }
  public void visit(TimesExpression tree) { visit((BinaryValueExpression)tree); }
  
  public void visit(UnaryValueExpression tree) 
  { 
    visit((ValueExpressionBase)tree);
  }
  
  public void visit(NegationExpression tree) { visit((UnaryValueExpression)tree); }
  public void visit(NotExpression tree) { visit((UnaryValueExpression)tree); }
  
  // --------------------------------------------------------------------------
  //
  // ASTLogicBase
  //
  // --------------------------------------------------------------------------
  
  public void visit(ASTLogicBase tree)
  {
    visit((ASTBase)tree);
  }
  
  public void visit(TypedIdentifier tree) { visit((ASTLogicBase)tree); }
  public void visit(ValuedIdentifier tree) { visit((ASTLogicBase)tree); }
  
  public void visit(fmrisc.ProgramExplorer.Syntax.Logic.DeclarationBase tree)
  {
    visit((ASTLogicBase)tree);
  }
  
  public void visit(DefinitionBase tree)
  {
    visit((fmrisc.ProgramExplorer.Syntax.Logic.DeclarationBase)tree);
  }
  
  public void visit(ValueDefinitionClass tree) { visit((DefinitionBase)tree); }
  
  public void visit(FormulaDefinition tree)
  {
    visit((DefinitionBase)tree);
  }
  
  public void visit(AxiomDefinition tree) { visit((FormulaDefinition)tree); }
  
  public void visit(TypeDeclaration tree)
  {
    visit((fmrisc.ProgramExplorer.Syntax.Logic.DeclarationBase)tree);
  }
  
  public void visit(TypeDefinition tree) { visit((TypeDeclaration)tree); }
  
  public void visit(ValueDeclarationClass tree)
  {
    visit((fmrisc.ProgramExplorer.Syntax.Logic.DeclarationBase)tree);
  }
  
  public void visit(ValueDeclarationFormula tree) 
  { visit((ValueDeclarationClass)tree); }
  public void visit(ValueDeclarationInitialized tree) 
  { visit((ValueDeclarationClass)tree); }
  
  public void visit(ExpressionBase tree)
  {
    visit((ASTLogicBase)tree);
  }
  
  public void visit(Reference tree) { visit((ExpressionBase)tree); }
  
  public void visit(ApplicationExpression tree)
  {
    visit((ExpressionBase)tree);
  }
  
  public void visit(TermApplication tree)
  {
    visit((ApplicationExpression)tree);
  }
  
  public void visit(AtomicFormula tree) { visit((TermApplication)tree); }
  public void visit(FunctionApplication tree) { visit((TermApplication)tree); }
  
  public void visit(BinaryExpression tree)
  {
    visit((ExpressionBase)tree);
  }
  
  public void visit(BinaryAtomicFormula tree)
  {
    visit((BinaryExpression)tree);
  }
  
  public void visit(EqualsFormula tree) { visit((BinaryAtomicFormula)tree); }
  public void visit(GreaterEqualFormula tree) { visit((BinaryAtomicFormula)tree); }
  public void visit(GreaterFormula tree) { visit((BinaryAtomicFormula)tree); }
  public void visit(LessEqualFormula tree) { visit((BinaryAtomicFormula)tree); }
  public void visit(LessFormula tree) { visit((BinaryAtomicFormula)tree); }
  public void visit(NotEqualsFormula tree) { visit((BinaryAtomicFormula)tree); }
  public void visit(SimilarFormula tree) { visit((BinaryAtomicFormula)tree); }
  
  public void visit(BinaryFormula tree)
  {
    visit((BinaryExpression)tree);
  }
  
  public void visit(AndFormula tree) { visit((BinaryFormula)tree); }
  public void visit(EquivalentFormula tree) { visit((BinaryFormula)tree); }
  public void visit(ImpliesFormula tree) { visit((BinaryFormula)tree); }
  public void visit(NotEquivalentFormula tree) { visit((BinaryFormula)tree); }
  public void visit(OrFormula tree) { visit((BinaryFormula)tree); }
  
  public void visit(BinaryTerm tree)
  {
    visit((BinaryExpression)tree);
  }
  
  public void visit(DividesTerm tree) { visit((BinaryTerm)tree); }
  public void visit(MinusTerm tree) { visit((BinaryTerm)tree); }
  public void visit(PlusTerm tree) { visit((BinaryTerm)tree); }
  public void visit(TimesTerm tree) { visit((BinaryTerm)tree); }
  public void visit(BitOrTerm tree) { visit((BinaryTerm)tree); }
  public void visit(BitAndTerm tree) { visit((BinaryTerm)tree); }
  
  public void visit(Constant tree)
  {
    visit((ExpressionBase)tree);
  }
  
  public void visit(fmrisc.ProgramExplorer.Syntax.Logic.IntLiteral tree) 
  { visit((Constant)tree); }
  public void visit(fmrisc.ProgramExplorer.Syntax.Logic.StringLiteral tree) 
  { visit((Constant)tree); }
  
  public void visit(fmrisc.ProgramExplorer.Syntax.Logic.BooleanLiteral tree) 
  { visit((Constant)tree); }
  public void visit(BitLiteral tree) { visit((Constant)tree); }
  
  public void visit(True tree) 
  { visit((fmrisc.ProgramExplorer.Syntax.Logic.BooleanLiteral)tree); }
  public void visit(False tree) 
  { visit((fmrisc.ProgramExplorer.Syntax.Logic.BooleanLiteral)tree); }
  public void visit(BitTrue tree) { visit((BitLiteral)tree); }
  public void visit(BitFalse tree) { visit((BitLiteral)tree); }
  
  public void visit(StateLiteral tree) 
  { visit((Constant)tree); }
  
  public void visit(OldState tree) { visit((StateLiteral)tree); }
  public void visit(NewState tree) { visit((StateLiteral)tree); }
  
  public void visit(FormulaBase tree)
  {
    visit((ExpressionBase)tree);
  }
  
  public void visit(WritesOnlyFormula tree)
  {
    visit((FormulaBase)tree);
  }
  
  public void visit(ReadsOnlyFormula tree) { visit((WritesOnlyFormula)tree); }
  
  public void visit(IfThenElseExpression tree)
  {
    visit((ExpressionBase)tree);
  }
  
  public void visit(IfThenElseFormula tree) { visit((IfThenElseExpression)tree); }
  public void visit(IfThenElseTerm tree) { visit((IfThenElseExpression)tree); }
  
  public void visit(LetExpression tree)
  {
    visit((ExpressionBase)tree);
  }
  
  public void visit(LetFormula tree) { visit((LetExpression)tree); }
  public void visit(LetTerm tree) { visit((LetExpression)tree); }
  
  public void visit(PostfixApplication tree)
  {
    visit((ExpressionBase)tree);
  }
  
  public void visit(PostfixFormula tree)
  {
    visit((PostfixApplication)tree);
  }
  
  public void visit(StateBreaks tree) { visit((PostfixFormula)tree); }
  public void visit(StateContinues tree) { visit((PostfixFormula)tree); }
  public void visit(StateExecutes tree) { visit((PostfixFormula)tree); }
  public void visit(StateReturns tree) { visit((PostfixFormula)tree); }
  public void visit(StateThrows tree) { visit((PostfixFormula)tree); }
  public void visit(StateThrowsException tree) { visit((PostfixFormula)tree); }
  
  public void visit(PostfixTerm tree)
  {
    visit((PostfixApplication)tree);
  }
  
  public void visit(StateValue tree) { visit((PostfixTerm)tree); }
  public void visit(StateMessage tree) { visit((PostfixTerm)tree); }
  
  public void visit(QuantifiedExpression tree)
  {
    visit((ExpressionBase)tree);
  }
  
  public void visit(QuantifiedFormula tree)
  {
    visit((QuantifiedExpression)tree);
  }
  
  public void visit(ForallFormula tree) { visit((QuantifiedFormula)tree); }
  public void visit(ExistsFormula tree) { visit((QuantifiedFormula)tree); }
  
  public void visit(QuantifiedTerm tree)
  {
    visit((QuantifiedExpression)tree);
  }
  
  public void visit(ArrayTerm tree) { visit((QuantifiedTerm)tree); }
  public void visit(LambdaTerm tree) { visit((QuantifiedTerm)tree); }
  public void visit(LambdaFormula tree) { visit((QuantifiedTerm)tree); }
  
  public void visit(TermBase tree)
  {
    visit((ExpressionBase)tree);
  }
  
  public void visit(RecordTerm tree) { visit((TermBase)tree); }
  public void visit(SelectionTerm tree) { visit((TermBase)tree); }
  public void visit(TupleTerm tree) { visit((TermBase)tree); }
  public void visit(UpdateTerm tree) { visit((TermBase)tree); }
  
  public void visit(ProgramVariableBase tree)
  {
    visit((TermBase)tree);
  }
  
  public void visit(PostVariable tree) { visit((ProgramVariableBase)tree); }
  public void visit(PreVariable tree) { visit((ProgramVariableBase)tree); }
  
  public void visit(TypeBase tree)
  {
    visit((ExpressionBase)tree);
  }
  
  public void visit(fmrisc.ProgramExplorer.Syntax.Logic.ArrayType tree) 
  { visit((TypeBase)tree); }
  public void visit(FunctionType tree) { visit((TypeBase)tree); }
  public void visit(RecordType tree) { visit((TypeBase)tree); }
  public void visit(StateType tree) { visit((TypeBase)tree); }
  public void visit(SubrangeType tree) { visit((TypeBase)tree); }
  public void visit(SubType tree) { visit((TypeBase)tree); }
  public void visit(TupleType tree) { visit((TypeBase)tree); }
  
  public void visit(AtomicType tree) 
  {
    visit((TypeBase)tree);
  }
  
  public void visit(fmrisc.ProgramExplorer.Syntax.Logic.BooleanType tree) 
  { visit((AtomicType)tree); }
  public void visit(BitType tree) 
  { visit((AtomicType)tree); }
  public void visit(fmrisc.ProgramExplorer.Syntax.Logic.IntType tree) 
  { visit((AtomicType)tree); }
  public void visit(NatType tree) 
  { visit((AtomicType)tree); }
  public void visit(RealType tree) 
  { visit((AtomicType)tree); }
  public void visit(Real0Type tree) 
  { visit((AtomicType)tree); }
  public void visit(fmrisc.ProgramExplorer.Syntax.Logic.StringType tree) 
  { visit((AtomicType)tree); }
  
  public void visit(UnaryExpression tree)
  {
    visit((ExpressionBase)tree);
  }
  
  public void visit(UnaryFormula tree)
  {
    visit((UnaryExpression)tree);
  }
  
  public void visit(NotFormula tree) { visit((UnaryFormula)tree); }
  
  public void visit(UnaryTerm tree)
  {
    visit((UnaryExpression)tree);
  }
  
  public void visit(NegationTerm tree) { visit((UnaryTerm)tree); }
  public void visit(BitNegationTerm tree) { visit((UnaryTerm)tree); }
  
  public void visit(fmrisc.ProgramExplorer.Syntax.Logic.SelectorBase tree)
  {
    visit((ASTLogicBase)tree);
  }
  
  public void visit(SelectorIdentifier tree) 
  { visit((fmrisc.ProgramExplorer.Syntax.Logic.SelectorBase)tree); }
  public void visit(SelectorIndex tree) 
  { visit((fmrisc.ProgramExplorer.Syntax.Logic.SelectorBase)tree); }
  public void visit(SelectorNumber tree) 
  { visit((fmrisc.ProgramExplorer.Syntax.Logic.SelectorBase)tree); }
}
