// ----------------------------------------------------------------------------
// Identifier.java
// A simple (unstructured) identifier.
// $Id: Identifier.java,v 1.18 2017/07/04 08:04:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Semantics.*;

/**
 * A simple (unstructured) identifier.
 */
public final class Identifier extends ASTSourceBase
{
  // names are made canonical such that they can be used as keys in hash tables
  private final String name;
  
  // the table of unique names
  private final static Map<String,String> names = 
    new MapClass<String,String>();
  
  // the symbol denoted by the identifier
  private Symbol symbol;
  
  /***************************************************************************
   * Construct identifier.
   * @param name the textual representation of the identifier.
   * @return the identifier.
   **************************************************************************/
  public static Identifier construct(String name)
  {
    return new Identifier(name);
  }
  
  /***************************************************************************
   * Construct identifiers.
   * @param names the textual representations of the identifiers.
   * @return the identifier.
   **************************************************************************/
  public static Identifier[] construct(String[] names)
  {
    int n = names.length;
    Identifier[] idents = new Identifier[n];
    for (int i=0; i<n; i++) 
      idents[i] = new Identifier(names[i]);
    return idents;
  }
  
  /***************************************************************************
   * Construct unique (canonical) version of a string
   * @param name a string
   * @return a canonical representation of the name
   **************************************************************************/
  private static String unique(String name)
  {
    String key = names.get(name);
    if (key == null)
    {
      names.put(name, name);
      key = name;
    }
    return key;
  }
  
  /***************************************************************************
   * Construct identifier.
   * @param name the textual representation of the identifier.
   **************************************************************************/
  private Identifier(String name)
  {
    this.name = unique(name);
    this.symbol = null;
  }
  
  /***************************************************************************
   * Create a new identifier.
   * @param name the (non-empty) base name of the new identifier
   *             (must start with a non-digit character)
   * @param used a collection of names that must not be used.
   * @return an identifier whose name is derived from "name" 
   *         but does not appear in "used"
   **************************************************************************/
  public static Identifier constructNew(String name, Collection<String> used)
  {
    // base itself is suitable
    if (!used.contains(unique(name))) return new Identifier(name);
    
    // search for position of last non-digit
    int n = name.length();
    int i = n-1;
    while (i >= 0)
    {
      char ch = name.charAt(i);
      if (ch < '0' || ch > '9') break;
      i = i-1;
    }
    
    // construct identifier 
    String prefix = name.substring(0, i+1);
    int ext = -1;
    if (i+1 < n) ext = Integer.parseInt(name.substring(i+1));
    while(true)
    {
      ext = ext+1;
      name = prefix+ext;
      if (!used.contains(unique(name))) return new Identifier(name);
    }
  }
  
  /***************************************************************************
   * Get string representation of identifier.
   * @return the string representation.
   **************************************************************************/
  public String getString()
  {
    return name;
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitorOld visitor)
  {
    boolean done = visitor.visit(this);
    if (done) return;
    super.accept(visitor);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitor visitor)
  {
    visitor.visit(this);
  }
  
  /***************************************************************************
   * Set the position of this identifier in the source code.
   * @param position the source code position.
   **************************************************************************/
  public void setPosition(SourcePosition position)
  {
    // also link position to identifier
    super.setPosition(position);
    position.setIdentifier(this);
  }
  
  /***************************************************************************
   * Set symbol denoted by identifier (must not yet be set).
   * @param symbol the symbol.
   **************************************************************************/
  public void setSymbol(Symbol symbol)
  {
    // if (this.symbol != null && this.symbol != symbol)
    //   Main.internalError("identifier has already symbol");
    this.symbol = symbol;
  }
  
  /****************************************************************************
   * Get symbol denoted by identifier.
   * @return the symbol.
   ***************************************************************************/
  public Symbol getSymbol()
  {
    return symbol;
  }
}
