// ----------------------------------------------------------------------------
// ClassDeclaration.java
// A class declaration.
// $Id: ClassDeclaration.java,v 1.6 2017/07/04 08:04:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax.Program;

import java.util.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Spec.*;

/**
 * A class declaration
 */
public final class ClassDeclaration extends CompilationUnitBase
{
  public final TopDeclaration[] decls;
  private ClassSpec classSpec;
  
  /****************************************************************************
   * Construct a class declaration
   * @param pname package name
   * @param imports import statements
   * @param name class identifier
   * @param decls sequence of declarations
   * @return the declaration.
   ***************************************************************************/
  public static ClassDeclaration construct(Name pname, Vector<Import> imports,
    Identifier name, Vector<TopDeclaration> decls)
  {
    return construct(pname,
        imports.toArray(new Import[imports.size()]), 
        name, 
        decls.toArray(new TopDeclaration[decls.size()]));
  }
  
  /****************************************************************************
   * Construct a class declaration
   * @param pname package name
   * @param imports import statements
   * @param name class identifier
   * @param decls sequence of declarations
   * @return the declaration.
   ***************************************************************************/
  public static ClassDeclaration construct(Name pname, Import[] imports,
    Identifier name, TopDeclaration[] decls)
  {
    return new ClassDeclaration(pname, imports, name, decls);
  }
  
  /****************************************************************************
   * Construct a class declaration
   * @param pname package name
   * @param imports import statements
   * @param name class identifier
   * @param decls sequence of declarations
   ***************************************************************************/
  private ClassDeclaration(Name pname, Import[] imports,
    Identifier name, TopDeclaration[] decls)
  {
    super(pname, imports, name);
    this.decls = decls;
    this.classSpec = null;
  }
  
  /***************************************************************************
   * Set class specification.
   * @param spec the specification.
   **************************************************************************/
  public void setClassSpec(ClassSpec spec)
  {
    this.classSpec = spec;
  }
  
  /***************************************************************************
   * Get class specification
   * @return the specification.
   **************************************************************************/
  public ClassSpec getClassSpec()
  {
    return classSpec;
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitorOld visitor)
  {
    boolean done = visitor.visit(this);
    if (done) return;
    super.accept(visitor);
    accept(decls, visitor);
    if (classSpec != null) classSpec.accept(visitor);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitor visitor)
  {
    visitor.visit(this);
  }
  
  /***************************************************************************
   * Accept visitor for a visit to the children of this node.
   * @param visitor the visitor who is accepted by the children of this node.
   **************************************************************************/
  public void acceptChildren(ASTVisitor visitor)
  {
    super.acceptChildren(visitor);
    accept(decls, visitor);
    if (classSpec != null) classSpec.accept(visitor);
  }
}
