// ----------------------------------------------------------------------------
// SelectorExpression.java
// Selection of an object component.
// $Id: SelectorExpression.java,v 1.5 2017/07/04 08:04:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax.Program;

import java.util.*;
import fmrisc.ProgramExplorer.Syntax.*;

/**
 * Selection of an object component
 */
public final class SelectorExpression extends ValueExpressionBase
{
  public static final int precedence = 15;
  
  public final ValueExpression base;
  public final Selector[] sels;
  private TypeExpression btype;
  
  /****************************************************************************
   * Selection of an object component.
   * @param base the object expression.
   * @param sels the selector sequence.
   * @return the expression.
   ***************************************************************************/
  public static SelectorExpression construct
    (ValueExpression base, Vector<Selector> sels)
  {
    return new SelectorExpression(base, 
        sels.toArray(new Selector[sels.size()]));
  }
  
  /****************************************************************************
   * Selection of an object component.
   * @param base the object expression.
   * @param sels the selector sequence.
   * @return the expression.
   ***************************************************************************/
  public static SelectorExpression construct
    (ValueExpression base, Selector[] sels)
  {
    return new SelectorExpression(base, sels);
  }
  
  /****************************************************************************
   * Set type of base expression.
   * @param btype the type of the base expression.
   ***************************************************************************/
  public void setBaseType(TypeExpression btype)
  {
    this.btype = btype;
  }
  
  /****************************************************************************
   * Get type of base expression.
   * @return the type of the base expression.
   ***************************************************************************/
  public TypeExpression getBaseType()
  {
    return btype;
  }
  
  /****************************************************************************
   * Selection of an object component.
   * @param base the object expression.
   * @param sels the selector sequence.
   ***************************************************************************/
  private SelectorExpression(ValueExpression base, Selector[] sels)
  {
    this.base = base;
    this.sels = sels;
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitorOld visitor)
  {
    boolean done = visitor.visit(this);
    if (done) return;
    base.accept(visitor);
    accept(sels, visitor);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitor visitor)
  {
    visitor.visit(this);
  }
  
  /***************************************************************************
   * Accept visitor for a visit to the children of this node.
   * @param visitor the visitor who is accepted by the children of this node.
   **************************************************************************/
  public void acceptChildren(ASTVisitor visitor)
  {
    super.acceptChildren(visitor);
    base.accept(visitor);
    accept(sels, visitor);
  }
  
  /***************************************************************************
   * Get precedence value of the expression operator 
   * (higher precedence value means higher binding power)
   **************************************************************************/
  public int getPrecedence()
  {
    return precedence;
  }
}

