// ----------------------------------------------------------------------------
// CorrectnessTask.java
// The task of proving the partial correctness of a method/constructor.
// $Id: CorrectnessTask.java,v 1.15 2012/10/18 16:12:58 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.io.*;
import java.util.*;

import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Proving.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Judgements.*;

/**
 * 
 * The task of proving the partial correctness of a method/constructor.
 *
 */
public class CorrectnessTask extends StateTaskBase
{
  private ParamSymbol method;
  private StateType stateType;
  private Name[] stateVars;
  private Name[] stateExcs;
  private Vector<TheorySymbol> theories;
  private Declaration[][] decls;
  private ErrorStream out;
  
  /***************************************************************************
   * Construct a partial correctness task for a given method.
   * @param method the symbol of a method with a specification.
   * @param parent the parent of the new directory for this task
   * @param stateType the type of the current state (null, if none)
   * @param stateVars the names of the modifiable variables (null, if none)
   * @param stateExcs the names of the throwable exceptions (null, if none)
   * @param theories the theories on which the following depend in dependence order.
   * @param decls a sequence of declaration lists.
   * @param out the stream to which to write error messages.
   **************************************************************************/
  public CorrectnessTask(ParamSymbol method, File parent,
    StateType stateType, Name[] stateVars, Name[] stateExcs,
    Vector<TheorySymbol> theories, Declaration[][] decls,
    ErrorStream out)
  {
    super("[" + method.getLongName() + "] postcondition", false,
        method.getDeclaration().getPosition(), parent, 
        new ManualStrategy());
    this.method = method;
    this.stateType = stateType;
    this.stateVars = stateVars;
    this.stateExcs = stateExcs;
    this.theories = theories;
    this.decls = decls;
    this.out = out;
  }
  
  /****************************************************************************
   * Print linear representation of task.
   * @param out the stream to write to.
   ***************************************************************************/
  public void print(PrintWriter out)
  {
    super.print(out);
    out.println("Type: verify partial correctness");
  }
     
  /****************************************************************************
   * Translate task into a proving problem in state logic.
   * @return the translation result
   ***************************************************************************/
  public StateProblem stateProblem()
  {
    Formula pre = Normalization.preCondition(method, true);
    StatJudgement judgement = method.getBodyJudgement();
    Formula assumption = judgement.sformula;
    if (pre != null) assumption = AndFormula.construct(pre, assumption);
    Formula goal = Normalization.postCondition(method);
    if (goal == null) 
      goal = ExpressionLogic.booleanFormula(true);
    else
      goal = ImpliesFormula.construct(assumption, goal);
    ClassSymbol csymbol = method.getClassSymbol();
    FlattenDeclarations.Result result = 
      FlattenDeclarations.flatten(theories, decls, goal, csymbol.getLocalTheory());
    Type stringType = csymbol.getPackage().getTypeTranslator().getStringType();
    return new StateProblem(result.decls, result.goal,
        result.map, stateType, stringType,
        Normalization.modifiableVariables(stateVars), 
        NameUtils.getClassSymbols(stateExcs), 
        out);    
  }
}
