// ----------------------------------------------------------------------------
// Task.java
// The interface to a verification task.
// $Id: Task.java,v 1.15 2017/07/04 08:04:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.io.*;

import fmrisc.ProgramExplorer.InOut.*;

/**
 * 
 * The interface to a verification task.
 *
 */
public interface Task
{
  /**
   * 
   * A task status.
   *
   */
  // NEW:     task has not yet been attempted
  // OPEN:    task has been already (partially) processed
  // ALMOST:  task is essentially done (but one can do slightly more)
  // DONE:    task is successfully completed
  // FAILED:  task is unsuccessfully completed
  // REVOKED: task is revoked
  public enum Status {NEW, OPEN, ALMOST, DONE, FAILED, REVOKED};
  
  /**
   * 
   * A task observer.
   *
   */
  public static interface Observer 
  {
    /**************************************************************************
     * Notify observer that the status of the denoted task has changed.
     * @param task the task whose status has changed.
     *************************************************************************/
    public void sendNotification(Task task);
  }
  
  /****************************************************************************
   * Get the name of the task.
   * @return the task name.
   ***************************************************************************/
  public String getName();
  
  /****************************************************************************
   * Determine whether new task is to be resumed.
   * @return true if new task is to be resumed.
   ***************************************************************************/
  public boolean resumeOnNew();
  
  /****************************************************************************
   * Determine whether task is optional.
   * @return true if task is optional.
   ***************************************************************************/
  public boolean isOptional();
  
  /****************************************************************************
   * Get status of task.
   * @return the task status.
   ***************************************************************************/
  public Status getStatus();
  
  /****************************************************************************
   * Determine whether task status is final.
   * @return true iff task status is DONE or FAILED.
   ***************************************************************************/
  public boolean isFinal();
  
  /****************************************************************************
   * Attempt to perform task which is in status NEW or OPEN.
   * @return the resulting task status (OPEN, DONE, or FAILED).
   ***************************************************************************/
  public Status resume();
  
  /****************************************************************************
   * Get source code position linked to this task (null, if none)
   * @return the position.
   ***************************************************************************/
  public SourcePosition getPosition();
  
  /****************************************************************************
   * Set folder containing current task.
   * @param parent the parent folder (may be null).
   ***************************************************************************/
  public void setParent(TaskFolder parent);
  
  /****************************************************************************
   * Get folder containing current task.
   * @return the parent folder (may be null).
   ***************************************************************************/
  public TaskFolder getParent();
  
  /****************************************************************************
   * Get directory to be used for writing persistent data
   * @return the directory.
   ***************************************************************************/
  public File getDirectory();
  
  /****************************************************************************
   * Print linear representation of task.
   * @param out the writer.
   ***************************************************************************/
  public void print(PrintWriter out);
  
  /****************************************************************************
   * Set task status.
   * @param status the task status.
   * @param evidence the evidence for the status (may be null).
   * @param info additional information (may be null)
   ***************************************************************************/
  public void setStatus(Status status, String evidence, Object info);
  
  /****************************************************************************
   * Get evidence for task status.
   * @return the evidence (may be null)
   ***************************************************************************/
  public String getEvidence();
  
  /****************************************************************************
   * Get additional information for task status.
   * @return the information (may be null)
   ***************************************************************************/
  public Object getInfo();
  
  /****************************************************************************
   * Set observer for this task.
   * @param observer a task observer
   ***************************************************************************/
  public void setObserver(Observer observer);
}
