// ----------------------------------------------------------------------------
// TerminationTasks.java
// The tasks of proving that a method is well-behaved with respect to termination.
// $Id: TerminationTasks.java,v 1.16 2013/01/22 10:31:34 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.io.*;
import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.Judgements.*;

/**
 * 
 * The tasks of proving that a method is well-behaved with respect to termination.
 *
 */
public final class TerminationTasks
{
  /**
   * 
   * The info necessary for delayed task creation.
   *
   */
  private final static class Info
  {
    public final ParamSymbol method;
    public final StateType stateType;
    public final Name[] stateVars;
    public final Name[] stateExcs;
    public final Vector<TheorySymbol> theories;
    public final Declaration[][] decls;
    public final ErrorStream out;
    
    public Info(ParamSymbol method,
      StateType stateType, Name[] stateVars, Name[] stateExcs,
      Vector<TheorySymbol> theories, Declaration[][] decls,
      ErrorStream out)
    {
      this.method = method;
      this.stateType = stateType;
      this.stateVars = stateVars;
      this.stateExcs = stateExcs;
      this.theories = theories;
      this.decls = decls;
      this.out = out;
    }
  }
  
  // the info registered for delayed task creation
  private static Vector<Info> infos = new Vector<Info>();
  
  /***************************************************************************
   * Register the construction of termination tasks for a given method.
   * @param method the symbol of a method with a specification.
   * @param stateType the type of the current state (null, if none)
   * @param stateVars the names of the modifiable variables (null, if none)
   * @param stateExcs the names of the throwable exceptions (null, if none)
   * @param theories the theories on which the following depend in dependence order.
   * @param decls a sequence of declaration lists.
   * @param out the stream to which to write error messages.
   **************************************************************************/
  public static void register(ParamSymbol method,
    StateType stateType, Name[] stateVars, Name[] stateExcs,
    Vector<TheorySymbol> theories, Declaration[][] decls,
    ErrorStream out)
  {
    infos.add(new Info(method, stateType, stateVars, stateExcs, theories, decls, out));
  }
 
  /****************************************************************************
   * Clear the registered tasks.
   ***************************************************************************/
  public static void clearTasks()
  {
    infos = new Vector<Info>();
  }
  
  /****************************************************************************
   * Determine whether there are tasks.
   * @return true iff there are tasks.
   ***************************************************************************/
  public static boolean exist()
  {
    return !infos.isEmpty();
  }
  
  /***************************************************************************
   * Generate those tasks for which the information has been previously registered.
   * @param out the stream to which messages are written
   ***************************************************************************/
  public static void generateTasks(PrintWriter out)
  {
    // determine the processed methods and their call sets
    Map<ParamSymbol, Set<ParamSymbol>> methods = 
      new MapClass<ParamSymbol, Set<ParamSymbol>>();
    for (Info info : infos)
      methods.put(info.method, info.method.getCall());
    
    // compute transitive closure of the call sets
    boolean changed;
    do
    {
      changed = false;
      Map<ParamSymbol, Set<ParamSymbol>> methods0 = 
        new MapClass<ParamSymbol, Set<ParamSymbol>>(methods);
      for (Map.Entry<ParamSymbol, Set<ParamSymbol>> entry : methods0.entrySet())
      {
        Set<ParamSymbol> called = entry.getValue();
        Set<ParamSymbol> called0 = new SetClass<ParamSymbol>(called);
        for (ParamSymbol method0 : called0)
        {
          if (!methods.containsKey(method0))
          {
            methods.put(method0, method0.getCall());
            changed = true;
          }
          Set<ParamSymbol> called1 = method0.getCall();
          for (ParamSymbol method1 : called1)
          {
            if (!called.contains(method1))
            {
              called.add(method1);
              changed = true;
            }
          }
        }
      }
    } while (changed);
    
    // compute recursion set for each method and generate termination tasks
    infos: for (Info info : infos)
    {
      ParamSymbol method = info.method; 
      Set<ParamSymbol> recursive = new SetClass<ParamSymbol>();
      method.setRecursion(recursive);
      Set<ParamSymbol> called = method.getCall();
      for (ParamSymbol method0 : called)
      {
        Set<ParamSymbol> called0 = methods.get(method0);
        if (called0.contains(method)) recursive.add(method0);
      }
      
      // check that recursive method has measure
      Term[] termination = null;
      MethodSpec spec = method.getSpecification();
      if (spec != null) termination = spec.getTerminationTerm();
      if (!recursive.isEmpty() && termination == null)
      {
        String message =
          "method " + method.getLongName() + 
          " is recursive but has no recursion measure;\n" +
          "the recursion set is " + toString(methods.get(method));
        TaskFolder tasks = method.getObligations();
        tasks.add(new FailedTask(method.getName().toString(), 
            "invalid measure", message, 
            method.getDeclaration().getPosition(), tasks.getDirectory()));
        out.print(".");
        out.flush();
        continue;
      }
      
      // check that all methods in recursion set have (same kind of) measure
      int n = termination == null ? -1 : termination.length;
      for (ParamSymbol method0 : recursive)
      {
        Term[] termination0 = null;
        int n0 = -1;
        MethodSpec spec0 = method0.getSpecification();
        if (spec0 != null) termination0 = spec0.getTerminationTerm();
        if (termination0 != null) n0 = termination0.length;
        if (n0 == -1)
        {
          String message =
            "method " + method.getLongName() + " calls recursively method " + 
            method0.getLongName() + " which has no recursion measure;\n" +
            "the recursion set is " + toString(methods.get(method));
          TaskFolder tasks = method.getObligations();
          tasks.add(new FailedTask(method0.getName().toString(), 
              "invalid measure", message, 
              method.getDeclaration().getPosition(), tasks.getDirectory()));
          out.print(".");
          out.flush();
          continue infos;
        }
        if (n0 != n)
        {
          String message =
            "method " + method.getLongName() + 
            " has recursion measure of length " + n + " but\n" + 
            "calls recursively method " + method0.getLongName() + 
            " with recursion measure of length " + n0 + ";\n" +
            "the recursion set is " + toString(methods.get(method));
          TaskFolder tasks = method.getObligations();
          tasks.add(new FailedTask(method0.getName().toString(), 
              "invalid measure", message, 
              method.getDeclaration().getPosition(), tasks.getDirectory()));
          out.print(".");
          out.flush();
          continue infos;
        }
      }
      
      // if everything is fine, generate termination tasks
      generate(method, info.stateType, info.stateVars, info.stateExcs,
          info.theories, info.decls, info.out, out);
    }
    
    // reset registered information
    clearTasks();
  }
  
  /***************************************************************************
   * Convert set of method symbols to string.
   * @param set the set.
   * @return the string representation of the set.
   **************************************************************************/
  private static String toString(Set<ParamSymbol> set)
  {
    StringBuffer b = new StringBuffer();
    b.append("{");
    int n = set.size();
    int i = 0;
    for (ParamSymbol method : set)
    {
      b.append(method.getLongName());
      if (i+1 < n) b.append(", ");
      i = i+1;
    }
    b.append("}");
    return b.toString();
  }
  
  // as far as possible, avoid tasks in the application of the judgement
  public final static boolean avoidTasks = true;
  
  /***************************************************************************
   * Construct the termination tasks for a given method.
   * @param method the symbol of a method with a specification.
   * @param stateType the type of the current state (null, if none)
   * @param stateVars the names of the modifiable variables (null, if none)
   * @param stateExcs the names of the throwable exceptions (null, if none)
   * @param theories the theories on which the following depend in dependence order.
   * @param decls a sequence of declaration lists.
   * @param out the stream to which to write error messages.
   * @param progress the stream to which to write progress information
   **************************************************************************/
  private static void generate(ParamSymbol method,
    StateType stateType, Name[] stateVars, Name[] stateExcs,
    Vector<TheorySymbol> theories, Declaration[][] decls,
    ErrorStream out, PrintWriter progress)
  { 
    // get task folder of method
    TaskFolder tasks = method.getObligations();
    
    // the variables modified and exceptions thrown by the method
    Set<VariableSymbol> variables = Normalization.modifiableVariables(stateVars);
    Set<ClassSymbol> exceptions = NameUtils.getClassSymbols(stateExcs);
    
    // verify well-formedness of termination measure
    MethodSpec spec = method.getSpecification();
    Formula preCondition = null;
    Term[] termination = null;
    Formula diverges = null;
    if (spec != null)
    {
      preCondition = Normalization.preCondition(method, true);
      termination = spec.getTerminationTerm();
      if (termination != null)
      {
        int n = termination.length;
        Term[] termination0 = new Term[n];
        for (int i=0; i<n; i++)
          termination0[i] = Normalization.perform(termination[i]);
        termination = termination0;
      }
      diverges = spec.getDivergesCondition();
      if (diverges != null) diverges = Normalization.perform(diverges);
    }
    if (termination != null)
    {
      Formula goal = null;
      int n = termination.length;
      for (int i=n-1; i>=0; i--)
      {
        Term zero = ExpressionLogic.intLiteral("0");
        zero.setType(IntType.construct());
        Formula notNegative =
          GreaterEqualFormula.construct(termination[i], zero);
        if (goal == null)
          goal = notNegative;
        else
          goal = AndFormula.construct(notNegative, goal);
      }
      if (preCondition != null)
        goal = ImpliesFormula.construct(preCondition, goal);
      tasks.add(new TerminationTask("measure",
          method, goal,
          null, "measure", tasks.getDirectory(), stateType,
          variables, exceptions,
          theories, decls, out));
      progress.print(".");
      progress.flush();
    }
    
    // verify termination of method body according to specification
    ValueSymbol[] symbols = null;
    TypedIdentifier[] tids = null;
    Formula preState = null;
    if (termination != null)
    {
      int n = termination.length;
      symbols = new ValueSymbol[n];
      tids = new TypedIdentifier[n];
      Type type = IntType.construct();
      for (int i=n-1; i>=0; i--)
      {
        Identifier name = Identifier.construct("measure" + (i == 0 ? "" : i));
        TypedIdentifier tid = TypedIdentifier.construct(name, type);
        symbols[i] = ValueSymbol.construct(name, type, tid);
        tids[i] = tid;
        Formula equals = EqualsFormula.construct(
            Satisfies.reference(name), termination[i]);
        preState = Satisfies.andFormula(equals, preState);
      }
    }
    if (diverges != null)
      preState = Satisfies.andFormula(NotFormula.construct(diverges), preState);
    if (preCondition != null)
      preState = Satisfies.andFormula(preCondition, preState);
    TaskFolder tasks0 = method.getLoops();
    Terminates terminates = 
      new Terminates(method, stateType, tasks0, theories, decls, out);
    Formula preGoal =
      terminates.derive(method.getDeclaration().getBody(), symbols, 
          preState == null ? ExpressionLogic.booleanFormula(true) : preState);
    if (avoidTasks && preGoal != null)
    {
      Formula goal =
        Satisfies.forallFormula(
            tids == null ? new TypedIdentifier[] { } : tids,
                Satisfies.impliesFormula(preState, preGoal));
      tasks.add(new TerminationTask("termination", 
          method, goal,
          null, "body", tasks.getDirectory(), stateType,
          variables, exceptions,
          theories, decls, out));
      progress.print(".");
      progress.flush();
    }
  }
}
