// ----------------------------------------------------------------------------
// Simplify.java
// The "simplify" command
// $Id: Simplify.java,v 1.21 2008/10/13 09:32:00 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import fmrisc.ProofNavigator.State;
import fmrisc.ProofNavigator.Proving.*;

import java.io.*;
import org.w3c.dom.*;

/**
 * 
 * Command "simplify": simplify a proof state.
 *
 */
public final class Simplify extends ProofCommandBase
{
  public final static String Name = "simplify";
  private String label; // the label of the formula to be simplified (or null)
  
  /***************************************************************************
   * Create a "simplify" command.
   * @param label the label of the formula to be simplified
   *        (null, if all formulas in proof state are to be simplified)
   **************************************************************************/
  public Simplify(String label)
  {
    super(Name);
    this.label = label;
  }
  
  /****************************************************************************
   * Process command "simplify". 
   * Generates a single child with formula(s) simplified.
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    if (state.useAutoSimplify())
    {
      error("simplify called in proof state with automatic simplification");
      return null;
    }
    
    // create new state from existing state
    ProofState state0 = new ProofState(state, state.getEnvironment());
    Formula[] assumptions0 = state0.copy(state.getAssumptions());
    Formula[] goals0 = state0.copy(state.getGoals());
    state0.setAssumptions(assumptions0);
    state0.setGoals(goals0);
    
    // simplify formula with automatic simplification switched on
    if (label == null)
    {
      boolean oldAuto = State.useAutoSimplify();
      State.setAutoSimplify(true);
      boolean okay = ProverReasoning.simplify(state0, false, false);
      State.setAutoSimplify(oldAuto);
      if (!okay) return null;
    }
    else
    {
      Formula formula0 = state0.getFormula(label);
      if (formula0 == null)
      {
        error("No formula with label " + label);
        return null;
      }
      boolean oldAuto = State.useAutoSimplify();
      State.setAutoSimplify(true);
      boolean okay = ProverReasoning.simplify(state0, formula0, false);
      State.setAutoSimplify(oldAuto);
      if (!okay) return null;
    }
     
    // return simplified state
    return new ProofState[] { state0 };
  }
  
  /***************************************************************************
   * Set external representation of command in a DOM node.
   * @param node the node in which the representation is to be set
   *        (by default none, to be overridden by command class)
   **************************************************************************/
  protected void setRepresentation(Node node)
  {
    if (label != null)
      node.appendChild(newTextNode(label));
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    NodeList children = node.getChildNodes();
    int n = children.getLength();
    if (n > 1) 
    {
      State.setError("at most one child expected: " + node.toString());
      return null;
    }
    if (n == 0) return new Simplify(null);
    String label = toString(children.item(0));
    if (label == null) return null;
    return new Simplify(label);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
    if (label == null) return;
    out.print(" " + label);
  }
}

