// ----------------------------------------------------------------------------
// Presenter.java
// Interface to visual presentation of declarations and proof states.
// $Id: Presenter.java,v 1.51 2017/07/04 08:04:35 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Communication;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Commands.*;
import fmrisc.ProofNavigator.Proving.*;
import fmrisc.ProofNavigator.Semantics.*;
import fmrisc.ProofNavigator.Syntax.*;

import org.w3c.dom.*;

import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * 
 * Interface to visual presentation of declarations and proof states.
 *
 */
public final class Presenter
{
  // directory where presentation files are stored
  private File directory;
  
  // document where declarations are written
  private Document declDocument;
  
  // node where declarations are added
  private Element declNode;
  
  // file where declarations are written
  private File declFile;
  
  // MathML generator used for writing declarations
  private PrettyMathML declMathML;
  
  // the sequence of declarations written so far
  private Vector<Declaration> declarations;
  
  // fontsize used in documents
  private String fontSize;
  
  // colors used for states
  private String openColor = "#ff0040";
  private String closedColor = "blue";
  
  /***************************************************************************
   * Create instance of presenter with files in denoted directory.
   * @param directory for storing the presentation files.
   **************************************************************************/
  public Presenter(File directory)
  {
    this.directory = directory;
    if (State.useBigFont())
      fontSize = "14px";
    else
      fontSize = "12px";
    boolean okay = newDeclarations();
    if (!okay)
    {
      System.out.println("Error: " + State.getError());
      System.exit(-1);
    }
  }
  
  /***************************************************************************
   * Update font size depending on user settings.
   **************************************************************************/
  public void resize()
  {
    if (State.useBigFont())
      fontSize = "14px";
    else
      fontSize = "12px";
    updateDeclarations();
  }
  
  /***************************************************************************
   * Creates new declaration document.
   * @return true iff everything went well.
   **************************************************************************/
  public boolean newDeclarations()
  {
    // base name of file must not overlap with declaration name, thus "-"
    declDocument = createDocument("Declarations", true);
    if (declDocument == null) return false;
    declNode = declDocument.createElement("table");
    declNode.setAttribute("width", "100%");
    Element root = getRoot(declDocument);
    root.appendChild(declNode);
    declNode.setAttribute("cellspacing", "0");
    declNode.setAttribute("cellpadding", "0");
    scrollToBottom(declDocument);
    declFile = new File(directory, "decl-list.xhtml");
    boolean okay = writeDocument(declDocument, declFile);
    if (!okay) return false;
    declarations = new Vector<Declaration>();
    declMathML = new PrettyMathML(declDocument);
    return true;
  }
  
  /***************************************************************************
   * Modify document such that on loading it is automatically scrolled
   * to bottom. Must be called after all content has been added to document.
   * @param document a document.
   **************************************************************************/
  private void scrollToBottom(Document document)
  {
    Element root = getRoot(document);
    root.setAttribute("onload", 
        "document.getElementById('bottom').scrollIntoView(false);");
    Element bottom = document.createElement("a");
    bottom.setAttribute("id", "bottom");
    root.appendChild(bottom);
  }
  
  /***************************************************************************
   * Get file where all declarations are stored
   * @return the declaration file
   **************************************************************************/
  public File getDeclFile()
  {
    return declFile;
  }
  
  /****************************************************************************
   * Construct file name for proof state
   * @param state the proof state.
   * @return the name of the file for visualizing the proof state.
   ***************************************************************************/
  public File getFile(ProofState state)
  {
    return new File(directory, state.getPath() + ".xhtml");
  }
  
  /****************************************************************************
   * Construct file name for proof state listing typed constants.
   * @param state the proof state.
   * @return the name of the file for listing the typed constants of the state.
   ***************************************************************************/
  private File getTypeFile(ProofState state)
  {
    return new File(directory, state.getPath() + "-const.xhtml");
  }
  
  /***************************************************************************
   * Create new XHTMl document.
   * @param title the document title
   * @param type (0 = MathML, 1 = Frameset)
   * @param bottom if true some extra space is added at bottom for type 0
   * @return the new document (null if an error occured)
   **************************************************************************/
  private Document createDocument(String title, int type, boolean bottom)
  {
    try
    {
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      factory.setNamespaceAware(true);
      DocumentBuilder builder = factory.newDocumentBuilder();
      DOMImplementation impl = builder.getDOMImplementation();
      DocumentType doctype = null;
      if (type == 0) 
        doctype = impl.createDocumentType("html", 
          "-//W3C//DTD XHTML 1.1 plus MathML 2.0//EN",
          "http://www.w3.org/TR/MathML2/dtd/xhtml-math11-f.dtd"); 
      else if (type == 1)
        doctype = impl.createDocumentType("html", 
            "-//W3C//DTD XHTML 1.0 Frameset//EN",
        "http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd");
      Document document = 
        impl.createDocument("http://www.w3.org/1999/xhtml", "html", doctype);
      Element root = document.getDocumentElement();
      Element head = document.createElement("head");
      root.appendChild(head);
      Element tnode = document.createElement("title");
      head.appendChild(tnode);
      tnode.appendChild(document.createTextNode(title));
      Element style = document.createElement("style");
      style.setAttribute("type", "text/css");
      head.appendChild(style);
      style.appendChild(document.createTextNode(
       "math { font-size: 110%; } " +
       "a { text-decoration: none; } " +
       "a.open { text-decoration: none; color: " + openColor + "; } " +
       "a.closed { text-decoration: none; color: " + closedColor + "; } " +
       "a.open:hover { text-decoration: none; color: white; background: " +
         openColor + "; } " +
       "a.closed:hover { text-decoration: none; color: white; background: " +
         closedColor + "; } " + 
       "div.dlblock { position: relative; } " +
       "td.dlabel { border: 1px outset; padding-right: 15px; margin: 0px; } " +
       "td.dlabel:hover { background: blue; color: white; } " +
       "td.dcontent { padding-left: 3px; } " +
       "div.lblock { position: relative; } " +
       "td.label { border: 1px outset; padding: 1px; } " +
       "td.label:hover { background: blue; color: white; } " +
       "table.menu { display: none; } " +
       "td:hover table.menu { display: block; position: absolute; " +
       "z-index: 1000; width: 300px; top: 0px; left: 12px; " +
       "border: 2px outset; background: #eeeeee; color: black; " +
       "border-collapse: collapse; } " +
       ".menu caption { width: 100%; color: " + openColor + ";" + 
         " border-bottom: 1px solid; border-color: black; font-weight: bold; } " +
       ".menu td { width: 1%; } " +
       "td.entry:hover { background: blue; color: white; } " +
       ".entry a { text-decoration: none; color: black; } " +
       ".entry a:hover { color: white; }"
      ));
      if (type == 0)
      {
        Element body = document.createElement("body");
        root.appendChild(body);
        String b = "0px";
        if (bottom) b = "60px";
        body.setAttribute("style",
            "margin:0 0 " + b + " 0; font-family: Helvetica, sans; font-size: " + 
            fontSize + ";");
      }
      else if (type == 1)
      {
        root.appendChild(document.createElement("frameset"));
      }
      return document;
    }
    catch(ParserConfigurationException e)
    {
      State.setError("exception " + e.getMessage());
      return null;
    } 
  }
  
  /***************************************************************************
   * Create new XHTML/MathML document.
   * @param title the document title
   * @param bottom true if extra space is added at bottom
   * @return the new document (null if an error occured)
   **************************************************************************/
  private Document createDocument(String title, boolean bottom)
  {
    return createDocument(title, 0, bottom);
  }
  
  /***************************************************************************
   * Create new XHTML frameset document.
   * @param title the document title
   * @return the new document (null if an error occured)
   **************************************************************************/
  private Document createFrameset(String title)
  {
    return createDocument(title, 1, false);
  }
  
  /***************************************************************************
   * Get root node of document to which new nodes should be appended.
   * @param document the document.
   * @return its root node.
   **************************************************************************/
  public Element getRoot(Document document)
  {
    return (Element)document.getDocumentElement().getChildNodes().item(1);
  }
  
  /***************************************************************************
   * Write document to denoted file.
   * @param the document.
   * @param file
   * @return true iff everything went okay.
   **************************************************************************/
  private boolean writeDocument(Document document, File file)
  {
    try
    {
      TransformerFactory tfactory = TransformerFactory.newInstance();
      Transformer transformer = tfactory.newTransformer();
      transformer.setOutputProperty(OutputKeys.METHOD, "xml");
      transformer.setOutputProperty(OutputKeys.INDENT, "yes");
      transformer.setOutputProperty(OutputKeys.DOCTYPE_PUBLIC,  
        document.getDoctype().getPublicId());
      transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM,
        document.getDoctype().getSystemId());
      DOMSource source = new DOMSource(document);
      OutputStream out = new BufferedOutputStream(new FileOutputStream(file));
      StreamResult result = new StreamResult(out);
      transformer.transform(source, result);
      out.close();
      return true;
    }
    catch(TransformerException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
    catch(IOException e)
    {
      State.setError("exception " + e.getMessage());
      return false;
    }
  }
  
  /***************************************************************************
   * Write proof state and return corresponding file.
   * @param state the state to be written to file.
   * @return the new file (null, if an error occured).
   **************************************************************************/
  public File writeProofState(ProofState state)
  {
    File file = getFile(state);
    File tfile = getTypeFile(state);
    
    Document document = createDocument(state.getPath(), false);
    if (document == null) return null;
    appendProofState(document, tfile, state);
    scrollToBottom(document);
    boolean okay = writeDocument(document, file);
    if (!okay) return null;
    
    Document tdocument = 
      createDocument(state.getPath() + ": Constants", false);
    if (tdocument == null) return null;
    appendProofStateConstants(tdocument, file, state);
    okay = writeDocument(tdocument, tfile);
    if (!okay) return null;
    
    return file;
  }
  
  /***************************************************************************
   * Write to documents the DOM representation of the proof state.
   * @param document the document to write the proof state to.
   * @param tfile the file holding the associated constant declarations.
   * @param state the proof state that is to be presented.
   **************************************************************************/
  private void appendProofState(Document document, 
    File tfile, ProofState state)
  {
    Node parent = getRoot(document);
    Environment env = state.getEnvironment();
    Formula[] axioms = state.getAxioms();
    Formula[] assumptions = state.getAssumptions();
    Formula[] goals = state.getGoals();
    Command command = state.getCommand();
    Expression[] counterExample = state.getCounterExample();
    
    Element hsection = document.createElement("div");
    parent.appendChild(hsection);
    hsection.setAttribute("style", 
      "padding: 6px; background: #f0f0f0; border-style: solid;" +
      "border-color: black; border-width: thin;");
    
    final Element header = document.createElement("b");
    hsection.appendChild(header);
    String tags;
    if (state.useAutoSimplify())
      tags = "autosimp";
    else
      tags = "no autosimp";
    State.ProverVersion proverVersion = state.proverVersion();
    if (proverVersion == State.ProverVersion.CVCL)
      tags = tags + ",CVCL";
    else if (proverVersion == State.ProverVersion.CVC3)
      tags = tags + ",CVC3";
    else if (proverVersion == State.ProverVersion.CVC4)
      tags = tags + ",CVC4";
    else
      tags = tags + ",UNKNOWN";
    if (state.firstOrder())
      tags = tags + ",first order";
    else
      tags = tags + ",higher order";
    if (state.toLogical())
      tags = tags + ",logical";
    else
      tags = tags + ",boolean";
    Node title = document.createTextNode(
      "Formula [" + state.getProof().getSymbol().getIdentifier().toString() +
      "] proof state [" + state.getLabel() + "] (" + tags + ")");
    header.appendChild(title);
    String color;
    if (isClosed(state))
      color = closedColor;
    else
      color = openColor;
    header.setAttribute("style", "color: " + color + ";");
    
    if (command != null) 
      header.appendChild(document.createTextNode(": " + command.toString()));
    
    Element psection0 = null;
    if (env != null)
    {
      Collection<Symbol> symbols = env.getValueSymbols();
      if (!symbols.isEmpty())
      {
        psection0 = document.createElement("div");
        parent.appendChild(psection0);
        psection0.setAttribute("style", "padding: 6px 6px 0 6px;");
        psection0.appendChild(document.createTextNode("Constants "));
        Element anode = document.createElement("a");
        psection0.appendChild(anode);
        anode.appendChild(document.createTextNode("(with types)"));
        anode.setAttribute("href", tfile.getName());
        psection0.appendChild(document.createTextNode(": "));
        Iterator<Symbol> iterator = symbols.iterator();
        ValueSymbol symbol = (ValueSymbol)iterator.next();
        MathML mathml = new MathML(document);
        Element mnode = MathML.createMathNode(document);
        psection0.appendChild(mnode);
        mathml.appendIdentifier(mnode, symbol.getValueDeclIdentifier());
        while (iterator.hasNext())
        {
          psection0.appendChild(document.createTextNode(", "));
          mnode = MathML.createMathNode(document);
          psection0.appendChild(mnode);
          symbol = (ValueSymbol)iterator.next();
          mathml.appendIdentifier(mnode, symbol.getValueDeclIdentifier());
        }
        psection0.appendChild(document.createTextNode("."));
        psection0.appendChild(document.createElement("br"));
      }
    }
    
    if (axioms.length != 0)
    {
      if (psection0 == null)
      { 
        psection0 = document.createElement("div");
        parent.appendChild(psection0);
      }
      psection0.setAttribute("style", "padding: 6px 6px 0 6px;");
      psection0.appendChild(document.createTextNode("Axioms: "));
      MathML mathml = new MathML(document);
      Element mathnode = MathML.createMathNode(document);
      psection0.appendChild(mathnode);
      int n = axioms.length;
      for (int i=0; i<n; i++)
      {
        mathml.appendExp(mathnode, axioms[i].getExpression());
        if (i+1 < n) mathnode.appendChild(document.createTextNode(", "));
      }
      psection0.appendChild(document.createTextNode("."));
      psection0.appendChild(document.createElement("br"));
    }
    
    Element psection = document.createElement("div");
    parent.appendChild(psection);
    psection.setAttribute("style", "padding: 12px 6px 12px 6px;");
    
    appendFormulas(document, psection, assumptions);
    
    Element hrnode1 = document.createElement("hr");
    psection.appendChild(hrnode1);
    hrnode1.setAttribute("style", 
      "width: 240px; text-align: left; margin: 4px auto 4px 0px;");
   
    appendFormulas(document, psection, goals);
    
    if (counterExample != null)
    {
      Element psection1 = document.createElement("div");
      parent.appendChild(psection1);
      psection1.setAttribute("style", "padding: 6px;");
      psection1.appendChild(
       document.createTextNode("This may be a counterexample:"));
      psection1.appendChild(
       document.createElement("br"));
      appendExpressions(document, psection1, counterExample);
    }
    
    ProofState pstate = state.getParent();
    ProofState[] children = state.getChildren();
    if (pstate == null && children == null) 
    {
      Element tsection = document.createElement("div");
      parent.appendChild(tsection);
      tsection.setAttribute("style", "margin-bottom: 80px;");
      return;
    }
    
    Element tsection = document.createElement("div");
    parent.appendChild(tsection);
    tsection.setAttribute("style", 
        "padding: 6px; background: #f0f0f0; border-style: solid;" +
        "border-color: black; border-width: thin; margin-bottom: 60px; ");
    
    Element bnode = document.createElement("b");
    tsection.appendChild(bnode);
    
    if (pstate != null)
    {
      bnode.appendChild(document.createTextNode("Parent: "));
      Element anode = document.createElement("a");
      bnode.appendChild(anode);
      anode.appendChild(document.createTextNode(
          "[" + pstate.getLabel() + "]"));
      anode.setAttribute("href", getFile(pstate).getName());
      bnode.appendChild(document.createTextNode(" "));
    }

    if (children != null)
    {
      bnode.appendChild(document.createTextNode("Children: "));
      int n = children.length;
      for (int i=0; i<n; i++)
      {
        ProofState child = children[i];
        Element anode = document.createElement("a");
        bnode.appendChild(anode);
        anode.appendChild(document.createTextNode(
            "[" + child.getLabel() + "]"));
        anode.setAttribute("href", getFile(child).getName());
        bnode.appendChild(document.createTextNode(" "));
      }
    }
  }
  
  /***************************************************************************
   * Write to documents the DOM representation of the constants
   * associated to the proof state.
   * @param document the document to write the proof state to.
   * @param file the file holding the associated proof state.
   * @param state the proof state that is to be presented.
   **************************************************************************/
  private void appendProofStateConstants(Document document, 
    File file, ProofState state)
  {
    Node parent = getRoot(document);
    Element hsection = document.createElement("div");
    parent.appendChild(hsection);
    hsection.setAttribute("style", 
      "padding: 6px; background: #f0f0f0; border-style: solid;" +
      "border-color: black; border-width: thin;");
    
    final Element header = document.createElement("b");
    hsection.appendChild(header);
    Node title = document.createTextNode(
      "Formula [" + state.getProof().getSymbol().getIdentifier().toString() +
      "] proof state [" + state.getLabel() + "]: Constant Declarations ");
    header.appendChild(title);
    String color;
    if (isClosed(state))
      color = closedColor;
    else
      color = openColor;
    header.setAttribute("style", "color: " + color + ";");
    
    Environment env = state.getEnvironment();
    Element psection0 = null;
    if (env != null)
    {
      Collection<Symbol> symbols = env.getValueSymbols();
      if (!symbols.isEmpty())
      {
        psection0 = document.createElement("div");
        parent.appendChild(psection0);
        psection0.setAttribute("style", "padding: 6px 6px 0 6px;");
        Element anode = document.createElement("a");
        psection0.appendChild(anode);
        anode.appendChild(document.createTextNode("[To Proof State]"));
        anode.setAttribute("href", file.getName());
        Element list = document.createElement("ul");
        psection0.appendChild(list);
        MathML mathml = new MathML(document);
        Iterator<Symbol> iterator = symbols.iterator();
        while (iterator.hasNext())
        {
          Element item = document.createElement("li");
          list.appendChild(item);
          Element mnode = MathML.createMathNode(document);
          item.appendChild(mnode);
          ValueSymbol symbol = (ValueSymbol)iterator.next();
          mathml.appendTypedIdentifier(mnode,
              Construct.typedIdentifier(symbol.getValueDeclIdentifier(), 
                  symbol.getType()));
        }
      }
    }
    
    Element tsection = document.createElement("div");
    parent.appendChild(tsection);
    tsection.setAttribute("style", 
        "padding: 6px; background: #f0f0f0; border-style: solid;" +
        "border-color: black; border-width: thin;");
    Element bnode = document.createElement("b");
    tsection.appendChild(bnode);
    Element anode = document.createElement("a");
    bnode.appendChild(anode);
    anode.appendChild(document.createTextNode("[To Proof State]"));
    anode.setAttribute("href", file.getName());
  }
  
  /***************************************************************************
   * Auxiliary class for determingthe state color.
   **************************************************************************/
  /*
  private final static class StateColor implements Runnable
  {
    private TreeItem item;
    public boolean isClosed;
    public StateColor(TreeItem item)
    {
      this.item = item;
    }
    public void run() 
    {
      if (item.getForeground().equals(MainSWT.getClosedColor()))
        isClosed = true;
      else
        isClosed = false;
    }
  }
  */
  
  /***************************************************************************
   * Determine whether state and its children are closed.
   * @param state the state.
   * @return true iff state and its children are closed.
   **************************************************************************/
  private boolean isClosed(ProofState state)
  {
    // it is too costly to update state colors on files
    // thus we leave them in one standard color
    return false;
    // ProofStateListener listener = state.getListener();
    // if (listener == null) return false;
    // TreeItem item = ((ProofTreeItem)listener).getTreeItem();
    // StateColor scolor = new StateColor(item);
    // item.getDisplay().syncExec(scolor);
    // return scolor.isClosed;
  }
  
  /***************************************************************************
   * Append to parent node the DOM representation of the denoted formulas.
   * @param the document to which the nodes belong.
   * @param parent the parent node.
   * @param formulas the formulas to be presented.
   **************************************************************************/
  private void appendFormulas(Document document, 
    Node parent, Formula[] formulas)
  {
    int n = formulas.length;
    if (n == 0) return;
    Element tnode = document.createElement("table");
    tnode.setAttribute("width", "100%");
    tnode.setAttribute("cellspacing", "0");
    tnode.setAttribute("cellpadding", "0");
    parent.appendChild(tnode);
    for (int i=0; i<n; i++)
    {
      Element trnode = document.createElement("tr");
      tnode.appendChild(trnode);
      Formula f = formulas[i];
      Element tdnode0 = document.createElement("td");
      tdnode0.setAttribute("class", "label");
      tdnode0.setAttribute("valign", "top");
      trnode.appendChild(tdnode0);
      Element divnode = document.createElement("div");
      divnode.setAttribute("class", "lblock");
      tdnode0.appendChild(divnode);
      String label = f.getLabel();
      divnode.appendChild(document.createTextNode(label));
      appendMenu(document, divnode, f);
      ProofState pstate = f.getState().getParent();
      boolean newFormula = pstate != null && pstate.getFormula(label) == null;
      Element tdnode1 = document.createElement("td");
      if (newFormula)
        tdnode1.setAttribute("style", 
          "padding-left:5px; width: 100%; border-right-style: solid; " +
          "border-right-width: thick; border-color: red;");
      else
        tdnode1.setAttribute("style", "padding-left:5px; width: 100%;");
      trnode.appendChild(tdnode1);
      PrettyMathML mathml = new PrettyMathML(document);
      mathml.appendExp(tdnode1, f.getExpression());
    }
  }
  
  /***************************************************************************
   * Append to parent node a menu for dealing with denoted formula.
   * @param document the document to which the nodes belong.
   * @param parent the parent node.
   * @param formula the formula to which the menu belongs.
   **************************************************************************/
  private void appendMenu(Document document, Node parent, Formula formula)
  {
    Element tnode = document.createElement("table");
    parent.appendChild(tnode);
    tnode.setAttribute("class", "menu");
    Element cnode = document.createElement("caption");
    tnode.appendChild(cnode);
    cnode.appendChild(document.createTextNode("Formula [" + 
      formula.getLabel() + "]"));
    
    String label = formula.getLabel();
    
    // simplify formula
    if (!formula.getState().useAutoSimplify())
      appendCommand(document, tnode, 
          "simplify " + label + ": Simplify Formula",
          new Simplify(label));
    
    // split formula
    Expression exp = formula.getExpression();
    if ((formula.isGoal() && exp instanceof AndFormula) ||
        (formula.isGoal() && exp instanceof EquivalentFormula) ||
        (formula.isGoal() && exp instanceof NotEquivalentFormula) ||
        (!formula.isGoal() && exp instanceof OrFormula) ||
        (!formula.isGoal() && exp instanceof ImpliesFormula) ||
        exp instanceof IfThenElseExpression)
      appendCommand(document, tnode, 
        "split " + label + ": Split Formula",
        new Split(label));
    
    // instantiate formula
    if ((formula.isGoal() && 
          formula.getExpression() instanceof ExistsFormula) ||
        (!formula.isGoal() && 
          formula.getExpression() instanceof ForallFormula))
    {
      appendCommand(document, tnode, 
        "auto " + label + ": Automatically Instantiate Formula", 
          new Auto(new String[] { label }));
      appendPattern(document, tnode, 
          "instantiate [] in " + label + ": Instantiate Variable(s) in Formula", 
          "instantiate [] in " + label);
    }
    
    // expand definitions in formula   
    appendPattern(document, tnode, 
        "expand [] in " + label + ": Expand Definition(s) in Formula", 
        "expand [] in " + label);
    
    // induction on goal
    if (formula.isGoal())
    {
      appendPattern(document, tnode, 
          "induction [] in " + label + ": Induction on Formula", 
          "induction [] in " + label);
    }
    
    // make formula single goal
    if (!formula.isGoal())
    {
      appendCommand(document, tnode, 
          "goal " + label + ": Make Formula Goal", new Goal(label));
    }
    
    // flip formula
    appendCommand(document, tnode, 
        "flip " + label + ": Flip Formula", new Flip(label));
  }
 
  /***************************************************************************
   * Append to parent node a menu entry for executing a particular command.
   * @param document the document to which the nodes belong.
   * @param parent the parent node.
   * @param label the label appearing in the command menu.
   * @param command the command to be executed.
   **************************************************************************/
  private void appendCommand(Document document, Node parent, 
    String label, Command command)
  {
    appendCommand(document, parent, label, command.toString(), true);
  }
  
  /***************************************************************************
   * Append to parent node a menu entry for inputting a particular pattern
   * @param document the document to which the nodes belong.
   * @param parent the parent node.
   * @param label the label appearing in the command menu.
   * @param pattern the pattern to be inputted
   **************************************************************************/
  private void appendPattern(Document document, Node parent, 
    String label, String pattern)
  {
    appendCommand(document, parent, label, pattern + ";", false);
  }
  
  /***************************************************************************
   * Append to parent node a menu entry for executing a particular command.
   * @param document the document to which the nodes belong.
   * @param parent the parent node.
   * @param label the label appearing in the command menu.
   * @param command the command to be executed.
   * @param execute false iff command only becomes input
   **************************************************************************/
  private void appendCommand(Document document, Node parent, 
    String label, String command, boolean execute)
  {
    Element trnode0 = document.createElement("tr");
    parent.appendChild(trnode0);
    Element tdnode0 = document.createElement("td");
    trnode0.appendChild(tdnode0);
    tdnode0.setAttribute("class", "entry");
    Element anode0 = document.createElement("a");
    tdnode0.appendChild(anode0);
    try
    {
      String key;
      if (execute)
        key = "/execute?command=";
      else
        key = "/input?pattern=";
      anode0.setAttribute("href", 
        "http://localhost:" + State.getServerPort() + key + 
        URLEncoder.encode(command, "UTF-8"));
    }
    catch(UnsupportedEncodingException e)
    {
      System.out.println("Error in encoding command: " + e);
      Main.exit(-1);
    }
    anode0.appendChild(document.createTextNode(label));
  }
  
  /***************************************************************************
   * Append to parent node the DOM representation of the denoted expressions.
   * @param document the document to which the nodes belong.
   * @param parent the parent node.
   * @param exps the expressions to be presented.
   **************************************************************************/
  private void appendExpressions(Document document, 
    Node parent, Expression[] exps)
  {
    MathML mathml = new MathML(document);
    Element mathnode = MathML.createMathNode(document);
    parent.appendChild(mathnode);
    mathml.appendExp(mathnode, ASTUtil.toConjunction(exps));
  }
  
  /****************************************************************************
   * Write declaration to declaration file.
   * @param decl the declaration to be written.
   * @return true iff everything went okay.
   ***************************************************************************/
  public boolean writeDeclaration(Declaration decl)
  {
    Element trnode = declDocument.createElement("tr");
    declNode.appendChild(trnode);
    Element tdnode0 = declDocument.createElement("td");
    tdnode0.setAttribute("class", "dlabel");
    tdnode0.setAttribute("valign", "top");
    trnode.appendChild(tdnode0);
    Element divnode = declDocument.createElement("div");
    divnode.setAttribute("class", "dlblock");
    tdnode0.appendChild(divnode);
    //divnode.appendChild(declDocument.createTextNode("\u00A0"));
    appendMenu(declDocument, divnode, decl);
    Element tdnode = declDocument.createElement("td");
    tdnode.setAttribute("style", "padding-left:5px; width: 100%;");
    trnode.appendChild(tdnode);
    declMathML.appendDecl(tdnode, decl);
    boolean okay = writeDocument(declDocument, declFile);
    if (okay) declarations.add(decl);
    return okay;
  }
  
  /***************************************************************************
   * Update the declarations written so far (rewriting the declaration file).
   * This is needed if the proof status of a formula (shown in display)
   * has changed by a proof.
   * @return true iff everything went okay.
   **************************************************************************/
  public boolean updateDeclarations()
  {
    Vector<Declaration> d = declarations;
    boolean okay = newDeclarations();
    if (!okay) return false;
    int n = d.size();
    for (int i=0; i<n; i++)
    {
      okay = writeDeclaration((Declaration)d.get(i));
      if (!okay) return false;
    }
    return true;
  }
  
  /***************************************************************************
   * Append to parent node a menu for dealing with denoted declaration.
   * @param document the document to which the nodes belong.
   * @param parent the parent node.
   * @param decl the declaration to which the menu belongs.
   **************************************************************************/
  private void appendMenu(Document document, Node parent, Declaration decl)
  {
    Element tnode = document.createElement("table");
    parent.appendChild(tnode);
    tnode.setAttribute("class", "menu");
    Element cnode = document.createElement("caption");
    tnode.appendChild(cnode);
    String kind;
    if (decl instanceof TypeDeclaration)
      kind = "Type";
    else if (decl instanceof ValueDeclaration)
      kind = "Value";
    else if (decl instanceof AxiomDeclaration)
      kind = "Axiom";
    else if (decl instanceof FormulaDeclaration)
      kind = "Formula";
    else
      kind = "Unknown";
    Identifier ident = decl.getIdentifier();
    String label = kind + " " + ident.toString();
    if (!(decl instanceof AxiomDeclaration) && 
        decl instanceof FormulaDeclaration)
    {
      FormulaSymbol symbol = (FormulaSymbol)decl.getSymbol();
      Proof proof = symbol.getProof();
      if (proof != null)
        label = label + " (" + proof.getStatus() + ")";
    }
    cnode.appendChild(document.createTextNode(label));
    
    if (decl instanceof TypeDeclaration)
    {
      appendCommand(document, tnode, 
        "printt " + ident.getName() + ": Print Type", new TypeC(ident));
    }
    if (decl instanceof ValueDeclaration)
    {
      appendCommand(document, tnode, 
        "printv " + ident.getName() + ": Print Value", new ValueC(ident));
    }
    if (decl instanceof AxiomDeclaration)
    {
      appendCommand(document, tnode, 
        "printf " + ident.getName() + ": Print Axiom", new FormulaC(ident));
    }
    else if (decl instanceof FormulaDeclaration)
    {
      appendCommand(document, tnode, 
        "prove " + ident.getName() + ": Construct Proof", new Prove(ident));
      appendCommand(document, tnode, 
        "proof " + ident.getName() + ": Show Proof", new ProofC(ident));
      appendCommand(document, tnode, 
        "printf " + ident.getName() + ": Print Formula", new FormulaC(ident));
    }
  }
  
  /***************************************************************************
   * Write index of proof to file suitable for later display of proof states.
   * @return file suitable for display (null, if an error occured)
   * @param proof the proof.
   *************************************************************************/
  public File writeProofIndex(Proof proof)
  {
    // names of index file, menu file, root state file
    String baseName = proof.getSymbol().getIdentifier().getName();
    String indexName = baseName + "_index.xhtml";
    String menuName = baseName + "_menu.xhtml";
    
    // root state
    ProofState root = proof.getRoot();
    
    // create index file
    File index = new File(directory, indexName);
    Document idoc = createFrameset("Proof of " + 
      proof.getSymbol().getIdentifier().toString());
    if (idoc == null) return null;
    Element idocNode = getRoot(idoc);
    idocNode.setAttribute("cols", "20%, 80%");
    Element frame1 = idoc.createElement("frame");
    idocNode.appendChild(frame1);
    frame1.setAttribute("src", menuName);
    Element frame2 = idoc.createElement("frame");
    frame2.setAttribute("id", "state");
    idocNode.appendChild(frame2);
    frame2.setAttribute("src", getFile(root).getName());
    boolean okay = writeDocument(idoc, index);
    if (!okay) return null;
    
    // create menu file
    File menu = new File(directory, menuName);
    Document mdoc = createDocument("Menu", false);
    if (mdoc == null) return null;
    Element mdocNode = getRoot(mdoc);
    mdocNode.setAttribute("style",
    "font-family: Helvetica, sans; font-size: 12px;");
    Element pnode = mdoc.createElement("p");
    mdocNode.appendChild(pnode);
    Element anode = mdoc.createElement("a");
    pnode.appendChild(anode);
    anode.setAttribute("href", "decl-list.xhtml");
    anode.setAttribute("target", "state");
    anode.setAttribute("class", "closed");
    anode.appendChild(mdoc.createTextNode("[Declarations]"));
    Element list = mdoc.createElement("dl");
    mdocNode.appendChild(list);
    list.setAttribute("style", "margin: 0 0 0 0;");
    appendStates(list, root);
    okay = writeDocument(mdoc, menu);
    if (!okay) return null;
    
    // update declarations file
    updateDeclarations();
    
    // handle to index file
    return index;
  }
  
  /***************************************************************************
   * Append to parent node subtree of proof states rooted at denoted state.
   * @param parent the parent node
   * @param state the root of the state tree
   *************************************************************************/
  private void appendStates(Node parent, ProofState state)
  {
    Document document = parent.getOwnerDocument();
    Element linode = document.createElement("dt");
    parent.appendChild(linode);
    Element anode = document.createElement("a");
    linode.appendChild(anode);
    anode.setAttribute("href", getFile(state).getName());
    anode.setAttribute("target", "state");
    if (isClosed(state))
      anode.setAttribute("class", "closed");
    else
      anode.setAttribute("class", "open");
    String label = "[" + state.getLabel() + "]";
    Command command = state.getCommand();
    if (command != null) label = label + ": " + command.toString();
    anode.appendChild(document.createTextNode(label));
    Element ddnode = document.createElement("dd");
    ddnode.setAttribute("style", "margin: 0;");
    parent.appendChild(ddnode);
    
    ProofState[] children = state.getChildren();
    if (children == null) return;
    
    Element ulnode = document.createElement("dl");
    ddnode.appendChild(ulnode);
    ulnode.setAttribute("style", "margin: 0 0 0 1em;");
    int n = children.length;
    for (int i=0; i<n; i++)
      appendStates(ulnode, children[i]);
  }
}
