// ----------------------------------------------------------------------------
// InvalidAnswerCVCL.java
// CVCL Answer that formula is invalid.
// $Id: InvalidAnswerCVCL.java,v 1.11 2012/10/24 16:13:33 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Proving.CVCL;

import java.util.*;

import fmrisc.ProgramExplorer.Main;
import fmrisc.ProofNavigator.State;
import fmrisc.ProofNavigator.Proving.*;
import fmrisc.ProofNavigator.Semantics.*;
import fmrisc.ProofNavigator.Syntax.*;

/**
 * 
 * Answer that formula is invalid.
 *
 */
public final class InvalidAnswerCVCL extends InvalidAnswer
{
  private Expression[] counterExample; // counterexample (may be null)
  
  /****************************************************************************
   * Create answer that formula is valid with denoted justification.
   * @param cvcl CVCL instance that gave the answer
   * @param justification why is formula invalid (may be null)
   ***************************************************************************/
  public InvalidAnswerCVCL(CVCL cvcl, String justification)
  {
    super(justification);
    if (cvcl.getVersion() == State.ProverVersion.CVC4)
      makeCounterExampleDecls(cvcl);
    else
      makeCounterExample(cvcl);
  }
  
  /****************************************************************************
   * Interpret justification as a counterexample.
   * @return sequence of formulas whose conjunction represents a refutation
   *         (null, if answer cannot be interpreted in this way)
   ***************************************************************************/
  public Expression[] getCounterExample()
  {
    return counterExample;
  }
  
  /****************************************************************************
   * Interpret justification as a counterexample.
   * @param cvcl the CVCL instance used for the interpretation.
   ***************************************************************************/
  private void makeCounterExample(CVCL cvcl)
  {
    String string = getJustification();
    if (string == null) return;
    Vector<Expression> result = new Vector<Expression>();
    int pos = 0;
    int n = string.length();
    while (pos < n)
    {
      int begin = string.indexOf("ASSERT", pos);
      if (begin == -1) break;
      begin = begin+6;
      int end = string.indexOf(";", begin);
      if (end == -1) return;
      pos = end+1;
      String estring = string.substring(begin, end);
      Expression exp = cvcl.parseExpression(estring);
      if (exp == null) return;

      // rewrite expression according to our rules
      Checking.setWeakerTypes(true);
      Type type = Checking.processFormula(exp);
      Checking.setWeakerTypes(false);
      if (type == null) return;
      exp = Rewriting.rewrite(exp);
      
      // ignore quantified formulas as counterexamples
      if (exp instanceof QuantifiedExpression) continue;
      if (exp instanceof LetExpression) continue;
      if (exp instanceof NotFormula)
      {
        NotFormula exp0 = (NotFormula)exp;
        Expression base = exp0.getBase();
        if (base instanceof QuantifiedExpression) continue;
        if (base instanceof LetExpression) continue;
      }
      // ignore unhelpful identities as counterexample
      if (exp instanceof EqualsFormula)
      {
        EqualsFormula exp0 = (EqualsFormula)exp;
        Expression exp1 = exp0.getFirst();
        Expression exp2 = exp0.getSecond();
        if (exp1 instanceof LambdaTerm && exp2 instanceof LambdaTerm)
          continue;
        if (exp1 instanceof Identifier && exp2 instanceof Identifier)
        {
          Identifier ident1 = (Identifier)exp1;
          Identifier ident2 = (Identifier)exp2;
          if (ident1.getName().equals(ident2.getName())) continue;
        }
        if (exp1.getClass() == exp2.getClass())
        {
          String text1 = ASTUtil.canonicalText(exp1);
          String text2 = ASTUtil.canonicalText(exp2);
          if (text1.equals(text2)) continue;
        }
      }
      result.add(exp);
    }
    if (result.size() == 0) return;
    counterExample = ASTUtil.toExpressionArray(result);
  }
  
  /****************************************************************************
   * Interpret justification as a counterexample.
   * @param cvcl the CVCL instance used for the interpretation.
   ***************************************************************************/
  private void makeCounterExampleDecls(CVCL cvcl)
  {
    String string = getJustification();
    if (string == null) return;
    Vector<Expression> result = new Vector<Expression>();
    int pos = 0;
    int n = string.length();
    while (pos < n)
    {
      int begin = pos;
      int end = string.indexOf(";", pos);
      if (end == -1) return;
      pos = end+1;
      String dstring = string.substring(begin, end);
      Declaration decl = cvcl.parseDeclaration(dstring);
      if (decl == null) return;
      if (!(decl instanceof ValueDeclaration))
        Main.internalError("counterexample is not a value declaration");
      ValueDeclaration decl0 = (ValueDeclaration)decl;
      ValueSymbol symbol = (ValueSymbol)decl0.getSymbol();
      Type type = Checking.canonical(decl0.getType());
      symbol.setType(type);
      EqualsFormula equality =
          Construct.equalsFormula(
              Construct.reference(symbol), decl0.getValue());
      result.add(equality);
    }
    if (result.size() == 0) return;
    counterExample = ASTUtil.toExpressionArray(result);
  }
}