// ----------------------------------------------------------------------------
// Substitute.java
// Substitute in an expression all free occurences of denoted variables.
// $Id: Substitute.java,v 1.13 2015/09/17 09:31:03 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Proving;

import fmrisc.ProofNavigator.Semantics.*;
import fmrisc.ProofNavigator.Syntax.*;

import java.util.*;

/**
 * 
 * Substitute in an expression all free occurences of denoted variables.
 *
 */
public final class Substitute extends ASTVisitorBase
{
  /****************************************************************************
   * Return expression with simultaneous variable substitutions performed.
   * @param exp an expression.
   * @param ident an array of identifiers.
   * @param subst an array of expressions of the same length as ident;
   *        if repeat is true, none of the expressions contains a free 
   *        occurence of any identifier in ident.
   * @param repeat true iff the substitution is to be iterated until
   *        no more applicable.
   * @return exp[subst/ident], i.e., a version of exp where all free 
   * occurences of any element of ident have been substituted by the 
   * corresponding expression in subst with some locally bound variables in 
   * exp possibly renamed to avoid name capture.
   ***************************************************************************/
  public static Expression substitute(Expression exp, 
    ValueDeclIdentifier[] ident, Expression[] subst, boolean repeat)
  {
    // perform at least one substitution (most likely needed)
    Vector<ValueSymbol> fv = null;
    do
    {
      exp = (Expression)exp.accept(new Substitute(ident, subst));
      if (!repeat) return exp;
      fv = FreeVariables.compute(exp, true);
    }
    while (occur(fv, ident));
    return exp;
  }
  
  /****************************************************************************
   * Determine whether symbol names occur among given identifiers.
   * @param symbols a vector of ValueSymbol objects.
   * @param idents an array of identifiers.
   * @return true if any symbol in symbol has a name that occurs among idents.
   ***************************************************************************/
  private static boolean occur(Vector<ValueSymbol> symbols, ValueDeclIdentifier[] idents)
  {
    int n = idents.length;
    for (int i=0; i<n; i++)
    {
      String name = idents[i].getName();
      int l = symbols.size();
      for (int j=0; j<l; j++)
      {
        ValueSymbol symbol = (ValueSymbol)symbols.elementAt(j);
        if (name.equals(symbol.getIdentifier().getName())) return true;
      }
    }
    return false;
  }
  
  // info for the substitution itself
  private String[] name;      // names of identifiers to be substituted
  private Expression[] subst; // the substitution expressions
  private boolean[] free;     // are identifiers still free for substitution?
  
  // the names of the identifiers occuring in the substitution expressions
  private String[] oldName; // the names occuring in the expression
  private String[] newName; // the new names for bound variables with same names
  
  /***************************************************************************
   * 
   * Auxiliary functions.
   *
   **************************************************************************/
  
  /**************************************************************************
   * Create visitor for substituting identifiers by expressions.
   * @param ident the array of identifiers
   * @param subst an array of same length as ident holding the substitutions.
   *************************************************************************/
  private Substitute(ValueDeclIdentifier[] ident, Expression[] subst)
  {
    // info for the core substitution
    int n = ident.length;
    this.name = new String[n];
    this.subst = subst;
    this.free = new boolean[n];
    Vector<ValueSymbol> fvars = new Vector<ValueSymbol>();
    for (int i=0; i<n; i++)
    {
      this.name[i] = ident[i].getName();
      this.free[i] = true;
      FreeVariables.append(subst[i], fvars, true);
    }
    
    // info for renaming bound variables
    int m = fvars.size();
    oldName = new String[m];
    newName = new String[m];
    for (int i=0; i<m; i++)
    {
      ValueSymbol symbol = (ValueSymbol)fvars.elementAt(i);
      oldName[i] = symbol.getIdentifier().getName();
      newName[i] = null; // no renaming required up to now
    }
  }
  
  /****************************************************************************
   * Construct a suitable renaming for visiting a quantified expression.
   * @param bound a list of names bound by the expression.
   * @param exp the quantified expression.
   * @return a renaming suitable for processing the expression.
   ***************************************************************************/
  private String[] newRenaming(String[] bound, Expression exp)
  {
    Vector<ValueSymbol> fvars = FreeVariables.compute(exp, true);
    int n = oldName.length;  
    String[] renaming = new String[n];
    for (int i=0; i<n; i++)
    {
      String name = oldName[i];
      int j = search(bound, name);
      if (j == -1)
        renaming[i] = newName[i];
      else
        renaming[i] = newName(name, fvars); 
    }
    return renaming;
  }
  
  /***************************************************************************
   * Construct a new name for a bound variable.
   * @param name the name of the bound variable.
   * @param fvars alist of free variables (vector of ValueSymbol objects).
   * @return a new name that does not collide with the names in fvars
   *    and not with the free variables in the substitution expressions.
   **************************************************************************/
  private String newName(String name, Vector<ValueSymbol> fvars)
  {
    int ext = 0;
    while (true)
    {
      String name0 = name+ext;
      int pos = search(oldName, name0);
      if (pos == -1)
      {
        pos = search(fvars, name0);
        if (pos == -1) return name0;
      }
      ext = ext+1;
    }
  }
  
  /***************************************************************************
   * Search name in vector of free variables.
   * @param fvars a vector of free variables (ValueSymbol objects).
   * @param name a name.
   * @return the index of a symbol in fvars that has name (-1, if none).
   **************************************************************************/
  private static int search(Vector<ValueSymbol> fvars, String name)
  {
    int n = fvars.size();
    for (int i=0; i<n; i++)
    {
      ValueSymbol symbol = (ValueSymbol)fvars.elementAt(i);
      if (name.equals(symbol.getIdentifier().getName())) return i;
    }
    return -1;
  }
  
  /************************************************************************
   * Search for name in sequence of names.
   * @param seq a sequence of names.
   * @param name a variable name.
   * @return the index of name in seq (-1, if there is no such index).
   ***********************************************************************/
  private static int search(String[] seq, String name)
  {
    int n = seq.length;
    for (int i=0; i<n; i++)
    {
      if (seq[i].equals(name)) return i;
    }
    return -1;
  }
  
  /************************************************************************
   * 
   * The visitor functions.
   *
   ***********************************************************************/
  
  /************************************************************************
   * Signal that visitor shall clone tree.
   ***********************************************************************/
  public boolean isCloned()
  {
    return true;
  }
  
  /************************************************************************
   * Possibly rename or substitute reference
   * @param ident a reference.
   ***********************************************************************/
  protected AST visit(Reference ident) 
  {
    // only value references are affected
    Symbol symbol = ident.getSymbol();
    if (!(symbol instanceof ValueSymbol)) return null;

    // first check whether reference has to be renamed
    ValueSymbol vsymbol = (ValueSymbol)symbol;
    ValueDeclIdentifier vident = vsymbol.getValueDeclIdentifier();
    if (vident.getSubstitution() != null) return null;
    
    // now check whether reference denotes free variable with substitution
    // toString() gives the name of the ident symbol
    String iname = ident.toString(); // ident.getName();
    int pos = search(name, iname);
    if (pos == -1 || !free[pos]) return null;
    
    // we have a free variable for which there exists a substitution
    // the substitution is cloned as well
    return subst[pos].instantiate();
  }

  /****************************************************************************
   * Process selector identifier (which is not a free variable).
   * @param sel the selector identifier to be processed.
   * @return sel (signalling that no subtree remains to be visited).
   ***************************************************************************/
  protected AST visit(SelectorIdentifier sel)
  {
    // identifier is not free, so we are done
    return sel;
  }
  
  /************************************************************************
   * Take care of bound variables.
   ***********************************************************************/
  protected AST visit(QuantifiedExpression exp)
  {
    // create new renaming
    TypedIdentifier[] vars = exp.getVariables();
    int l = vars.length;
    String[] vnames = new String[l];
    for (int i=0; i<l; i++)
      vnames[i] = vars[i].getIdentifier().getName();
    String[] newName0 = newName;
    newName = newRenaming(vnames, exp);
    
    // clone variable list
    int n = free.length;
    boolean[] switched = new boolean[n];
    TypedIdentifier[] vars0 = new TypedIdentifier[l];
    for (int i=0; i<l; i++)
    {
      TypedIdentifier tident = vars[i];
      ValueDeclIdentifier ident = tident.getIdentifier();
      Type type = tident.getType();
      
      // clone identifier and potentially rename it
      ValueDeclIdentifier ident0 = (ValueDeclIdentifier)ident.accept(this);
      ident.setSubstitution(ident0);
      String name = ident0.getName();
      int j = search(oldName, name);
      if (j != -1) 
      { 
        name = newName[j];
        ident0.setName(name);
      }
      j = search(this.name, name);
      if (j != -1 && free[j])
      {
        free[j] = false;
        switched[j] = true;
      }
      
      // clone type and construct new typed identifier
      Type type0 = (Type)type.accept(this);
      vars0[i] = Construct.typedIdentifier(ident0, type0);
    }
    
    // clone base expression performing substitutions and renamings
    Expression base = exp.getBase();
    Expression base0 = (Expression)base.accept(this);
    
    // result is cloned quantified expression
    // (construct before resetting substitution
    //  since also hidden information may be substituted)
    QuantifiedExpression result = exp.construct(vars0, base0);
    
    // hack for dealing with replacement of array terms
    if (exp instanceof ArrayTerm)
    {
      ArrayTerm exp0 = (ArrayTerm)exp;
      ArrayTerm result0 = (ArrayTerm)result;
      result0.setConstant(exp0.getConstant());
      result0.setAxiom(exp0.getAxiom());
      Expression term0 = (Expression)exp0.getTerm().accept(this);
      result0.setTerm(term0); // also rename free vars in replacement term
      ValueDeclIdentifier[] free = exp0.getFree();
      int n0 = free.length;
      ValueDeclIdentifier[] free0 = new ValueDeclIdentifier[n0];
      for (int i=0; i<n0; i++)
      {
        free0[i] = (ValueDeclIdentifier)free[i].accept(this);
      }
      result0.setFree(free0);
    }
    
    // reset original state and return result
    newName = newName0;
    for (int j=0; j<l; j++)
      vars[j].getIdentifier().resetSubstitution();
    for (int j=0; j<n; j++)
    {
      if (switched[j]) free[j] = true;
    }
    return result;
  }
  
  /************************************************************************
   * Take care of bound variables.
   ***********************************************************************/
  protected AST visit(LetExpression exp)
  {
    // create new renaming
    ValueDeclaration[] decls = exp.getDeclarations();
    int l = decls.length;
    String[] vnames = new String[l];
    for (int i=0; i<l; i++)
      vnames[i] = decls[i].getIdentifier().getName();
    String[] newName0 = newName;
    newName = newRenaming(vnames, exp);
    
    // clone variable list
    int n = free.length;
    boolean[] switched = new boolean[n];
    ValueDeclaration[] decls0 = new ValueDeclaration[l];
    for (int i=0; i<l; i++)
    {
      ValueDeclaration decl = decls[i];
      ValueDeclIdentifier ident = decl.getValueDeclIdentifier();
      Type type = decl.getType();
      Expression value = decl.getValue();
      
      // clone identifier and potentially rename it
      ValueDeclIdentifier ident0 = (ValueDeclIdentifier)ident.accept(this);
      ident.setSubstitution(ident0);
      String name = ident0.getName();
      int j = search(oldName, name);
      if (j != -1) 
      { 
        name = newName[j];
        ident0.setName(name);
      }
      j = search(this.name, name);
      if (j != -1 && free[j])
      {
        free[j] = false;
        switched[j] = true;
      }
      
      // clone type and value and construct new typed identifier
      Type type0 = null;
      if (type != null) type0 = (Type)type.accept(this);
      Expression value0 = (Expression)value.accept(this);
      decls0[i] = Construct.valueDeclaration(ident0, type0, value0);
    }
    
    // clone base expression performing substitutions and renamings
    Expression base = exp.getBase();
    Expression base0 = (Expression)base.accept(this);
    
    // reset original state
    newName = newName0;
    for (int j=0; j<l; j++)
      decls[j].getValueDeclIdentifier().resetSubstitution();
    for (int j=0; j<n; j++)
    {
      if (switched[j]) free[j] = true;
    }
    
    // result is cloned let-expression
    return Construct.letExpression(decls0, base0);
  }
}
